unit SelectionBar;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, Buttons;

type
  TSelectionControlState = (bsUp, bsDisabled, bsDown);

  TSelectionControl = class(TGraphicControl)
  private
    FLayout : TButtonLayout;
    FFlat,
    FTransparent : Boolean;
    FMouseInControl: Boolean;
    FDown : Boolean;
    FState : TSelectionControlState;
    FFont : TFont;
    FImageIndex : Integer;
  protected
    procedure SetFlat(Value : Boolean);
    procedure SetTransparent(Value : Boolean);
    procedure SetDown(Value : Boolean);
    procedure SetFont(Value : TFont);
(*
    procedure WMLButtonDblClk(var Message: TWMLButtonDown); message WM_LBUTTONDBLCLK;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure CMSysColorChange(var Message: TMessage); message CM_SYSCOLORCHANGE;
*)
    procedure CMButtonPressed(var Message: TMessage); message CM_BUTTONPRESSED;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;

    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
(*
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
*)
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure Paint; override;

    property Parent;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Click; override;
    property Flat : Boolean read FFlat write SetFlat;
    property Font : TFont read FFont write SetFont;
    property Transparent : Boolean read FTransparent write SetTransparent;
    property Down : Boolean read FDown write SetDown;
    property ImageIndex : Integer read FImageIndex write FImageIndex;
    property Caption;
    property OnClick;
  end;

  TDirection = (dirUp, dirDown);

  TScrollControl = class(TSelectionControl)
  private
    FDirection : TDirection;
  public
    procedure Click; override;
    property Direction : TDirection read FDirection write FDirection;
  end;

  TSelectionBar = class(TWinControl)
  private
    FSelectionControls : TList;

    FScrollUp,
    FScrollDown : TScrollControl;

    FYOffs : Integer;

    FAutoScroll : Boolean;
    FBorderStyle : TBorderStyle;
    FPosition : Integer;
    FOnResize : TNotifyEvent;
    FImageList : TImageList;
  protected
    procedure SetAutoScroll(Value : Boolean);
    procedure SetBorderStyle(Value : TBorderStyle);
    procedure SetPosition(Value : Integer);
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure Resize; override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    function AddSelectionControl : TSelectionControl;
    procedure Clear;    
    procedure ScrollUp;
    procedure ScrollDown;
  published
    property Align;
    property AutoScroll: Boolean read FAutoScroll write SetAutoScroll default True;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Color nodefault;
    property Ctl3D;
    property Font;
    property ImageList : TImageList read FImageList write FImageList;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property Position : Integer read FPosition write SetPosition;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize: TNotifyEvent read FOnResize write FOnResize;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Win32', [TSelectionBar]);
end;

 //-----------------------------------------------------------------------------

constructor TSelectionControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetBounds(0, 0, 25, 25);
  ControlStyle := [csCaptureMouse, csOpaque, csDoubleClicks];
//  FGlyph := TButtonGlyph.Create;
//  TButtonGlyph(FGlyph).OnChange := GlyphChanged;
  ParentFont := True;
//  FSpacing := 4;
//  FMargin := -1;
  FLayout := blGlyphLeft;
  FFlat := True;
  FTransparent := False;
//  Inc(ButtonCount);
  FFont := TFont.Create;
  FImageIndex := -1;
end;

destructor TSelectionControl.Destroy;
begin
  FFont.Free;
  inherited Destroy;
end;

procedure TSelectionControl.Paint;
const
  DownStyles: array[Boolean] of Integer = (BDR_RAISEDINNER, BDR_SUNKENOUTER);
  FillStyles: array[Boolean] of Integer = (BF_MIDDLE, 0);
var
  PaintRect : TRect;
  DrawFlags: Integer;
  XOffs, YOffs, XAt, YAt : Integer;
  R : TRect;
begin
  Canvas.Font := Self.Font;
  PaintRect := Rect(0, 0, Width, Height);
  if not FFlat then
    begin
      DrawFlags := DFCS_BUTTONPUSH or DFCS_ADJUSTRECT;
      if FState = bsDown then
        DrawFlags := DrawFlags or DFCS_PUSHED;
      DrawFrameControl(Canvas.Handle, PaintRect, DFC_BUTTON, DrawFlags);
    end
  else
    begin
      Canvas.Font.Assign(FFont);
      if not Transparent or FMouseInControl then
        begin
          Canvas.Brush.Style := bsSolid;
          Canvas.Brush.Color := Color;
          Canvas.FillRect(PaintRect);
          Canvas.Font.Color := clBlack;
        end
      else
        begin
          Canvas.Brush.Style := bsSolid;
          Canvas.Brush.Color := TSelectionBar(Parent).Color;
          Canvas.FillRect(PaintRect);
        end;

      if (FState = bsDown) then
        begin XOffs := 1; YOffs := 1; end
      else
        begin XOffs := 0; YOffs := 0; end;

      if (TSelectionBar(Parent).ImageList <> Nil) and (ImageIndex >= 0) then
        begin
           // Has an image
          XAt := (Width-TSelectionBar(Parent).ImageList.Width) div 2 + XOffs;
          YAt := (Height-(TSelectionBar(Parent).ImageList.Height+Canvas.TextHeight(Caption)*2)) div 2 + YOffs;
          TSelectionBar(Parent).ImageList.Draw(Canvas, XAt, YAt, ImageIndex);

          YAt := YAt + TSelectionBar(Parent).ImageList.Height;
          R := Rect(XOffs+5, YAt, Width+XOffs-5, Height+YOffs);
          DrawText(Canvas.Handle, PChar(Caption), Length(Caption), R, DT_CENTER or DT_WORDBREAK);
          //Canvas.TextOut(XAt, YAt, Caption);
        end
      else
        begin
           // Does not have an image
          XAt := (Width-Canvas.TextWidth(Caption)) div 2 + XOffs;
          YAt := (Height-Canvas.TextHeight(Caption)) div 2 + YOffs;
          Canvas.TextOut(XAt, YAt, Caption);
        end;

      if (FState = bsDown) or
        (FMouseInControl and (FState <> bsDisabled)) or
        (csDesigning in ComponentState) then
        DrawEdge(Canvas.Handle, PaintRect, DownStyles[FState = bsDown],
          FillStyles[FFlat] or BF_RECT);
      InflateRect(PaintRect, -1, -1);
    end;

  if FState = bsDown then
    begin
      if (not FFlat or not FMouseInControl) then
      begin
//        if Pattern = nil then CreateBrushPattern;
//        Canvas.Brush.Bitmap := Pattern;
        Canvas.FillRect(PaintRect);
      end;
//      Offset.X := 1;
//      Offset.Y := 1;
    end
  else
    begin
//      Offset.X := 0;
//      Offset.Y := 0;
    end;
   // Draw glyph
end;

procedure TSelectionControl.SetFlat(Value : Boolean);
begin
  FFlat := Value;
end;

procedure TSelectionControl.SetTransparent(Value : Boolean);
begin
  FTransparent := Value;
end;

procedure TSelectionControl.SetFont(Value : TFont);
begin
  FFont.Assign(Value);
end;

procedure TSelectionControl.SetDown(Value: Boolean);
begin
  if Value <> FDown then
    begin
      if FDown then Exit;
      FDown := Value;
      if Value then
        begin
          if FState = bsUp then Invalidate;
          FState := bsDown
        end
      else
        begin
          FState := bsUp;
          Repaint;
        end;
    end;
end;

procedure TSelectionControl.Click;
begin
  inherited Click;
end;

procedure TSelectionControl.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if (Button = mbLeft) and Enabled then
  begin
    if not FDown then
    begin
      FState := bsDown;
      Invalidate;
    end;
  end;
end;

procedure TSelectionControl.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  DoClick: Boolean;
begin
  inherited MouseUp(Button, Shift, X, Y);
  DoClick := (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight);

  { Redraw face in-case mouse is captured }
  FState := bsUp;
  if DoClick then
    begin
      if FState <> bsDown then Invalidate;
      Click;
    end
  else
    Invalidate;
end;

procedure TSelectionControl.CMButtonPressed(var Message: TMessage);
var
  Sender: TSelectionControl;
begin
  Sender := TSelectionControl(Message.LParam);
  if Sender = Self then
    Down := True;
end;

procedure TSelectionControl.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  if FFlat and not FMouseInControl and Enabled then
  begin
    FMouseInControl := True;
    Repaint;
  end;
end;

procedure TSelectionControl.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if FFlat and FMouseInControl and Enabled then
  begin
    FMouseInControl := False;
    Invalidate;
  end;
end;

 //-----------------------------------------------------------------------------

procedure TScrollControl.Click;
begin
  if FDirection = dirUp then
    TSelectionBar(Parent).ScrollDown
  else
    TSelectionBar(Parent).ScrollUp;
end;

 //-----------------------------------------------------------------------------

constructor TSelectionBar.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := [csAcceptsControls, csCaptureMouse, csClickEvents,
    csSetCaption, csDoubleClicks];
  FAutoScroll := True;
  Align := alLeft;
  Width := 72;
  Height := 41;
  FBorderStyle := bsSingle;
  FYOffs := 0;

  Color := clBtnShadow;

  FScrollUp := TScrollControl.Create(Self);
  with FScrollUp do
    begin
      Parent := Self;
      Direction := dirUp;
      Color := clBtnFace;
      Flat := True;
      Height := 12;
      Top := 0;
      Visible := False;
      Caption := 't';
      Font.Name := 'Marlett';
      Font.Size := 8;      
    end;

  FScrollDown := TScrollControl.Create(Self);
  with FScrollDown do
    begin
      Parent := Self;
      Direction := dirDown;
      Color := clBtnFace;
      Flat := True;
      Height := 12;
      Caption := 'u';
      Font.Name := 'Marlett';
      Font.Size := 8;
    end;

  FSelectionControls := TList.Create;
end;

destructor TSelectionBar.Destroy;
begin
  Clear;
  FSelectionControls.Free;

  FScrollDown.Free;
  FScrollUp.Free;
  inherited Destroy;
end;

procedure TSelectionBar.SetAutoScroll(Value : Boolean);
begin
end;

procedure TSelectionBar.SetBorderStyle(Value : TBorderStyle);
begin
  if Value <> FBorderStyle then
  begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TSelectionBar.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TBorderStyle] of DWORD = (0, WS_BORDER);
begin
  inherited CreateParams(Params);
  with Params do
    begin
      Style := Style or BorderStyles[FBorderStyle];
      WindowClass.style := WindowClass.style or CS_HREDRAW or CS_VREDRAW;
      if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then
      begin
        Style := Style and not WS_BORDER;
        ExStyle := ExStyle or WS_EX_CLIENTEDGE;
      end;
    end;
end;

procedure TSelectionBar.Resize;
var
  I : Integer;
begin
  FScrollUp.Width := ClientWidth;
  FScrollDown.Width := ClientWidth;
  FScrollDown.Top := ClientHeight-FScrollDown.Height;
  For I := 0 to FSelectionControls.Count-1 do
    TSelectionControl(FSelectionControls[I]).Width := ClientWidth;
  FScrollDown.Visible :=
     (FSelectionControls.Count > 0) and
     (TSelectionControl(FSelectionControls[FSelectionControls.Count-1]).Top +
      TSelectionControl(FSelectionControls[FSelectionControls.Count-1]).Height >= ClientHeight);
  if Assigned(FOnResize) then FOnResize(Self);
end;

procedure TSelectionBar.WMSize(var Message: TWMSize);
begin
  inherited;
  if not (csLoading in ComponentState) then Resize;
end;

function TSelectionBar.AddSelectionControl : TSelectionControl;
var
  I,
  YPos : Integer;
begin
  YPos := 0;
  For I := 0 to FSelectionControls.Count-1 do
    Inc(YPos, TSelectionControl(FSelectionControls[I]).Height);

  Result := TSelectionControl.Create(Self);
  with Result do
    begin
      Parent := Self;
      Color := clBtnFace;
      Caption := 'Unknown' + IntToStr(FSelectionControls.Count);
      Flat := True;
      Transparent := True;
      Top := YPos;
      Left := 0;
      Width := Self.ClientWidth;
      Height := 64;
      Font.Color := clWhite;
      ImageIndex := -1;
    end;
  FSelectionControls.Add(Result);

  FScrollDown.Visible :=
     (FSelectionControls.Count > 0) and
     (TSelectionControl(FSelectionControls[FSelectionControls.Count-1]).Top +
      TSelectionControl(FSelectionControls[FSelectionControls.Count-1]).Height >= ClientHeight);

  FScrollUp.BringToFront;
  FScrollDown.BringToFront;
end;

procedure TSelectionBar.Clear;
var
  I : Integer;
begin
  For I := 0 to FSelectionControls.Count-1 do
    TSelectionControl(FSelectionControls[I]).Free;

  While FSelectionControls.Count > 0 do
    FSelectionControls.Delete(0);
end;

procedure TSelectionBar.ScrollUp;
var
  I, Shift, ATop : Integer;
begin
  ATop := FScrollUp.Height;

   // Determine the shift to get the next control to to the top
  I := 0;
  while (I < FSelectionControls.Count) and
        (TSelectionControl(FSelectionControls[I]).Top <= ATop) do
    Inc(I);
  if I >= FSelectionControls.Count then Exit;
  Shift := TSelectionControl(FSelectionControls[I]).Top - ATop;

   // Determine the bottom control
  while (I < FSelectionControls.Count) and
        (TSelectionControl(FSelectionControls[I]).Top - Shift +
         TSelectionControl(FSelectionControls[I]).Height < ClientHeight) do
    Inc(I);

   // If the bottom block will be above the bottom then...
  if (I >= FSelectionControls.Count) then
    begin
      I := FSelectionControls.Count-1;
       // ... if it already IS above the bottom, then don't do anything.
      if (TSelectionControl(FSelectionControls[I]).Top +
          TSelectionControl(FSelectionControls[I]).Height < ClientHeight) then
        Exit
       // ... otherwise, shift the bottom control snugly to the bottom.
      else
        begin
          Shift :=  TSelectionControl(FSelectionControls[I]).Top +
                    TSelectionControl(FSelectionControls[I]).Height -
                    ClientHeight;
          FScrollDown.Visible := False;
        end;
    end;

   // Make it so
  ScrollBy(0, -Shift);
  FScrollUp.Visible := True;

  FScrollUp.Top := 0;
  FScrollDown.Top := ClientHeight-FScrollDown.Height;
end;

procedure TSelectionBar.ScrollDown;
var
  I, Shift, ATop : Integer;
begin
  if FScrollUp.Visible then ATop := FScrollUp.Height else ATop := 0;

   // Determine the shift to get the previous control to to the top
  I := 0;
  while (I < FSelectionControls.Count) and
        (TSelectionControl(FSelectionControls[I]).Top < ATop) do
    Inc(I);
  if I-1 < 0 then Exit;
  if (I-1 = 0) then
    begin
      FScrollUp.Visible := False;
      Shift := TSelectionControl(FSelectionControls[I-1]).Top;
    end
  else
    Shift := TSelectionControl(FSelectionControls[I-1]).Top - ATop;

   // Determine the bottom control
  while (I < FSelectionControls.Count) and
        (TSelectionControl(FSelectionControls[I]).Top - Shift +
         TSelectionControl(FSelectionControls[I]).Height < ClientHeight) do
    Inc(I);

  FScrollDown.Visible :=
    (I < FSelectionControls.Count) and 
    (TSelectionControl(FSelectionControls[I]).Top - Shift +
     TSelectionControl(FSelectionControls[I]).Height > ClientHeight);

   // Make it so
  ScrollBy(0, -Shift);

  FScrollUp.Top := 0;
  FScrollDown.Top := ClientHeight-FScrollDown.Height;
end;

procedure TSelectionBar.SetPosition(Value : Integer);
begin
  if Value < 0 then Value := 0;
  if Value <> FPosition then
    begin
      ScrollBy(0, Value - FPosition);
      FPosition := Value;

      FScrollUp.Top := 0;
      FScrollDown.Top := ClientHeight-FScrollDown.Height;
    end;
end;

end.
