unit UOMapComp;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, UOMap, MemMapFile;

type
  TMapZoom = (zm1x1, zm1x8);
  TMapFlags = set of (flXRay, flStatics, flNulWater);
  TMapAlt = (asNormal, asNormalWithAlt, asAltitude);
  TMapMouseEvent = procedure(Sender: TObject; Button: TMouseButton;
                             Shift: TShiftState; X, Y, AX, AY : Integer)
                             of object;

  // Core map view area in blocks structure.
  RMapViewArea = record
    Left, Top, Width, Height : Integer;
  end;

  TMapViewArea = class
  public
    constructor Create(AArea : RMapViewArea);
    function GetArea : RMapViewArea;
    function GetBoundingRect(AZoom : TMapZoom) : TRect;    
  private
    FArea : RMapViewArea;
  end;

  TUOMap = class(TCustomPanel)
  private
    { Private declarations }
   {$IfNDef UseMMF}
    MapStr : TFileStream;
    Statics : TFileStream;
    MapDifLoc : TFileStream;
    MapDif : TFileStream;
    StaDifLoc : TFileStream;
    StaDif : TFileStream;
    StaticsDif : TFileStream;
   {$Else}
    MapStr : TMMFStream;
    Statics : TMMFStream;  //TFileStream;
   {$EndIf}
    StaticsIndexStr : TMemoryStream;        // 4.6 meg
    ColorLookupFilled : Boolean;
    ColorLookup : Array[0..65535] of Word;  // 131 K

    PB : TPaintBox;
    ScrollBox : TScrollBox;

    FMapPath : String;
    FStaticsIndexPath : String;
    FStaticsPath : String;
    FRadarPath : String;
    FMapDifLocPath : String;
    FMapDifPath : String;
    FStaDifLocPath : String;
    FStaDifPath : String;
    FStaticsDifPath : String;
    FZoom : TMapZoom;
    FFlags : TMapFlags;
    FAltSetting : TMapAlt;
    FOnMapClick : TMapMouseEvent;

    FCurrentView : Integer;
    FCurrentArea : TMapViewArea;
    FOverallArea : RMapViewArea;
  protected
    { Protected declarations }
    function CheckFiles : Boolean;
    procedure SetMapPath(Value : String);
    procedure SetStaticsIndexPath(Value : String);
    procedure SetStaticsPath(Value : String);
    procedure SetRadarPath(Value : String);
    procedure SetMapDifLocPath(Value : String);
    procedure SetMapDifPath(Value : String);
    procedure SetStaDifLocPath(Value : String);
    procedure SetStaDifPath(Value : String);
    procedure SetStaticsDifPath(Value : String);
    procedure SetZoom(Value : TMapZoom);
    procedure SetFlags(Value : TMapFlags);
    procedure SetAltSetting(Value : TMapAlt);
    procedure DoPaint(Sender : TObject);
    procedure DoMouseDown(Sender: TObject; Button: TMouseButton;
                          Shift: TShiftState; X, Y: Integer);
    procedure AddStaticsToBuffer(AZoom : TMapZoom; AFlags : TMapFlags; var BlBuffer : Array of TBlock; Block, Count : Integer);
//    function GetCellColor(AZoom : TMapZoom; AFlags : TMapFlags; Cell : TCell) : Word;
//    function GetBlockColor(AZoom : TMapZoom; AFlags : TMapFlags; Block : TBlock) : Word;
  public
    // Hate to make this public, but MapSave has to get to it.
    FViews : TStringList;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Save(AFileName : String; ARect : TRect; AFlags : TMapFlags;
                   AAlt : TMapAlt; AArea : TMapViewArea; AZoom : TMapZoom);
    procedure GetStaticsAt(X, Y : Integer; List : TList);
    procedure OutputToBitmap(ARect : TRect; AFlags : TMapFlags;
                             AAlt : TMapAlt; AZoom : TMapZoom;
                             ABitmap : TBitmap);
    procedure ResetViews;
    procedure AddView(const AName : String; const AView : RMapViewArea);
    procedure ResetMap(AView : Integer);
    procedure SetOverallArea(const AMapViewArea : RMapViewArea);
  published
    { Published declarations }
    property MapPath : String read FMapPath write SetMapPath;
    property StaticsIndexPath : String read FStaticsIndexPath write SetStaticsIndexPath;
    property StaticsPath : String read FStaticsPath write SetStaticsPath;
    property RadarPath : String read FRadarPath write SetRadarPath;
    property MapDifLocPath : String read FMapDifLocPath write SetMapDifLocPath;
    property MapDifPath : String read FMapDifPath write SetMapDifPath;
    property StaDifLocPath : String read FStaDifLocPath write SetStaDifLocPath;
    property StaDifPath : String read FStaDifPath write SetStaDifPath;
    property StaticsDifPath : String read FStaticsDifPath write SetStaticsDifPath;
    property Zoom : TMapZoom read FZoom write SetZoom;
    property Flags : TMapFlags read FFlags write SetFlags;
    property AltSetting : TMapAlt read FAltSetting write SetAltSetting;
    property OnMapClick : TMapMouseEvent read FOnMapClick write FOnMapClick;
  end;

  function ParseArea(Str : String) : RMapViewArea;

implementation

uses Utility;

function ParseArea(Str : String) : RMapViewArea;
type
  TState = (Left, Top, Width, Height);
var
  State : TState;
  C : Char;
  S : String;
  I : Integer;
begin
  Result.Left := 0;
  Result.Top := 0;
  Result.Width := 0;
  Result.Height := 0;
  I := 1;
  S := '';
  State := Left;
  repeat
    if I <= Length(Str) then
      C := Str [I]
    else
      C := #00;
    if C in [',',#00] then
      case State of
        Left :   begin Result.Left := StrToInt(S); S := ''; State := Top; end;
        Top :    begin Result.Top := StrToInt(S); S := ''; State := Width; end;
        Width :  begin Result.Width := StrToInt(S); S := ''; State := Height; end;
        Height : begin Result.Height := StrToInt(S); S := ''; Exit; end;
      end
    else
      S := S + C;
    Inc(I);
  until C = #00;
end;

constructor TMapViewArea.Create(AArea : RMapViewArea);
begin
  FArea := AArea;
end;

function TMapViewArea.GetArea : RMapViewArea;
begin
  Result := FArea;
end;

function TMapViewArea.GetBoundingRect(AZoom : TMapZoom) : TRect;
begin
  case AZoom of
    zm1x1 :
      begin
        Result.Left := FArea.Left * 8;
        Result.Top := FArea.Top * 8;
        Result.Right := Result.Left + FArea.Width * 8;
        Result.Bottom := Result.Top + FArea.Height * 8;
      end;
    zm1x8:
      begin
        Result.Left := FArea.Left;
        Result.Top := FArea.Top;
        Result.Right := Result.Left + FArea.Width;
        Result.Bottom := Result.Top + FArea.Height;
      end;
  end;
end;

constructor TUOMap.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FViews := TStringList.Create;

  FFlags := [flStatics];

  ScrollBox := TScrollBox.Create(Nil);
  with ScrollBox do
    begin
      Parent := Self;
      Color := clBlack;
      Align := alClient;
    end;

  PB := TPaintBox.Create(Nil);
  with PB do
    begin
      Parent := ScrollBox;
      Left := 0;
      Top := 0;
      BevelOuter := bvNone;
      with Canvas.Brush do
        begin
          Style := bsSolid;
          Color := clBlack;
        end;
      OnPaint := DoPaint;
      OnMouseDown := DoMouseDown;
    end;
end;

destructor TUOMap.Destroy;
begin
  ResetViews;
  FViews.Free; FViews := Nil;

  // Free up open file streams
  MapStr.Free; MapStr := Nil;
  StaticsIndexStr.Free; StaticsIndexStr := Nil;
  Statics.Free; Statics := Nil;
  MapDifLoc.Free; MapDifLoc := Nil;
  MapDif.Free; MapDif := Nil;
  StaDifLoc.Free; StaDifLoc := Nil;
  StaDif.Free; StaDif := Nil;
  StaticsDif.Free; StaticsDif := Nil;

  ColorLookupFilled := False;
  PB.Free;
  ScrollBox.Free;
  inherited Destroy;
end;

procedure TUOMap.ResetViews;
var
  I : Integer;
begin
  for I := 0 to FViews.Count-1 do
    FViews.Objects[I].Free;
  FViews.Clear;
end;

procedure TUOMap.AddView(const AName : String; const AView : RMapViewArea);
var
  Area : TMapViewArea;
begin
  Area := TMapViewArea.Create(AView);
  FViews.AddObject(AName, Area);
end;

procedure TUOMap.ResetMap(AView : Integer);
begin
  FCurrentView := AView;
  FCurrentArea := TMapViewArea(FViews.Objects[AView]);
  PB.BoundsRect := FCurrentArea.GetBoundingRect(FZoom);
  Invalidate;
end;

procedure TUOMap.SetOverallArea(const AMapViewArea : RMapViewArea);
begin
  FOverallArea := AMapViewArea;
end;

procedure TUOMap.AddStaticsToBuffer(AZoom : TMapZoom; AFlags : TMapFlags; var BlBuffer : Array of TBlock; Block, Count : Integer);
var
  StaticIndex : TStaticIndex;
  I,J : Integer;
  Cell : ^TCell;
  StCount : Integer;
  Buffer2 : Array[0..1023] of TStaticPosition;
  StPos : TStaticPosition;
begin
  StaticsIndexStr.Seek(Block * SizeOf(TStaticIndex), soFromBeginning);
  For I := 0 to Count-1 do
    begin
      StaticsIndexStr.Read(StaticIndex, SizeOf(TStaticIndex));
      if StaticIndex.Length > 0 then
        begin
          Statics.Seek(StaticIndex.Lookup, soFromBeginning);
          StCount := StaticIndex.Length div SizeOf(TStaticPosition);
          If StCount > 1024 then StCount := 1024;
          Statics.Read(Buffer2, StCount * SizeOf(TStaticPosition));
          For J := 0 to StCount-1 do
            begin
              StPos := Buffer2[J];
              Cell := @BlBuffer[I].Cells[StPos.Y*8+StPos.X];

              if flXRay in AFlags then
                begin
                  If StPos.Altitude <= Cell^.Altitude then
                    begin
                      Cell^.Value := StPos.BlockNum + 16384;
                      Cell^.Altitude := StPos.Altitude;
                    end;
                end
              else
                begin
                  If StPos.Altitude >= Cell^.Altitude then
                    begin
                      Cell^.Value := StPos.BlockNum + 16384;
                      Cell^.Altitude := StPos.Altitude;
                    end;
                end;
            end;
        end;
    end;
end;

function GetCellColor(var ColorLookup : Array of Word;
                      AZoom : TMapZoom; AFlags : TMapFlags;
                      AAltSetting : TMapAlt; Cell : TCell) : Word;
 // Retrieves the color for a specified Cell.  This will only take into
 // account the Altitude flag, and return either a radar color or an
 // altitude color for the specified cell.
var
  BlockVal : Integer;
begin
  if AAltSetting = asAltitude then
    begin
      BlockVal := Cell.Value;
      if (flNulWater in AFlags) and
         (((BlockVal >= 168) and (BlockVal <= 171)) or
          ((BlockVal >= 16384+6038) and (BlockVal <= 16384+6066))) then
        Result := $0
      else
        Result := Cell.Altitude + 128;
    end
  else
    begin
      BlockVal := Cell.Value;
      if (flNulWater in AFlags) and
         (((BlockVal >= 168) and (BlockVal <= 171)) or
          ((BlockVal >= 16384+6038) and (BlockVal <= 16384+6066))) then
        Result := $0
      else
        Result := ColorLookup[BlockVal];
    end;
end;

function GetBlockColor(var ColorLookup : Array of Word;
                       AZoom : TMapZoom; AFlags : TMapFlags;
                       AAltSetting : TMapAlt; Block : TBlock) : Word;
 // Retrieves the average color for a specified block.  This will only take
 // into account the Altitude flag, and return either an average radar color
 // or an average altitude color for the specified cell.
var
  Red, Green, Blue : Integer;
  Height : Integer;
  I : Integer;
  ColorVal : Word;
begin
  if AAltSetting = asAltitude then
    begin
      Height := 0;
      For I := 0 to 63 do
        Inc(Height, GetCellColor(ColorLookup, AZoom, AFlags, AAltSetting,
                                 Block.Cells[I]));
      Result := Height div 64;
    end
  else
    begin
      Red := 0; Green := 0; Blue := 0;
      For I := 0 to 63 do
        begin
          ColorVal := GetCellColor(ColorLookup, AZoom, AFlags, AAltSetting,
                                   Block.Cells[I]);
          Inc(Red, (ColorVal shr 10) and $1F);
          Inc(Green, (ColorVal shr 5) and $1F);
          Inc(Blue, ColorVal and $1F);
        end;
      Result := ((Red div 64) shl 10) +
                ((Green div 64) shl 5) +
                (Blue div 64);
    end;
end;

function Lighting(C : Word; A1, A2 : ShortInt) : Word;
var
  Slope : Integer;
  Red, Green, Blue : Byte;
begin
  Slope := A2 - A1 + 256; { 0 - 512 }
  If Slope <= 256 then
    begin
      Slope := Slope - (256-4);
      if Slope < 0 then Slope := 0;
      Slope :=  49152 * Slope div 4;  { 0 -> 1 (shl 16) }
    end
  else
    begin
      Dec(Slope, 256);
      If Slope > 4 then Slope := 4;
      Slope := 49152 * Slope div 4 + 65536;  { 0 -> 1.2 (shl 16) }
    end;

  Red := (((C shr 10) and $1F) * Slope) shr 16; If Red > 31 then Red := 31;
  Green := (((C shr 5) and $1F) * Slope) shr 16; If Green > 31 then Green := 31;
  Blue := ((C and $1F) * Slope) shr 16; if Blue > 31 then Blue := 31;
  Result := (Red shl 10) + (Green shl 5) + Blue;
end;

type
  TPixelBlock = Array[0..63] of Word;

function ProcessColor(C : Word; Rise : Integer) : Word;
 // Adjusts a color based on its rise.  It is assumed that the
 // rise is a 16:16 fixed point value where 0 is perfectly flat,
 // > 0 is rising and < 0 is falling.
const
  MAXSL = 10;  //3
  MAXSSL = MAXSL * 65536;
var
  Red, Green, Blue : Byte;
  Mult : Integer;
begin
  if Rise < -MAXSSL then Rise := -MAXSSL
  else if Rise > MAXSSL then Rise := MAXSSL;
  Mult := Rise div MAXSL;
//  if Rise >= 0 then
//    Mult := Mult + Mult div 2  // Mult = Mult * 1.5
//  else
//    Mult := Mult - Mult div 2; // Mult = Mult * 0.5
  Mult := 65536 + Mult;
  If Mult < 0 then Mult := 0;

  Red := (((C shr 10) and $1F) * Mult) shr 16; If Red > 31 then Red := 31;
  Green := (((C shr 5) and $1F) * Mult) shr 16; If Green > 31 then Green := 31;
  Blue := ((C and $1F) * Mult) shr 16; if Blue > 31 then Blue := 31;
  Result := (Red shl 10) + (Green shl 5) + Blue;
end;

procedure ProcessQuad(var C1, C2, C3, C4 : Word; A1, A2, A3, A4 : ShortInt);
 // Processes a square of four colors and adjusts their values depending on
 // their altitudes.
var
  Slope1, Slope2 : Integer;
begin
//  Slope1 := (A3 + A2) * 65536 div 2 - A1 * 65536;
//  Slope2 := A4 * 65536 - (A3 + A2) * 65536 div 2;

  Slope1 := (A2-A1) * 65536;
  Slope2 := (A4-A3) * 65536;

  C1 := ProcessColor(C1, Slope1);
  C2 := ProcessColor(C2, Slope1);
  C3 := ProcessColor(C3, Slope2);
  C4 := ProcessColor(C4, Slope2);
end;

procedure ProcessBlock(var ColorLookup : Array of Word;
                       AZoom : TMapZoom; AFlags : TMapFlags;
                       AAltSetting : TMapAlt; Block : TBlock;
                       var PixelBlock : TPixelBlock);
var
  I, J : Integer;
begin
  For I := 0 to 63 do
    PixelBlock[I] := GetCellColor(ColorLookup, AZoom, AFlags, AAltSetting, Block.Cells[I]);

  if AAltSetting = asNormalWithAlt then
    For I := 0 to 15 do
      begin
        J := (I shr 2) * 16 + (I and $03) * 2;
        ProcessQuad(PixelBlock[J], PixelBlock[J+1],
                    PixelBlock[J+8], PixelBlock[J+9],
                    Block.Cells[J].Altitude, Block.Cells[J+1].Altitude,
                    Block.Cells[J+8].Altitude, Block.Cells[J+9].Altitude);
      end;
end;

procedure TUOMap.OutputToBitmap(ARect : TRect; AFlags : TMapFlags;
                                AAlt : TMapAlt; AZoom : TMapZoom;
                                ABitmap : TBitmap);
var
  I : Integer;
  StColumn, EdColumn : Integer;
  StRow, EdRow : Integer;
  BlockCount : Integer;
  Block : Integer;
  XOfs : Integer;
  X, Y : Integer;
  ColorVal : Word;
  Line : Pointer;
  XLeftOfs, XRightOfs, YTopOfs, YBottomOfs : Integer;
  BlBuffer : Array of TBlock;  // 102K
  Palette : TMaxLogPalette;
  PixelBlock : TPixelBlock;
  WidthInBlocks : Integer;
  HeightInBlocks : Integer;
begin
  SetLength(BlBuffer, FOverallArea.Height);

  if AAlt = asAltitude then
    begin
      Palette.palVersion := $300;
      Palette.palNumEntries := 256;
      For I := 0 to 255 do with Palette.palPalEntry[I] do
        begin
          peRed := I;
          peGreen := I;
          peBlue := I;
          peFlags := 0;
        end;
      ABitmap.PixelFormat := pf8bit;
      ABitmap.Palette := CreatePalette(PLogPalette(@Palette)^);
    end
  else
    ABitmap.PixelFormat := pf15bit;

  StColumn := 0;
  EdColumn := 0;
  StRow := 0;
  EdRow := 0;

  case AZoom of
    zm1x1 :
      begin
        StColumn := ARect.Left div 8;
        EdColumn := (ARect.Right+7) div 8;
        StRow := ARect.Top div 8;
        EdRow := (ARect.Bottom+7) div 8;
        ABitmap.Width := (EdColumn-StColumn)*8;
        ABitmap.Height := (EdRow-StRow)*8;
      end;
    zm1x8 :
      begin
        StColumn := ARect.Left;
        EdColumn := ARect.Right;
        if EdColumn-StColumn > FOverallArea.Width then EdColumn := FOverallArea.Width;
        StRow := ARect.Top;
        EdRow := ARect.Bottom;
        if EdRow-StRow > FOverallArea.Height then EdRow := StRow + FOverallArea.Height;
        ABitmap.Width := EdColumn-StColumn;
        ABitmap.Height := EdRow-StRow;
      end;
  end;
  BlockCount := EdRow-StRow;

  XLeftOfs := ARect.Left - StColumn * 8;
  XRightOfs := ARect.Right - StColumn * 8;

  For I := StColumn to EdColumn-1 do
    begin
      Block := I * FOverallArea.Height + StRow;
      XOfs := (I-StColumn) * 8;

//TODO:
{
mapdifl<X>.mul
--------------
Description: Provides the location to apply the replacement map blocks.
Format:
	4 byte Records - Provides the 8x8 block number to replace in the map<X>.mul file.


mapdif<X>.mul
-------------
Discription: The replacement blocks to be placed at the locations provided from the file above.
Format:
	196 byte Records - Replacement Map Block (same format as map<X>.mul blocks).


stadifl<X>.mul
--------------
Description: Provides the block location to apply the replacement Statics.
Format:
	4 byte Records - Provides the 8x8 block number to replace in the STAIDX<X>.mul file.

stadifi<X>.mul
--------------
Description: Replacement Static Indexes
Format:
	12 byte Records - Simular to staidx<X>.mul
		DWORD is the Start position (0xFFFFFFFF if no static objects exist for this block)
		DWORD is the Length of the data segment
		DWORD is of unknown use 

stadif<X>.mul
-------------
Description: Replacement Static Objects
Format:
	Simular to staticsX.mul.
}

      MapStr.Seek(Block * SizeOf(TBlock), soFromBeginning);
      MapStr.Read(BlBuffer[0], BlockCount * SizeOf(TBlock));

      if flStatics in AFlags then
        AddStaticsToBuffer(AZoom, AFlags, BlBuffer, Block, BlockCount);

      case AZoom of
        zm1x1 :
          begin
            YTopOfs := ARect.Top - StRow * 8;
            YBottomOfs := ARect.Bottom - StRow * 8;

            For Y := 0 to BlockCount * 8-1 do
              if (Y >= YTopOfs) and (Y < YBottomOfs) then
                begin
                  Line := ABitmap.ScanLine[Y-YTopOfs];
                  ProcessBlock(ColorLookup, AZoom, AFlags, AAlt, BlBuffer[Y shr 3], PixelBlock);

                  For X := 0 to 7 do
                    if (XOfs + X >= XLeftOfs) and (XOfs + X < XRightOfs) then
                      begin
                        if AAlt = asAltitude then
                          PByteArray(Line)^[(XOfs + X)-XLeftOfs] :=
                            PixelBlock[(Y and $07)*8+X]
                        else
                          PWordArray(Line)^[(XOfs + X)-XLeftOfs] :=
                            PixelBlock[(Y and $07)*8+X];
                      end;
                end;
          end;
        zm1x8 :
          begin
            For Y := 0 to BlockCount-1 do
              begin
                ColorVal := GetBlockColor(ColorLookup, AZoom, AFlags, AAlt, BlBuffer[Y]);
                Line := ABitmap.ScanLine[Y];
                if AAlt = asAltitude then
                  PByteArray(Line)^[I-StColumn] := ColorVal
                else
                  PWordArray(Line)^[I-StColumn] := ColorVal;
              end;
          end;
      end; // case
    end; // for
end;

procedure TUOMap.Save(AFileName : String; ARect : TRect; AFlags : TMapFlags;
                      AAlt : TMapAlt; AArea : TMapViewArea; AZoom : TMapZoom);
var
  Bitmap : TBitmap;
begin
  if not CheckFiles then
    raise Exception.Create('All necessary files have not been set for Save.');

  Bitmap := TBitmap.Create;
  try
    OutputToBitmap(ARect, AFlags, AAlt, AZoom, Bitmap);
    Bitmap.SaveToFile(AFileName);
  finally
    Bitmap.Free;
  end;
end;

procedure TUOMap.GetStaticsAt(X, Y : Integer; List : TList);
var
  SP : ^TStaticPosition;
  StCount : Integer;
  Buffer2 : Array[0..1023] of TStaticPosition;
  StaticIndex : TStaticIndex;
  Block : Integer;
  I : Integer;
  ABlock : TBlock;
begin
  Block := (X div 8) * FOverallArea.Height + (Y div 8);

  MapStr.Seek(Block * SizeOf(TBlock), soFromBeginning);
  MapStr.Read(ABlock, SizeOf(TBlock));
  GetMem(SP, SizeOf(TStaticPosition));
  with ABlock.Cells[(Y and $07) * 8 + (X and $07)] do
    begin
      SP^.BlockNum := Value;
      SP^.Altitude := Altitude;
    end;
  List.Add(SP);

  StaticsIndexStr.Seek(Block * SizeOf(TStaticIndex), soFromBeginning);
  StaticsIndexStr.Read(StaticIndex, SizeOf(TStaticIndex));
  if StaticIndex.Length > 0 then
    begin
      Statics.Seek(StaticIndex.Lookup, soFromBeginning);
      StCount := StaticIndex.Length div SizeOf(TStaticPosition);
      If StCount > 1024 then StCount := 1024;
      Statics.Read(Buffer2, StCount * SizeOf(TStaticPosition));
      For I := 0 to StCount-1 do
        begin
          If (Buffer2[I].X = (X and $07)) and (Buffer2[I].Y = (Y and $07)) then
            begin
              GetMem(SP, SizeOf(TStaticPosition));
              SP^ := Buffer2[I];
              List.Add(SP);
            end;
        end;
   end;
end;

procedure TUOMap.DoPaint(Sender : TObject);
var
  Bitmap : TBitmap;
  W, H : Integer;
begin
  if not CheckFiles then Exit;
  with PB.Canvas do
    begin
      W := ClipRect.Right-ClipRect.Left;
      H := ClipRect.Bottom-ClipRect.Top;

      if (W <= 0) or (H <= 0) then Exit;

      Bitmap := TBitmap.Create;
      try
        OutputToBitmap(ClipRect, FFlags, FAltSetting, FZoom, Bitmap);
        Draw(ClipRect.Left,ClipRect.Top,Bitmap);
      finally
        Bitmap.Free;
      end;
    end;
end;

procedure TUOMap.DoMouseDown(Sender: TObject; Button: TMouseButton;
                             Shift: TShiftState; X, Y: Integer);
var
  P : TPoint;
begin
  P := PB.ClientToScreen(Point(X, Y));
  If Assigned(OnMapClick) then OnMapClick(Self, Button, Shift, P.X, P.Y, X, Y);
end;

function TUOMap.CheckFiles : Boolean;
begin
  Result :=
    (MapStr <> Nil) and (StaticsIndexStr <> Nil) and
    (Statics <> Nil) and ColorLookupFilled;
end;

procedure TUOMap.SetMapPath(Value : String);
begin
  MapStr.Free; MapStr := Nil;
  If FileExists(Value) then
   {$IfNDef UseMMF}
    MapStr := TFileStream.Create(Value,fmOpenRead or fmShareDenyNone);
   {$Else}
    MapStr := TMMFStream.Create(Value);
   {$EndIf}
  FMapPath := Value;
end;

procedure TUOMap.SetStaticsIndexPath(Value : String);
var
  F : TFileStream;
begin
  StaticsIndexStr.Free; StaticsIndexStr := Nil;
  if FileExists(Value) then
    begin
      StaticsIndexStr := TMemoryStream.Create;
      F := TFileStream.Create(Value,fmOpenRead or fmShareDenyNone);
      try
        StaticsIndexStr.LoadFromStream(F);
      finally
        F.Free;
      end;
    end;
  FStaticsIndexPath := Value;
end;

procedure TUOMap.SetStaticsPath(Value : String);
begin
  Statics.Free; Statics := Nil;
  If FileExists(Value) then
{$IfNDef UseMMF}
    Statics := TFileStream.Create(Value,fmOpenRead or fmShareDenyNone);
{$Else}
    Statics := TMMFStream.Create(Value);
{$EndIf}
  FStaticsPath := Value;
end;

procedure TUOMap.SetRadarPath(Value : String);
var
  FS : TFileStream;
begin
  ColorLookupFilled := False;
  If FileExists(Value) then
    begin
      FS := TFileStream.Create(Value,fmOpenRead or fmShareDenyNone);
      try
        FS.Read(ColorLookup, SizeOf(ColorLookup));
      finally
        FS.Free;
      end;
      ColorLookupFilled := True;
    end;
  FRadarPath := Value;
end;

procedure TUOMap.SetMapDifLocPath(Value : String);
begin
  MapDifLoc.Free; MapDifLoc := Nil;
  if FileExists(Value) then
    MapDifLoc := TFileStream.Create(Value, fmOpenRead or fmShareDenyNone);
  FMapDifLocPath := Value;
end;

procedure TUOMap.SetMapDifPath(Value : String);
begin
  MapDif.Free; MapDif := Nil;
  if FileExists(Value) then
    MapDif := TFileStream.Create(Value, fmOpenRead or fmShareDenyNone);
  FMapDifPath := Value;
end;

procedure TUOMap.SetStaDifLocPath(Value : String);
begin
  StaDifLoc.Free; StaDifLoc := Nil;
  if FileExists(Value) then
    StaDifLoc := TFileStream.Create(Value, fmOpenRead or fmShareDenyNone);
  FStaDifLocPath := Value;
end;

procedure TUOMap.SetStaDifPath(Value : String);
begin
  StaDif.Free; StaDif := Nil;
  if FileExists(Value) then
    StaDif := TFileStream.Create(Value, fmOpenRead or fmShareDenyNone);
  FStaDifPath := Value;
end;

procedure TUOMap.SetStaticsDifPath(Value : String);
begin
  StaticsDif.Free; StaticsDif := Nil;
  if FileExists(Value) then
    StaticsDif := TFileStream.Create(Value, fmOpenRead or fmShareDenyNone);
  FStaticsDifPath := Value;
end;

procedure TUOMap.SetZoom(Value : TMapZoom);
begin
  if Value <> FZoom then
    begin
      ScrollBox.HorzScrollBar.Position := 0;
      ScrollBox.VertScrollBar.Position := 0;
      FZoom := Value;
      PB.BoundsRect := TMapViewArea(FViews.Objects[FCurrentView]).GetBoundingRect(Value);
      PB.Invalidate;
    end;
end;

procedure TUOMap.SetFlags(Value : TMapFlags);
begin
  if Value <> FFlags then
    PB.Invalidate;
  FFlags := Value;
end;

procedure TUOMap.SetAltSetting(Value : TMapAlt);
begin
  if Value <> FAltSetting then
    PB.Invalidate;
  FAltSetting := Value;
end;

procedure Register;
begin
  RegisterComponents('Graphics', [TUOMap]);
end;

end.
