unit Hues;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  UOHues, ComCtrls, Grids, StdCtrls, ExtCtrls;

type
  TfrmHues = class(TForm)
    HuesGrid: TDrawGrid;
    Panel1: TPanel;
    btnOK: TButton;
    btnCancel: TButton;
    btnClear: TButton;
    procedure FormShow(Sender: TObject);
    procedure HuesGridDrawCell(Sender: TObject; Col, Row: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure HuesGridSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure btnClearClick(Sender: TObject);
  private
    OldRow : Integer;
    GroupCount : Integer;
    Range : Array of THueGroup;
    DataLoaded : Boolean;
    IncomingRow : Integer;
    SelectedRow : Integer;
  end;

procedure ShowHues(var Selection : Integer; var Entry : THueEntry);

implementation

uses Options;

{$R *.DFM}

procedure ShowHues(var Selection : Integer; var Entry : THueEntry);
var
  frmHues: TfrmHues;
begin
  Application.CreateForm(TfrmHues, frmHues);
  if Selection >= 0 then
    frmHues.IncomingRow := Selection;
  if frmHues.ShowModal = mrOK then
    begin
      Selection := frmHues.SelectedRow;
      Entry := frmHues.Range[Selection div 8].Entries[Selection mod 8];
    end;
  frmHues.Release;
end;

procedure TfrmHues.FormShow(Sender: TObject);
var
  F : TFileStream;
  I : Integer;
begin
  HuesGrid.ColWidths[0] := 100;

  if not frmOptions.VerifyFile('HUES.MUL') then
    Exit;

  F := TFileStream.Create(frmOptions.GetPath('HUES.MUL'), fmOpenRead);
  try
    GroupCount := F.Size div SizeOf(THueGroup);
    //GroupCount := 128; // Apparently this file is half full of garbage
    SetLength(Range, GroupCount);
    For I := 0 to GroupCount-1 do
      F.Read(Range[I], SizeOf(THueGroup));
  finally
    F.Free;
  end;
  DataLoaded := True;
  HuesGrid.RowCount := GroupCount * 8;
  Refresh;
  HuesGrid.Row := IncomingRow;
  SelectedRow := IncomingRow;
end;

function Color16To32(C : Word) : TColor;
begin
  Result :=
//    ((C and $1F) * 8) or
//    ((((C shr 5) and $1F) * 8) shl 8) or
//    ((((C shr 10) and $1F) * 8) shl 16);
    (((C and $1F) * 8) shl 16) or
    ((((C shr 5) and $1F) * 8) shl 8) or
    ((((C shr 10) and $1F) * 8));
end;

procedure TfrmHues.HuesGridDrawCell(Sender: TObject; Col, Row: Integer;
  Rect: TRect; State: TGridDrawState);
var
  S : String;
  Group : Integer;
  Entry : Integer;
begin
  if not DataLoaded then Exit;
  Brush.Style := bsSolid;
  Group := Row div 8;
  Entry := Row mod 8;
  if Col = 0 then with HuesGrid.Canvas do
    begin
      Pen.Style := psClear;
      if HuesGrid.Row = Row then
        begin
          Brush.Color := clHighlight;
          Font.Color := clHighlightText;
        end
      else
        begin
          Brush.Color := clWindow;
          Font.Color := clWindowText;
        end;
      S := Range[Group].Entries[Entry].Name;
      FillRect(Rect);
      DrawText(Handle, PChar(S), Length(S), Rect, DT_LEFT or DT_VCENTER or DT_SINGLELINE);
    end
  else
    begin
      with HuesGrid.Canvas, Range[Group] do
        begin
          Brush.Color := Color16To32(Entries[Entry].ColorTable[Col-1]);
          FillRect(Rect);
        end;
      if (HuesGrid.Row = Row) and (OldRow <> Row) then
        begin
          Rect := HuesGrid.CellRect(0, Row);
          InvalidateRect(HuesGrid.Handle, @Rect, True);
          Rect := HuesGrid.CellRect(0, OldRow);
          InvalidateRect(HuesGrid.Handle, @Rect, True);
          OldRow := Row;
        end;
    end;
end;

procedure TfrmHues.HuesGridSelectCell(Sender: TObject; ACol, ARow: Integer;
  var CanSelect: Boolean);
begin
  Caption := 'Hues - ' + Range[ARow div 8].Entries[ARow mod 8].Name;
  SelectedRow := ARow;
end;

procedure TfrmHues.btnClearClick(Sender: TObject);
begin
  SelectedRow := -1;
  ModalResult := mrOK;
end;

end.
