unit IndexFileList;

{
Most of the files in Ultima Online have an index file and are therefore
referenced by an index file and a main data file.  The TIndexFileList class
can be used to keep track of these files pairs.
}

interface

uses
  Classes, IndexedFile;

type
  TIndexFile = class
  public
    Main : String;
    Index : String;
    IndexedFile : TIndexedFile;
    constructor Create(
      AMain, AIndex, AMainPath, AIndexPath : String;
      APatchFile : TStream; ADataPatchID : Integer);
    function Equals(AMain, AIndex : String) : Boolean;
  end;

  TIndexFileList = class
  private
    Data : TList;
  public
    constructor Create;
    destructor Destroy; override;
    function Find(AMain, AIndex : String) : Integer;
    function Add(
      AMain, AIndex, AMainPath, AIndexPath : String;
      APatchFile : TStream; ADataPatchID : Integer) : Integer;
    function Get(Index : Integer) : TIndexedFile;      
  end;

implementation

uses
  SysUtils;

constructor TIndexFile.Create(
  AMain, AIndex, AMainPath, AIndexPath : String;
  APatchFile : TStream; ADataPatchID : Integer);
begin
  Main := LowerCase(AMain);
  Index := LowerCase(AIndex);
  if FileExists(AMainPath) and FileExists(AIndexPath) then
    begin
      IndexedFile := TIndexedFile.Create(
        AIndexPath, AMainPath,
        APatchFile, ADataPatchID, True);
    end
  else
    IndexedFile := Nil;
end;

function TIndexFile.Equals(AMain, AIndex : String) : Boolean;
begin
  Result := (AMain = Main) and (AIndex = Index);
end;

constructor TIndexFileList.Create;
begin
  Data := TList.Create;
end;

destructor TIndexFileList.Destroy;
var
  I : Integer;
begin
  for I := 0 to Data.Count-1 do
    TIndexFile(Data[I]).Free;
  Data.Free;
end;

function TIndexFileList.Find(AMain, AIndex : String) : Integer;
begin
  AMain := LowerCase(AMain);
  AIndex := LowerCase(AIndex);
  Result := 0;
  while (Result < Data.Count) and not TIndexFile(Data[Result]).Equals(AMain, AIndex) do
    begin
      Inc(Result);
    end;
  if Result >= Data.Count then
    Result := -1;
end;

function TIndexFileList.Add(
  AMain, AIndex, AMainPath, AIndexPath : String;
  APatchFile : TStream; ADataPatchID : Integer) : Integer;
begin
  Result := Data.Add(TIndexFile.Create(
    AMain, AIndex, AMainPath, AIndexPath, APatchFile, ADataPatchID));
end;

function TIndexFileList.Get(Index : Integer) : TIndexedFile;
begin
  Result := TIndexFile(Data[Index]).IndexedFile;
end;

end.
