unit Main;

interface

uses
  Windows, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ExtCtrls, ComCtrls, SelectionBar, StdCtrls, Grids,
  // UO helper libraries
  UOAnim,
  UOArt,
  UOFonts,
  UOGumps,
  UOMapComp,
  UOMulti,
  UOSkill,
  UOSound,
  UOTexMaps,
  UOHues,
  // Misc helper libraries
  InsideUOText, IndexFileList,
  ToolWin, IndexedFile, ImgList, PNG_Graphic, XMLParser,
  ActnList;

type
  TAnimSaveMode = (asmBitmaps, asmAVI);
  TActiveSection = (asNone, asAnimation, asArtwork, asFonts, asGumps,
                    asHues, asMap, asMulti, asSkills, asSound, asTexMaps);
  TArtGroup = (agLandscape, agStatic, agCharacter);

  TCharData = class
  public
    IndexedFile : TIndexedFile;
    Index : Integer;
    IndexHandle : Integer;
    Model : Integer;
    constructor Create(AIndexedFile : TIndexedFile; AIndex, AIndexHandle, AModel : Integer);
  end;

  TfrmMain = class(TForm)
    MainMenu: TMainMenu;
    pmAnim: TPopupMenu;
    pmArt: TPopupMenu;
    pmGumps: TPopupMenu;
    pmMap: TPopupMenu;
    pmMulti: TPopupMenu;
    pmSound: TPopupMenu;
    pmTexMaps: TPopupMenu;
    mnuAnimSaveAsBitmaps: TMenuItem;
    mnuAnimSaveAsAVI: TMenuItem;
    mnuAnimDetails: TMenuItem;
    mnuArtView: TMenuItem;
    mnuArtCopy: TMenuItem;
    mnuArtSaveAs: TMenuItem;
    mnuArtN1: TMenuItem;
    mnuArtDetails: TMenuItem;
    mnuArtFindbyName: TMenuItem;
    mnuArtFindbyNumber: TMenuItem;
    mnuArtFindNext: TMenuItem;
    mnuGumpsSaveAs: TMenuItem;
    mnuGumpsDetails: TMenuItem;
    mnuMapZoom: TMenuItem;
    mnuMapZoom1x1: TMenuItem;
    mnuMapZoomWhole: TMenuItem;
    mnuMapN1: TMenuItem;
    mnuMapNormal: TMenuItem;
    mnuMapAltitude: TMenuItem;
    mnuMapN2: TMenuItem;
    mnuMapBlackWater: TMenuItem;
    mnuMapNormalWithAlt: TMenuItem;
    mnuMapDrawStatics: TMenuItem;
    mnuMapXRayView: TMenuItem;
    mnuMapN3: TMenuItem;
    mnuMapSave: TMenuItem;
    mnuMapN4: TMenuItem;
    mnuMapDetails: TMenuItem;
    mnuMultiSaveAs: TMenuItem;
    mnuMultiDetails: TMenuItem;
    mnuSoundPlay: TMenuItem;
    mnuSoundCopy: TMenuItem;
    mnuSoundSaveAs: TMenuItem;
    mnuSoundDetails: TMenuItem;
    mnuTexMapsView: TMenuItem;
    mnuTexMapsSaveAs: TMenuItem;
    mnuTexMapsCopy: TMenuItem;
    mnuTexMapsDetails: TMenuItem;
    ilSelection: TImageList;
    sbMain: TStatusBar;
    Help1: TMenuItem;
    About1: TMenuItem;
    View1: TMenuItem;
    Options1: TMenuItem;
    ilSound: TImageList;
    sdSound: TSaveDialog;
    SavePicture: TSaveDialog;
    AnimTimer: TTimer;
    sdAnim: TSaveDialog;
    ColorDialog: TColorDialog;
    sdAnimAVI: TSaveDialog;
    sbSelection: TSelectionBar;
    N1: TMenuItem;
    pnlMain: TPanel;
    pnlIntro: TPanel;
    imgInsideUO: TImage;
    pnlAnimation: TPanel;
    Label2: TLabel;
    Splitter1: TSplitter;
    tvAnim: TTreeView;
    pnlAnimDisplay: TPanel;
    imgAnimation: TImage;
    pnlAnimTop: TPanel;
    lblAnimFrameSet: TLabel;
    edtAnimFrameSet: TEdit;
    tbAnimSpeed: TTrackBar;
    ToolBar1: TToolBar;
    btnAnimate: TToolButton;
    btnSetBackColor: TToolButton;
    btnSetCharColor: TToolButton;
    cbFullReplace: TCheckBox;
    pnlArtwork: TPanel;
    pcArtwork: TPageControl;
    tsLandscape: TTabSheet;
    dgLandscape: TDrawGrid;
    tsStatic: TTabSheet;
    dgStatic: TDrawGrid;
    tsCharacter: TTabSheet;
    dgCharacter: TDrawGrid;
    pnlFonts: TPanel;
    Splitter3: TSplitter;
    lvFonts: TListView;
    pnlMap: TPanel;
    pnlMulti: TPanel;
    Splitter2: TSplitter;
    lvMulti: TListView;
    sbMulti: TScrollBox;
    imgMulti: TImage;
    pnlMultiTop: TPanel;
    Label1: TLabel;
    tbMulti: TTrackBar;
    pnlSound: TPanel;
    tvSound: TTreeView;
    pnlTexMaps: TPanel;
    dgTexMaps: TDrawGrid;
    pnlGumps: TPanel;
    pnlInGumps: TPanel;
    lvGumps: TListView;
    Splitter4: TSplitter;
    pnlSkills: TPanel;
    lvSkills: TListView;
    ilSkills: TImageList;
    pnlFontMain: TPanel;
    dgFonts: TDrawGrid;
    pnlFontHeader: TPanel;
    ToolBar2: TToolBar;
    tbMakePhrase: TToolButton;
    edtFontPhrase: TEdit;
    lblWelcome: TLabel;
    pnlSectionTitleBack: TPanel;
    pnlSectionTitle: TPanel;
    lblMapError: TLabel;
    pmMapsEmpty: TPopupMenu;
    imgGumps: TImage;
    ActionList: TActionList;
    ilActions: TImageList;
    actDrawPhrase: TAction;
    actAnimate: TAction;
    actSetBackColor: TAction;
    actSetCharColor: TAction;
    actPlaySound: TAction;
    actFilePaths: TAction;
    actAbout: TAction;
    tbFontColor: TToolButton;
    actFontColor: TAction;
    pnlHues: TPanel;
    HuesGrid: TDrawGrid;
    N2: TMenuItem;
    mnuArtSaveAll: TMenuItem;
    sdArtAll: TSaveDialog;
    mnuMapDrawPatched: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure mnuAnimSaveAsBitmapsClick(Sender: TObject);
    procedure mnuAnimSaveAsAVIClick(Sender: TObject);
    procedure mnuAnimDetailsClick(Sender: TObject);
    procedure mnuArtViewClick(Sender: TObject);
    procedure mnuArtCopyClick(Sender: TObject);
    procedure mnuArtSaveAsClick(Sender: TObject);
    procedure mnuArtFindbyNameClick(Sender: TObject);
    procedure mnuArtFindbyNumberClick(Sender: TObject);
    procedure mnuArtFindNextClick(Sender: TObject);    
    procedure mnuArtDetailsClick(Sender: TObject);
    procedure mnuGumpsSaveAsClick(Sender: TObject);
    procedure mnuGumpsDetailsClick(Sender: TObject);
    procedure mnuMapNormalClick(Sender: TObject);
    procedure mnuMapAltitudeClick(Sender: TObject);
    procedure mnuMapDrawStaticsClick(Sender: TObject);
    procedure mnuMapXRayViewClick(Sender: TObject);
    procedure mnuMapBlackWaterClick(Sender: TObject);
    procedure mnuMapNormalWithAltClick(Sender: TObject);
    procedure mnuMapSaveClick(Sender: TObject);
    procedure mnuMapDetailsClick(Sender: TObject);
    procedure mnuMapViewAlt(Sender: TObject);
    procedure mnuMultiSaveAsClick(Sender: TObject);
    procedure mnuMultiDetailsClick(Sender: TObject);
    procedure mnuSoundPlayClick(Sender: TObject);
    procedure mnuSoundCopyClick(Sender: TObject);
    procedure mnuSoundSaveAsClick(Sender: TObject);
    procedure mnuSoundDetailsClick(Sender: TObject);
    procedure mnuTexMapsViewClick(Sender: TObject);
    procedure mnuTexMapsCopyClick(Sender: TObject);
    procedure mnuTexMapsSaveAsClick(Sender: TObject);
    procedure mnuTexMapsDetailsClick(Sender: TObject);
    procedure tvSoundDblClick(Sender: TObject);
    procedure pnlArtworkResize(Sender: TObject);
    procedure dgLandscapeDrawCell(Sender: TObject; Col, Row: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure dgStaticDrawCell(Sender: TObject; Col, Row: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure dgLandscapeDblClick(Sender: TObject);
    procedure dgLandscapeMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure dgStaticDblClick(Sender: TObject);
    procedure dgStaticMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ZoomClick(Sender: TObject);
    procedure dgCharacterDrawCell(Sender: TObject; Col, Row: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure dgCharacterDblClick(Sender: TObject);
    procedure dgCharacterMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure tvSoundDeletion(Sender: TObject; Node: TTreeNode);
    procedure tvAnimChange(Sender: TObject; Node: TTreeNode);
    procedure AnimTimerTimer(Sender: TObject);
    procedure cbAnimRectClick(Sender: TObject);
    procedure tvAnimExpanding(Sender: TObject; Node: TTreeNode;
      var AllowExpansion: Boolean);
    procedure edtAnimFrameSetKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure edtAnimFrameSetKeyPress(Sender: TObject; var Key: Char);
    procedure tbAnimSpeedChange(Sender: TObject);
    procedure imgAnimationMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure imgAnimationMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pnlTexMapsResize(Sender: TObject);
    procedure dgTexMapsDblClick(Sender: TObject);
    procedure dgTexMapsDrawCell(Sender: TObject; Col, Row: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure lvGumpsDeletion(Sender: TObject; Item: TListItem);
    procedure lvGumpsChange(Sender: TObject; Item: TListItem;
      Change: TItemChange);
    procedure lvMultiChange(Sender: TObject; Item: TListItem;
      Change: TItemChange);
    procedure tbMultiChange(Sender: TObject);
    procedure MultiMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pnlFontsResize(Sender: TObject);
    procedure dgFontsDrawCell(Sender: TObject; Col, Row: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure FormDestroy(Sender: TObject);
    procedure lvFontsDeletion(Sender: TObject; Item: TListItem);
    procedure lvFontsChange(Sender: TObject; Item: TListItem;
      Change: TItemChange);
    procedure dgTexMapsMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure dgStaticKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure dgLandscapeKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure cbFullReplaceClick(Sender: TObject);
    procedure tvAnimContextPopup(Sender: TObject; MousePos: TPoint;
      var Handled: Boolean);
    procedure tvSoundContextPopup(Sender: TObject; MousePos: TPoint;
      var Handled: Boolean);
    procedure lvGumpsContextPopup(Sender: TObject; MousePos: TPoint;
      var Handled: Boolean);
    procedure edtFontPhraseKeyPress(Sender: TObject; var Key: Char);
    procedure pnlMapContextPopup(Sender: TObject; MousePos: TPoint;
      var Handled: Boolean);
    procedure actDrawPhraseExecute(Sender: TObject);
    procedure actAnimateExecute(Sender: TObject);
    procedure actSetBackColorExecute(Sender: TObject);
    procedure actSetCharColorExecute(Sender: TObject);
    procedure actFilePathsExecute(Sender: TObject);
    procedure actAboutExecute(Sender: TObject);
    procedure pnlInGumpsResize(Sender: TObject);
    procedure actFontColorExecute(Sender: TObject);
    procedure sbMultiResize(Sender: TObject);
    procedure dgFontsDblClick(Sender: TObject);
    procedure HuesGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure HuesGridSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure pnlHuesResize(Sender: TObject);
    procedure mnuArtSaveAllClick(Sender: TObject);
    procedure mnuMapDrawPatchedClick(Sender: TObject);
  private
    { Private declarations }
    ActiveSection : TActiveSection;
    ActiveFilesOK : Boolean;

    FPatchCount : Integer;
    FPatchFile : TFileStream;
    FIndex,
    FMain : TFileStream;

    FIdxMain : TIndexedFile;

    GlobalBitmap : TBitmap;

    FindBlockName : String;
    FindBlockNum : Integer;
    FindActive : (faNo, faName, faNumber);
    FindTag : Integer;

     // Animation
    AnimFiles : TIndexFileList;
    FrameSet : TFrameSet;
    DisablingAnimation : Boolean;
    Processing : Boolean;
    CurrFrame : Integer;
    CurrNode : Integer;
    FActiveAnimationNode : TTreeNode;
    TotalHueReplace : Boolean;
    AnimHue : Integer;
    AnimHueEntry : THueEntry;

     // Artwork
    FLandscapeStart,
    FLandscapeCount,
    FStaticStart,
    FStaticCount,
    FCharacterStart,
    FCharacterCount : Integer;
    FActiveArtGrid : TArtGroup;
    FActiveBNum : Integer;
    CenterX, CenterY : Integer;

     // Fonts
    ActiveItem : TListItem;
    FontHue : Integer;
    FontHueEntry : THueEntry;

     // Gumps
    FGumpItem : TListItem;
    FGumpCount : Integer;

     // Hues
    FOldRow : Integer;
    FRange : Array of THueGroup;
    FStartingRow : Integer;

     // Map
    FUOMap : TUOMap;
    WhichMap : String;
    WhichMapIdx : Integer;
    MapAX, MapAY : Integer;
    Maps : TStringList;
    MapsLoadedOnce : Boolean;

     // Multi
    MultiFiles : TIndexFileList;
    ArtIndex : TFileStream;
    ArtMain : TFileStream;
    ArtCacheIndex : TArtCacheIndex;
    ShowFloors : Integer;

     // Sound
    FActiveSoundNode : TTreeNode;

    procedure EstablishBitmap(W, H : Integer; C : TColor);

     // Animation
    procedure AddCharTreeNode(
      IndexedFile : TIndexedFile;
      ParentTreeNode : TTreeNode;
      Name : String;
      ModelSize, ModelStart,
      StartIndexHandle, IndexHandle : Integer); overload;
    procedure AddCharTreeNode(
      IndexedFile : TIndexedFile;
      ParentTreeNode : TTreeNode;
      Name : String;
      Model, StartIndex, IndexHandle : Integer); overload;
    procedure LoadCurrFrame;
    procedure LoadCurrNode;
    procedure ShowCharacter(Name : String; CharData : TCharData);
    procedure ShowAction(Name : String; CharData : TCharData);
    procedure ShowFrameSet(
      Name : String; IndexedFile : TIndexedFile; Index : Integer);
    procedure SaveCharacter(Name : String; CharData : TCharData;
                            SaveMode : TAnimSaveMode);
    procedure SaveAction(Name : String; CharData : TCharData;
                         SaveMode : TAnimSaveMode);
    procedure SaveFrameSet(Name : String; CharData : TCharData; SetIdx : Integer;
                           SaveMode : TAnimSaveMode);
    procedure DoSaveFrameSet(
      IndexedFile : TIndexedFile; Name : String; Index : Integer;
      SaveMode : TAnimSaveMode);
    procedure UseFrameSet(
      IndexedFile : TIndexedFile;
      Index : Integer);

     // Art
    procedure DrawArtCell(DrawGrid : TDrawGrid; Rect : TRect;
                          Handle : Integer; Style : TBlockStyle; W, H : Word;
                          Selected : Boolean);
    function GetArtIndexData : TIndexedData;

     // Fonts
    procedure DrawFontCell(C : TCanvas; R : TRect; Character : TCharacter);

     // Gumps

     // Map
    procedure DoMapClick(Sender: TObject; Button: TMouseButton;
                         Shift: TShiftState; X, Y, AX, AY : Integer);
    procedure BuildMapsMenu(AMenu : TMenu);

     // Multi
    procedure DrawMulti(Item : TListItem);

     // Sound
    procedure DoSelectionClick(Sender : TObject);
    procedure PlayRequestedSound(Item : TTreeNode);

     // TexMaps
    function BuildTexBitmap(Index : Integer) : TBitmap;
  protected
     // Deactivate
    procedure DeactivateOld;
    procedure DeactivateAnimation;
    procedure DeactivateArtwork;
    procedure DeactivateFonts;
    procedure DeactivateGumps;
    procedure DeactivateHues;
    procedure DeactivateMap;
    procedure DeactivateMulti;
    procedure DeactivateSkills;
    procedure DeactivateSound;
    procedure DeactivateTexMaps;
    function GetDescription(BlockNum : Integer; Default : String) : String;
  public
     // Activate
    procedure ActivateAnimation;
    procedure ActivateArtwork;
    procedure ActivateFonts;
    procedure ActivateGumps;
    procedure ActivateHues;
    procedure ActivateMap;
    procedure ActivateMulti;
    procedure ActivateSkills;
    procedure ActivateSound;
    procedure ActivateTexMaps;
     //
    function FindByName(Start : Integer; BlkName : String) : Integer;
    function FindByNumber(Start, BlkNum : Integer) : Integer;
  end;

var
  frmMain: TfrmMain;

implementation

uses About, Options, MMSystem, ArtView, FindBlockNum, FindBlockName,
  MapPointData, HashTable, Clipbrd, Utility, MapSave, UOMap, AVI,
  Hues, Details, Color15bit;

{$R *.DFM}

const
  SectionStart = 0;
  SectionCount = 10;
  Sections : Array[0..SectionCount-1] of String =
    ('Animation',
     'Artwork',
     'Fonts',
     'Gumps',
     'Hues',
     'Map',
     'Multi',
     'Skills',
     'Sound',
     'Texture Maps');
  SectionIconNames : Array[0..SectionCount-1] of String =
    ('MainMenu_Animation',
     'MainMenu_Art',
     'MainMenu_Fonts',
     'MainMenu_Gumps',
     'MainMenu_Hues',
     'MainMenu_Map',
     'MainMenu_Multi',
     'MainMenu_Skills',
     'MainMenu_Sound',
     'MainMenu_TexMaps');
  GridColors : Array[Boolean] of TColor = (clWindow, clHighlight);
  ArtGroupStrings : Array[TArtGroup] of String =
    ('Landscape', 'Static', 'CharacterTest');

procedure BoldIt(Ed : TRichEdit; St, Ln : Integer);
var
  Old : Integer;
begin
  Old := Ed.SelStart;
  Ed.SelStart := St; Ed.SelLength := Ln;
  Ed.SelAttributes.Style := [fsBold];
  Ed.SelStart := Old;
end;

function LoadBitmap(IL : TImageList; Name : String) : Boolean;
var
  BitMap, Mask : TBitMap;
begin
  Result := True;
  Bitmap := TBitmap.Create;
  Mask := TBitmap.Create;
  try
    // If this line fails, it is because the new area doesn't have a bitmap
    // in the provided image list.
    BitMap.LoadFromResourceName(HInstance, Name);
    Mask.LoadFromResourceName(HInstance, Name);
    Mask.Mask(clBlack);
    IL.Add(Bitmap, Mask);
  finally
    Mask.Free;
    Bitmap.Free;
  end;
end;

type
 {$A-}
  TPatchIndex = record
    FileID : Integer;
    Handle : Integer;
    Lookup : Integer;
    Length : Integer;
    Unknown : Integer;
  end;
 {$A+}

 //-----------------------------------------------------------------------------

procedure TfrmMain.FormCreate(Sender: TObject);
begin
   // Override in order to do animated minimize and maximize
//  Application.OnRestore := AppRestore;
  GlobalBitmap := TBitmap.Create;
  MapsLoadedOnce := False;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  Maps.Free;
  Maps := Nil;
  GlobalBitmap.Free;
end;

procedure TfrmMain.FormShow(Sender: TObject);
var
  SC : TSelectionControl;
  I : Integer;
begin
  try
    frmOptions.Startup;
  except
    on E: Exception do
      ShowMessage(E.Message);
  end;

  For I := 0 to SectionCount-1 do
    LoadBitmap(ilSelection, SectionIconNames[I]);
  AnimHue := -1;

//  ilAnimationOptions.ResourceLoad(rtBitmap, 'Animation_Animation', clOlive);
//  ilAnimationOptions.ResourceLoad(rtBitmap, 'Animation_Color', clOlive);

//  ilSound.ResourceLoad(rtBitmap, 'Sound_Folder', clOlive);
//  ilSound.ResourceLoad(rtBitmap, 'Sound_Speaker', clOlive);

   // Set full alignment for all panels
  pnlIntro.Align := alClient;
  pnlAnimation.Align := alClient;
  pnlArtwork.Align := alClient;
  pnlFonts.Align := alClient;
  pnlGumps.Align := alClient;
  pnlHues.Align := alClient;
  pnlMap.Align := alClient;
  pnlMulti.Align := alClient;
  pnlSkills.Align := alClient;
  pnlSound.Align := alClient;
  pnlTexMaps.Align := alClient;

  pnlIntro.BringToFront;

  // Build the selection control on the left--one button at a time
  for I := SectionStart to SectionCount-1 do
    begin
      SC := sbSelection.AddSelectionControl;
      with SC do
        begin
          Caption := Sections[I];
          ImageIndex := I;
          OnClick := DoSelectionClick;
        end;
    end;

//  DatFileName := ExtractFilePath(ParamStr(0)) + DatFileName;

   // Try to load the patch file
  if FileExists(frmOptions.GetPath('VERDATA.MUL')) then
    begin
      FPatchFile := TFileStream.Create(frmOptions.GetPath('VERDATA.MUL'),
                                       fmOpenRead or fmShareDenyNone);
      FPatchFile.Read(FPatchCount, SizeOf(FPatchCount));
    end
  else
    FPatchFile := Nil;
end;  // FormShow

procedure TfrmMain.FormHide(Sender: TObject);
begin
  FPatchFile.Free;
  DeactivateOld;
end;  // FormHide

procedure TfrmMain.DoSelectionClick(Sender : TObject);
var
  Section : Integer;
begin
  Section := (Sender as TSelectionControl).ImageIndex;
  if (Section < 0) or
     (Section >= SectionCount) or
     (Section = Ord(ActiveSection)-1) then
    Exit;
  pnlSectionTitle.Caption := Sections[Section];
  DeactivateOld;
  Case Section of
    0 : ActivateAnimation;
    1 : ActivateArtwork;
    2 : ActivateFonts;
    3 : ActivateGumps;
    4 : ActivateHues;
    5 : ActivateMap;
    6 : ActivateMulti;
    7 : ActivateSkills;
    8 : ActivateSound;
    9 : ActivateTexMaps;
  end;
end;

procedure TfrmMain.actAboutExecute(Sender: TObject);
begin
  frmAbout.ShowModal;
end;

procedure TfrmMain.actFilePathsExecute(Sender: TObject);
begin
  frmOptions.ShowModal;
end;

procedure TfrmMain.DeactivateOld;
begin
  try
    case ActiveSection of
      asAnimation : DeactivateAnimation;
      asArtwork : DeactivateArtwork;
      asFonts : DeactivateFonts;
      asGumps : DeactivateGumps;
      asHues : DeactivateHues;
      asMap : DeactivateMap;
      asMulti : DeactivateMulti;
      asSkills : DeactivateSkills;
      asSound : DeactivateSound;
      asTexMaps : DeactivateTexMaps;
    end;
  finally
    FIndex := Nil;
    FMain := Nil;
  end;
end;

procedure TfrmMain.EstablishBitmap(W, H : Integer; C : TColor);
begin
  GlobalBitmap.FreeImage;
  GlobalBitMap.Height := H;
  GlobalBitMap.Width := W;
  GlobalBitMap.PixelFormat := pf15bit;
  GlobalBitMap.HandleType := bmDIB;
  GlobalBitmap.Canvas.Brush.Style := bsSolid;
  GlobalBitmap.Canvas.Brush.Color := C;
  GlobalBitmap.Canvas.FillRect(Rect(0,0,W,H));
end;

function CreateBitmap(W, H : Integer; C : TColor) : TBitmap;
begin
  Result := TBitmap.Create;
  Result.Width := W;
  Result.Height := H;
  Result.PixelFormat := pf15bit;
  Result.HandleType := bmDIB;
  Result.Canvas.Brush.Style := bsSolid;
  Result.Canvas.Brush.Color := C;
  Result.Canvas.FillRect(Rect(0,0,W,H));
end;

 //-----------------------------------------------------------------------------
 //  ANIMATION
 //-----------------------------------------------------------------------------

// Terms
// -----
// Index : This is the entry offset in an index file with blank entries removed.
// Index Handle : This is the entry offset in an index file with no entries
//                removed.
// Character : This is all frame sets associated with a specific
//             creature, thing or person animation.  For example, the Titan
//             character is all of the frames associated with a Titan.
// Sequence : This is all frame sets assiocated with a particular action
//            in any direction.  For example, "Attack" in all possible
//            directions is a sequence.
//

constructor TCharData.Create(
  AIndexedFile : TIndexedFile;
  AIndex, AIndexHandle, AModel : Integer);
begin
  IndexedFile := AIndexedFile;
  Index := AIndex;
  IndexHandle := AIndexHandle;
  Model := AModel;
end;

// Adds a specified IndexHandle's entry to the animation tree if the entry
// exists in the index file.
//
// TODO:
// This isn't the most efficient way to do it since it has to do a search
// for each handle.  The best way would be to require that all nodes are in
// order and therefore the search would be much quicker.
procedure TfrmMain.AddCharTreeNode(
  IndexedFile : TIndexedFile;
  ParentTreeNode : TTreeNode;
  Name : String;
  ModelSize, ModelStart,
  StartIndexHandle, IndexHandle : Integer);
var
  TreeNode : TTreeNode;
  Index : Integer;
begin
  if IndexedFile = Nil then Exit;
  
  if IndexedFile.FindHandle(IndexHandle, Index) then
    begin
      TreeNode := tvAnim.Items.AddChildObject(ParentTreeNode, Name,
        TCharData.Create(
          IndexedFile, Index, IndexHandle,
          ModelStart + (IndexHandle - StartIndexHandle) div ModelSize));
      if IndexedFile[Index].Source = FPatchFile then
        TreeNode.Text := TreeNode.Text + '(p)';
      TreeNode.HasChildren := true;
    end;
end;

// Basically the same as the above except this one is specifically for
// actions (also called frame sets).
procedure TfrmMain.AddCharTreeNode(
  IndexedFile : TIndexedFile;
  ParentTreeNode : TTreeNode;
  Name : String;
  Model, StartIndex, IndexHandle : Integer);
var
  TreeNode : TTreeNode;
  Index : Integer;
begin
  if IndexedFile.FindHandle(IndexHandle, StartIndex, Index) then
    begin
      TreeNode := tvAnim.Items.AddChildObject(ParentTreeNode, Name,
        TCharData.Create(IndexedFile, Index, IndexHandle, Model));
      if IndexedFile[Index].Source = FPatchFile then
        TreeNode.Text := TreeNode.Text + '(p)';
      TreeNode.HasChildren := true;
    end;
end;

procedure TfrmMain.ActivateAnimation;
var
  I, J : Integer;
  Panel : TStatusPanel;
  CharTypeTreeNode : TTreeNode;
  AnimationNode : TXMLNode;
  CharTypeNode : TXMLNode;
  CharactersNode : TXMLNode;
  CharacterNode : TXMLNode;
  SequencesNode : TXMLNode;
  StartIndexHandle : Integer;
  Size : Integer;
  PrevIndexHandle : Integer;
  ModelStart : Integer;
  CharsStartIndexHandle : Integer; // Starting Index Handle of character type
  FileIndex : Integer;
  AnimFile : String;
  AnimIdxFile : String;
begin
  pnlAnimation.BringToFront;
  ActiveSection := asAnimation;
  ActiveFilesOK := False;
  DisablingAnimation := False;

   //
  sbMain.SimpleText := 'Loading animation information...';
  CurrFrame := 0;
  CurrNode := 0;
  FIndex := Nil; FMain := Nil;

  Screen.Cursor := crHourglass;
  Application.ProcessMessages;
  try
    Assert(AnimFiles = Nil);
    AnimFiles := TIndexFileList.Create;

    // Load up the tree view
    tvAnim.Items.BeginUpdate;
    try
      // Load animation information from the XML file.
      AnimationNode := frmOptions.InsideUOData.GetNode('InsideUO/Animations');
      try
        for I := 0 to AnimationNode.Count-1 do
          begin
            CharTypeNode := AnimationNode[I];
            if CharTypeNode.TagName = 'CHARACTERTYPE' then
              begin
                CharTypeTreeNode := tvAnim.Items.AddChildObject(
                  Nil, CharTypeNode.GetAttributeValue('NAME'), CharTypeNode);

                AnimFile := CharTypeNode.GetAttributeValue('ANIM_FILE');
                AnimIdxFile := CharTypeNode.GetAttributeValue('ANIMIDX_FILE');

                FileIndex := AnimFiles.Find(AnimFile, AnimIdxFile);
                if FileIndex < 0 then
                  begin
                    FileIndex := AnimFiles.Add(
                      AnimFile, AnimIdxFile,
                      frmOptions.GetPath(AnimFile),
                      frmOptions.GetPath(AnimIdxFile),
                      FPatchFile, $06);
                  end;

                // This value is apparently an offset for UOX?
                ModelStart :=
                  StrToInt(CharTypeNode.GetAttributeValue('STARTMODEL'));

                SequencesNode := CharTypeNode.GetChildNode('SEQUENCES');
                Size := StrToInt(SequencesNode.GetAttributeValue('SIZE'));

                CharactersNode := CharTypeNode.GetChildNode('CHARACTERS');
                CharsStartIndexHandle := StrToInt(CharactersNode.GetAttributeValue('START'));
                PrevIndexHandle := CharsStartIndexHandle - Size;
                for J := 0 to CharactersNode.Count-1 do
                  begin
                    CharacterNode := CharactersNode[J];
                    if CharacterNode.TagName = 'CHARACTER' then
                      begin
                        StartIndexHandle := StrToInt(CharacterNode.GetAttributeValue('START'));

                        // Add in any nodes that aren't represented
                        Inc(PrevIndexHandle, Size);
                        while(StartIndexHandle > PrevIndexHandle) do
                          begin
                            AddCharTreeNode(
                              AnimFiles.Get(FileIndex), CharTypeTreeNode,
                              IntToStr(PrevIndexHandle), Size, ModelStart,
                              CharsStartIndexHandle, PrevIndexHandle);
                            Inc(PrevIndexHandle, Size);
                          end;

                        // Add the specified node
                        if CharacterNode.Data = '' then
                          AddCharTreeNode(
                            AnimFiles.Get(FileIndex), CharTypeTreeNode,
                            IntToStr(PrevIndexHandle), Size, ModelStart,
                            CharsStartIndexHandle, StartIndexHandle)
                        else
                          AddCharTreeNode(
                            AnimFiles.Get(FileIndex), CharTypeTreeNode,
                            CharacterNode.Data, Size, ModelStart,
                            CharsStartIndexHandle, StartIndexHandle);
                      end;
                  end;
                if CharTypeTreeNode.Count = 0 then
                  tvAnim.Items.Delete(CharTypeTreeNode)
                else
                  CharTypeTreeNode.AlphaSort;
              end
          end;
      except
        on E: Exception do
          raise Exception.CreateFmt(ErrorXMLWrapper, [e.Message]);
      end;
    finally
      tvAnim.Items.EndUpdate;
    end;
    if tvAnim.Items.Count > 0 then
      begin
        tvAnim.Selected := tvAnim.Items[0];
      end;
  finally
    Screen.Cursor := crDefault;
  end;

   // Set up the status bar
  sbMain.SimplePanel := False;
  sbMain.Panels.Clear;

  Panel := sbMain.Panels.Add;
  Panel.Width := 60;
  Panel.Bevel := pbNone;
  Panel.Text := 'Character';
  Panel.Alignment := taRightJustify;

  Panel := sbMain.Panels.Add;
  Panel.Width := 150;
  Panel.Text := '';

  Panel := sbMain.Panels.Add;
  Panel.Width := 60;
  Panel.Bevel := pbNone;
  Panel.Text := 'Model No.';

  Panel := sbMain.Panels.Add;
  Panel.Width := 80;
  Panel.Text := '0';

  Panel := sbMain.Panels.Add;
  Panel.Width := 10;
  Panel.Bevel := pbNone;
  Panel.Text := '';

  ActiveFilesOK := True;
end;  // TfrmMain.ActivateAnimation

procedure TfrmMain.DeactivateAnimation;
begin
  DisablingAnimation := True;

  AnimFiles.Free;
  AnimFiles := Nil;

  tvAnim.Items.BeginUpdate;
  try
    tvAnim.Items.Clear;
  finally
    tvAnim.Items.EndUpdate;
  end;

  FrameSet.Free;
  FrameSet := Nil;

  AnimTimer.Enabled := False;

//  FIdxMain.Free;
  sbMain.SimplePanel := True;
  DisablingAnimation := False;
end;  // TfrmMain.DeactivateAnimation

procedure TfrmMain.tvAnimExpanding(Sender: TObject; Node: TTreeNode;
  var AllowExpansion: Boolean);
var
  I : Integer;
  SeqNode : TXMLNode;
  SequencesNode : TXMLNode;
  CharacterTypeNode : TXMLNode;
  StartIndexHandle : Integer;
  CharData : TCharData;
//  ModelStart : Integer;
//  Size : Integer;
begin
  AllowExpansion := True;
  if Node.Count > 0 then Exit;

   // Load children
  case Node.Level of
    1 : // Sequence level
      if Node.Count <= 0 then
        begin
          CharacterTypeNode := TXMLNode(Node.Parent.Data);
          // ModelStart := StrToInt(CharacterTypeNode.GetAttributeValue('STARTMODEL'));
          CharData := TCharData(Node.Data);

          // Reloading from the XML parser for ever node may be a
          // little inefficient.
          SequencesNode := CharacterTypeNode.GetChildNode('SEQUENCES');
          // Size := StrToInt(SequencesNode.GetAttributeValue('SIZE'));
          For I := 0 to SequencesNode.Count-1 do
            begin
              SeqNode := SequencesNode[I];
              If SeqNode.TagName = 'SEQ' then
                begin
                  StartIndexHandle := CharData.IndexHandle + StrToInt(SeqNode.GetAttributeValue('START'));
                  AddCharTreeNode(
                    CharData.IndexedFile,
                    Node, SeqNode.Data, CharData.Model,
                    CharData.Index, StartIndexHandle);
                end;
            end;
        end;
    2 : // Frame set level
      begin
        tvAnim.Items.AddChild(Node, 'Down');
        tvAnim.Items.AddChild(Node, 'Down-Left');
        tvAnim.Items.AddChild(Node, 'Left');
        tvAnim.Items.AddChild(Node, 'Up-Left');
        tvAnim.Items.AddChild(Node, 'Up');
      end;
  end;
end;

procedure TfrmMain.LoadCurrNode;
var
  Node : TTreeNode;
begin
  actAnimate.Enabled := False;

  if CurrNode >= tvAnim.Items.Count then Exit;
  Node := tvAnim.Items[CurrNode];
  if Node = Nil then Exit;

  case Node.Level of
    1 : ShowCharacter(Node.Text, TCharData(Node.Data));
    2 : ShowAction(Node.Text, TCharData(Node.Data));
    3 : ShowFrameSet(Node.Text, TCharData(Node.Parent.Data).IndexedFile, TCharData(Node.Parent.Data).Index + Node.Index);
  else
    Exit;
  end;

  actAnimate.Enabled := True;
end;

procedure TfrmMain.LoadCurrFrame;
begin
  if FrameSet = Nil then Exit;
  if AnimHue > 0 then
    begin
      FrameSet.UseHue := True;
      FrameSet.HueEntry := AnimHueEntry;
      FrameSet.TotalHueReplace := TotalHueReplace;
    end
  else
    FrameSet.UseHue := False;
  FrameSet.LoadFrameToImage(CurrFrame, imgAnimation, frmOptions.AnimBkColor,
                            CenterX, CenterY);
end;

// Update the current frameset and statusbar to reflect a specific selection
// of a character indicated by the CharData parameter.
procedure TfrmMain.ShowCharacter(Name : String; CharData : TCharData);
begin
  sbMain.Panels[0].Text := 'Character';
  sbMain.Panels[1].Text := Name;
  sbMain.Panels[3].Text := Format('%d (0x%x)', [CharData.Model, CharData.Model]);
  sbMain.Panels[2].Width := 60;
  sbMain.Panels[3].Width := 80;
  UseFrameSet(CharData.IndexedFile, CharData.Index+1);
  CurrFrame := 0;
  LoadCurrFrame;
end;

// Update the current frameset and statusbar to reflect a specific selection
// of a character action indicated by the CharData parameter.
procedure TfrmMain.ShowAction(Name : String; CharData : TCharData);
begin
  sbMain.Panels[0].Text := 'Action';
  sbMain.Panels[1].Text := Name;
  sbMain.Panels[2].Width := 0;
  sbMain.Panels[3].Width := 0;
  UseFrameSet(CharData.IndexedFile, CharData.Index + 1);
  CurrFrame := 0;
  LoadCurrFrame;
end;

// Update the current frameset and statusbar to reflect a specific selection
// of a frame set (ie. the lowest level--direction)
procedure TfrmMain.ShowFrameSet(Name : String; IndexedFile : TIndexedFile; Index : Integer);
begin
  sbMain.Panels[0].Text := 'Frame Set';
  sbMain.Panels[1].Text := Name;
  sbMain.Panels[2].Width := 0;
  sbMain.Panels[3].Width := 0;
  UseFrameSet(IndexedFile, Index);
  CurrFrame := 0;
  LoadCurrFrame;
end;

procedure TfrmMain.SaveCharacter(Name : String; CharData : TCharData;
                                 SaveMode : TAnimSaveMode);
begin
  DoSaveFrameSet(CharData.IndexedFile, Name, CharData.Index+1, SaveMode);
end;

procedure TfrmMain.SaveAction(Name : String; CharData : TCharData;
                              SaveMode : TAnimSaveMode);
begin
  DoSaveFrameSet(CharData.IndexedFile, Name, CharData.Index+1, SaveMode);
end;

procedure TfrmMain.SaveFrameSet(Name : String; CharData : TCharData;
                                SetIdx : Integer; SaveMode : TAnimSaveMode);
begin
  DoSaveFrameSet(CharData.IndexedFile, Name, CharData.Index + SetIdx, SaveMode);
end;

procedure TfrmMain.DoSaveFrameSet(
  IndexedFile : TIndexedFile;
  Name : String; Index : Integer;
  SaveMode : TAnimSaveMode);
var
  Data : TStream;
  FS : TFrameSet;
  I : Integer;
  AVIStream : TAVIStream;
//  PngImage : TPngImage;
begin
  Data := IndexedFile[Index].Data;
  Data.Seek(0, soFromBeginning);

   // Create the frame set
  FS := TFrameSet.Create;
  try
    FS.LoadFromStream(Data, Data.Size);
    if AnimHue > 0 then
      begin
        FS.UseHue := True;
        FS.HueEntry := AnimHueEntry;
        FS.TotalHueReplace := TotalHueReplace;
      end
    else
      FS.UseHue := False;

    case SaveMode of
      asmAVI :
        begin
          AVIStream := TAVIStream.Create(Name);
          try
            for I := 0 to FS.FrameCount-1 do
              begin
                EstablishBitmap(FS.X2-FS.X1, FS.Y2-FS.Y1, clWindow);
                FS.LoadFrameToBitmap(I, GlobalBitmap, Abs(FS.X1), Abs(FS.Y1));
                AVIStream.AddBitmap(GlobalBitmap);
              end;
          finally
            AVIStream.Free;
          end;
        end;
      asmBitmaps :
        for I := 0 to FS.FrameCount-1 do
          begin
            EstablishBitmap(FS.X2-FS.X1, FS.Y2-FS.Y1, clWindow);
            FS.LoadFrameToBitmap(I, GlobalBitmap, Abs(FS.X1), Abs(FS.Y1));
            // It'd be great to save these transparent, however, BMP doesn't
            // support transparency.  It would then be good to save as PNG
            // with transparency, but I don't have the Delphi version that
            // supports PNG.  The PNG_Graphic library that I created only
            // decompresses.
            GlobalBitmap.SaveToFile(Format(Name, [I]));
          end;
    end;
  finally
    FS.Free;
  end;
end;

// Loads a frame set into the Animation sections FrameSet variable.
// This frame set will be used for display on the next call to LoadCurrFrame.
procedure TfrmMain.UseFrameSet(
  IndexedFile : TIndexedFile;
  Index : Integer);
var
  Data : TStream;
begin
  // Toss the old frame set
  FrameSet.Free;
  FrameSet := Nil;

  // Get the cache block for the new frame set
  edtAnimFrameSet.Text := IntToStr(Index);
  Data := IndexedFile[Index].Data;
  Data.Seek(0, soFromBeginning);

   // Create the new frame set
  FrameSet := TFrameSet.Create;
  FrameSet.LoadFromStream(Data, Data.Size);

  CenterX := imgAnimation.Width div 2 - FrameSet.CX;
  CenterY := imgAnimation.Height div 2 - FrameSet.CY;
end;

procedure TfrmMain.tvAnimChange(Sender: TObject; Node: TTreeNode);
begin
  if not DisablingAnimation then
    begin
      CurrNode := Node.AbsoluteIndex;
      LoadCurrNode;
    end;
end;

procedure TfrmMain.edtAnimFrameSetKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = 13 then
    begin
      ShowMessage('Under construction');
      {
      UseFrameSet(StrToInt(edtAnimFrameSet.Text));
      CurrFrame := 0;
      LoadCurrFrame;
      }
    end;
end;

procedure TfrmMain.edtAnimFrameSetKeyPress(Sender: TObject; var Key: Char);
begin
  // If the key that was pressed is not a number or backspace, then disallow
  if not (Key in ['0'..'9', #08]) then Key := #00;
end;

procedure TfrmMain.actAnimateExecute(Sender: TObject);
begin
  if actAnimate.Checked then
    begin
      AnimTimer.Enabled := False;
      actAnimate.Checked := False;
      //btnAnimate.Caption := CapAnimateGo;
      //btnAnimate.Hint := CapAnimateGo;
    end
  else
    begin
      AnimTimer.Enabled := True;
      actAnimate.Checked := True;
      //btnAnimate.Caption := CapAnimateStop;
      //btnAnimate.Hint := CapAnimateStop;
    end;
end;

procedure TfrmMain.AnimTimerTimer(Sender: TObject);
begin
  if (FrameSet = Nil) or Processing or (FrameSet.FrameCount <= 0) then Exit;
  Processing := True;
  try
    LoadCurrFrame;
    Inc(CurrFrame);
    if CurrFrame >= FrameSet.FrameCount then CurrFrame := 0;
  finally
    Processing := False;
  end;
end;

procedure TfrmMain.cbAnimRectClick(Sender: TObject);
begin
  LoadCurrFrame;
end;

procedure TfrmMain.actSetBackColorExecute(Sender: TObject);
begin
  // Sets the background color for animations
  if ColorDialog.Execute then
    begin
      frmOptions.AnimBkColor := ColorDialog.Color;
      frmOptions.Save;
      LoadCurrFrame;
    end;
end;

procedure TfrmMain.actSetCharColorExecute(Sender: TObject);
begin
  ShowHues(AnimHue, AnimHueEntry);
  LoadCurrFrame;
end;

procedure TfrmMain.cbFullReplaceClick(Sender: TObject);
begin
  TotalHueReplace := cbFullReplace.Checked;
  LoadCurrFrame;
end;

procedure TfrmMain.tbAnimSpeedChange(Sender: TObject);
begin
  AnimTimer.Interval := tbAnimSpeed.Position;
end;

procedure TfrmMain.imgAnimationMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
  if ssLeft in Shift then
    begin
      CenterX := X;
      CenterY := Y;

      if not AnimTimer.Enabled then
        LoadCurrFrame;
    end;
end;

procedure TfrmMain.imgAnimationMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  imgAnimationMouseMove(Sender, Shift, X, Y);
end;

procedure TfrmMain.tvAnimContextPopup(Sender: TObject; MousePos: TPoint;
  var Handled: Boolean);
begin
  if not ActiveFilesOK then
    Exit;

  FActiveAnimationNode := tvAnim.GetNodeAt(MousePos.X, MousePos.Y);
  if FActiveAnimationNode <> Nil then
    begin
      tvAnim.Selected := FActiveAnimationNode;
      if FActiveAnimationNode.Parent = Nil then
        Handled := True;
    end
  else
    Handled := True;
end;

procedure TfrmMain.mnuAnimSaveAsBitmapsClick(Sender: TObject);
var
  FName, FFull, FExt : String;
begin
  if FActiveAnimationNode = Nil then Exit;

  sdAnim.FileName := FActiveAnimationNode.Text + '.bmp';
  if sdAnim.Execute then
    begin
      FName := ExtractFileName(sdAnim.FileName);
      FExt := ExtractFileExt(sdAnim.FileName);
      Delete(FName, Length(FName) - Length(FExt) + 1, Length(FExt));
      FFull := ExtractFilePath(sdAnim.FileName) + FName + '%0.2d' + FExt;
      If MessageDlg(Format(MsgAnimSaveVerify,[FName, FExt]),
                    mtConfirmation, mbOKCancel, 0) = mrOK then
        case FActiveAnimationNode.Level of
          1 : SaveCharacter(FFull, TCharData(FActiveAnimationNode.Data), asmBitmaps);
          2 : SaveAction(FFull, TCharData(FActiveAnimationNode.Data),asmBitmaps);
          3 : SaveFrameSet(FFull, TCharData(FActiveAnimationNode.Parent.Data), FActiveAnimationNode.Index, asmBitmaps);
        end;
    end;
end;

procedure TfrmMain.mnuAnimSaveAsAVIClick(Sender: TObject);
begin
  if FActiveAnimationNode = Nil then Exit;

  sdAnimAVI.FileName := FActiveAnimationNode.Text + '.avi';
  if sdAnimAVI.Execute and TestFileOverwrite(sdAnimAVI.FileName)then
    case FActiveAnimationNode.Level of
      1 : SaveCharacter(sdAnimAVI.FileName, TCharData(FActiveAnimationNode.Data),asmAVI);
      2 : SaveAction(sdAnimAVI.FileName, TCharData(FActiveAnimationNode.Data),asmAVI);
      3 : SaveFrameSet(sdAnimAVI.FileName, TCharData(FActiveAnimationNode.Parent.Data), FActiveAnimationNode.Index,asmAVI);
    end;
end;

procedure TfrmMain.mnuAnimDetailsClick(Sender: TObject);
var
  Data : TCharData;
  frmDetails : TfrmDetails;
  S : Integer;
begin
  if not ActiveFilesOK then Exit;

  // Pull up details on FActiveAnimationNode
  Data := Nil;
  case FActiveAnimationNode.Level of
    0 : Exit;
    1 : Data := TCharData(FActiveAnimationNode.Data);
    2 : Data := TCharData(FActiveAnimationNode.Parent.Data);
    3 : Data := TCharData(FActiveAnimationNode.Parent.Parent.Data);
  end;

   // Need to add on TActionData and index to get actual Index for frames lower
   // than the top level.

  Application.CreateForm(TfrmDetails, frmDetails);
  with frmDetails, Data do
    begin
      Caption := 'Details for ' + FActiveAnimationNode.Text;
      Memo.Clear;
      S := Memo.SelStart;
      Memo.Lines.Add(Format('Model Number: %d (0x%x)', [Data.Model, Data.Model]));
      BoldIt(Memo, S, 13); S := Memo.SelStart;
      Memo.Lines.Add(Format('Index: %d (0x%x)', [Index, Index]));
      BoldIt(Memo, S, 7); S := Memo.SelStart;
//      Memo.Lines.Add(Format('Type: %x', [ChType]));
//      BoldIt(Memo, S, 5); S := Memo.SelStart;
      Memo.Lines.Add(Format('Handle: %d (0x%x)', [Data.IndexedFile[Index].Handle, Data.IndexedFile[Index].Handle]));
      BoldIt(Memo, S, 7); //S := Memo.SelStart;
//      Memo.Lines.Add(Format('Tag: %d (0x%x)', [FIdxMain[Index].Tag, FIdxMain[Index].Tag]));
//      BoldIt(Memo, S, 4); S := Memo.SelStart;
//      Memo.Lines.Add(Format('Usage Number: %d (0x%x)', [Data.UsageNumber, Data.UsageNumber]));
//      BoldIt(Memo, S, 13);
      Memo.SelStart := 0;
      Show;
    end;
end;

 //-----------------------------------------------------------------------------
 //  ARTWORK
 //-----------------------------------------------------------------------------

procedure TfrmMain.ActivateArtwork;
var
  Panel : TStatusPanel;
begin
  if not frmOptions.VerifyFile('ARTIDX.MUL') then exit;
  if not frmOptions.VerifyFile('ART.MUL') then exit;

  ActiveSection := asArtwork;
  ActiveFilesOK := False;
  FIdxMain := Nil;

  sbMain.SimpleText := 'Loading artwork index...';
  Screen.Cursor := crHourglass;
  Application.ProcessMessages;
  try
    FIdxMain := TIndexedFile.Create(frmOptions.GetPath('ARTIDX.MUL'),
                                    frmOptions.GetPath('ART.MUL'),
                                    FPatchFile, $04, True);

    FIdxMain.GetRangeInfo(0, 16383, FLandscapeStart, FLandscapeCount);
    FIdxMain.GetRangeInfo(16384, 32767, FStaticStart, FStaticCount);
    FIdxMain.GetRangeInfo(32768, 2147483647, FCharacterStart, FCharacterCount);
  finally
    Screen.Cursor := crDefault;
    sbMain.SimpleText := '';
    Application.ProcessMessages;
  end;
  ActiveFilesOK := True;

   // Set up the status bar
  sbMain.SimplePanel := False;
  sbMain.Panels.Clear;

  Panel := sbMain.Panels.Add;
  Panel.Width := 40;
  Panel.Bevel := pbNone;
  Panel.Text := 'Name';
  Panel.Alignment := taRightJustify;

  Panel := sbMain.Panels.Add;
  Panel.Width := 150;
  Panel.Text := '';

  Panel := sbMain.Panels.Add;
  Panel.Width := 60;
  Panel.Bevel := pbNone;
  Panel.Text := 'Model No.';

  Panel := sbMain.Panels.Add;
  Panel.Width := 100;
  Panel.Text := '0';

  Panel := sbMain.Panels.Add;
  Panel.Width := 10;
  Panel.Bevel := pbNone;
  Panel.Text := '';

  dgLandscape.RowCount := Screen.Height;
  dgStatic.RowCount := Screen.Height;
  dgCharacter.RowCount := Screen.Height;
  pnlArtworkResize(Self);
  pnlArtwork.BringToFront;

  if pcArtwork.ActivePage = tsLandscape then
    dgLandscape.SetFocus
  else if pcArtwork.ActivePage = tsStatic then
    dgStatic.SetFocus
  else if pcArtwork.ActivePage = tsCharacter then
    dgCharacter.SetFocus;
end;

procedure TfrmMain.DeactivateArtwork;
begin
  sbMain.SimplePanel := True;
  ActiveFilesOK := False;
  FIdxMain.Free;
  FIdxMain := Nil;
  ActiveSection := asNone;
end;

procedure TfrmMain.pnlArtworkResize(Sender: TObject);
var
  W : Integer;
begin
  if ActiveSection <> asArtwork then Exit;

  if pcArtwork.ActivePage = tsLandscape then
    W := dgLandscape.ClientWidth - 16
  else if pcArtwork.ActivePage = tsStatic then
    W := dgStatic.ClientWidth - 16
  else if pcArtwork.ActivePage = tsCharacter then
    W := dgCharacter.ClientWidth - 16
  else
    Exit;

  dgLandscape.ColCount := W div dgLandscape.DefaultColWidth;
  dgStatic.ColCount := W div dgStatic.DefaultColWidth;
  dgCharacter.ColCount := W div dgCharacter.DefaultColWidth;

  if not ActiveFilesOK then
    begin
      dgLandscape.RowCount := 0;
      dgStatic.RowCount := 0;
      dgCharacter.RowCount := 0;
    end
  else
    begin
      dgLandscape.RowCount := FLandscapeCount div dgLandscape.ColCount + 1;
      dgStatic.RowCount := FStaticCount div dgStatic.ColCount + 1;
      dgCharacter.RowCount := FCharacterCount div dgCharacter.ColCount + 1;
    end;

  dgLandscape.Refresh;
  dgStatic.Refresh;
  dgCharacter.Refresh;
end;

procedure TfrmMain.DrawArtCell(DrawGrid : TDrawGrid; Rect : TRect;
                               Handle : Integer; Style : TBlockStyle; W, H : Word;
                               Selected : Boolean);
begin
  if Style = bsInvalid then
    with DrawGrid.Canvas do
      begin
        Brush.Style := bsSolid;
        Brush.Color := clSilver;
        FillRect(Rect);
      end
  else
    begin
      if FIdxMain[Handle].Source = FPatchFile then
      EstablishBitmap(Rect.Right - Rect.Left, Rect.Bottom - Rect.Top,
                      clRed)
      else
      EstablishBitmap(Rect.Right - Rect.Left, Rect.Bottom - Rect.Top,
                      GridColors[Selected]);
      DrawBlock(FIdxMain[Handle], GlobalBitmap, 0, 0);
      DrawGrid.Canvas.Draw(Rect.Left,Rect.Top,GlobalBitMap);
    end;
end;

procedure TfrmMain.dgLandscapeDrawCell(Sender: TObject; Col, Row: Integer;
  Rect: TRect; State: TGridDrawState);
var
  BNum : Integer;
  Style : TBlockStyle;
  Width, Height : Word;
begin
  if not ActiveFilesOK then Exit;
  BNum := Row * dgLandscape.ColCount + Col;
  if BNum >= FLandscapeCount then Exit;

  GetBlockInfo(TIndexedData(FIdxMain[FLandscapeStart + BNum]), Width, Height, Style);
  DrawArtCell(dgLandscape, Rect, FLandscapeStart + BNum, Style, Width, Height, gdSelected in State);
end;

procedure TfrmMain.dgStaticDrawCell(Sender: TObject; Col, Row: Integer;
  Rect: TRect; State: TGridDrawState);
var
  BNum : Integer;
  Style : TBlockStyle;
  Width, Height : Word;
begin
  if not ActiveFilesOK then Exit;
  BNum := Row * dgStatic.ColCount + Col;
  if BNum >= FStaticCount then Exit;

  GetBlockInfo(TIndexedData(FIdxMain[FStaticStart + BNum]), Width, Height, Style);
  DrawArtCell(dgStatic, Rect, FStaticStart + BNum, Style, Width, Height, gdSelected in State);
end;

procedure TfrmMain.dgCharacterDrawCell(Sender: TObject; Col, Row: Integer;
  Rect: TRect; State: TGridDrawState);
var
  BNum : Integer;
  Style : TBlockStyle;
  Width, Height : Word;
begin
  if not ActiveFilesOK then Exit;
  BNum := Row * dgCharacter.ColCount + Col;
  if BNum >= FCharacterCount then Exit;

  GetBlockInfo(TIndexedData(FIdxMain[FCharacterStart + BNum]), Width, Height, Style);
  DrawArtCell(dgCharacter, Rect, FCharacterStart + BNum, Style, Width, Height, gdSelected in State);
end;

procedure TfrmMain.dgLandscapeDblClick(Sender: TObject);
begin
  FActiveArtGrid := agLandscape;
  FActiveBNum := dgLandscape.Row * dgLandscape.ColCount + dgLandscape.Col;
  mnuArtViewClick(Self);
end;

procedure TfrmMain.dgLandscapeMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  P : TPoint;
  ACol, ARow : Integer;
begin
  if (Button = mbRight) and ActiveFilesOK then
    begin
      mnuArtSaveAs.Enabled := True;
      mnuArtFindByNumber.Enabled := True;
      mnuArtFindByName.Enabled := True;
      mnuArtFindNext.Enabled := FindActive <> faNo;

      dgLandscape.MouseToCell(X, Y, ACol, ARow);
      dgLandscape.Row := ARow;
      dgLandscape.Col := ACol;

      FActiveArtGrid := agLandscape;
      FActiveBNum := dgLandscape.Row * dgLandscape.ColCount + dgLandscape.Col;
      if FActiveBNum < FLandscapeCount then
        begin
          P := dgLandscape.ClientToScreen(Point(X + 2, Y + 2));
          pmArt.Popup(P.X, P.Y);
        end;
    end;
end;

procedure TfrmMain.dgStaticDblClick(Sender: TObject);
begin
  FActiveArtGrid := agStatic;
  FActiveBNum := dgStatic.Row * dgStatic.ColCount + dgStatic.Col;
  mnuArtViewClick(Self);
end;

procedure TfrmMain.dgStaticMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  P : TPoint;
  ACol, ARow : Integer;
begin
  if (Button = mbRight) and ActiveFilesOK then
    begin
      mnuArtSaveAs.Enabled := True;
      mnuArtFindByNumber.Enabled := True;
      mnuArtFindByName.Enabled := True;
      mnuArtFindNext.Enabled := FindActive <> faNo;

      dgStatic.MouseToCell(X, Y, ACol, ARow);
      dgStatic.Row := ARow;
      dgStatic.Col := ACol;

      FActiveArtGrid := agStatic;
      FActiveBNum := dgStatic.Row * dgStatic.ColCount + dgStatic.Col;
      if FActiveBNum < FStaticCount then
        begin
          P := dgStatic.ClientToScreen(Point(X + 2, Y + 2));
          pmArt.Popup(P.X, P.Y);
        end;
    end;
end;

procedure TfrmMain.dgCharacterDblClick(Sender: TObject);
begin
  FActiveArtGrid := agCharacter;
  FActiveBNum := dgCharacter.Row * dgCharacter.ColCount + dgCharacter.Col;
  mnuArtViewClick(Self);
end;

procedure TfrmMain.dgCharacterMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  P : TPoint;
  ACol, ARow : Integer;
begin
  if (Button = mbRight) and ActiveFilesOK then
    begin
      mnuArtSaveAs.Enabled := True;
      mnuArtFindByNumber.Enabled := False;
      mnuArtFindByName.Enabled := False;
      mnuArtFindNext.Enabled := False;

      dgCharacter.MouseToCell(X, Y, ACol, ARow);
      dgCharacter.Row := ARow;
      dgCharacter.Col := ACol;

      FActiveArtGrid := agCharacter;
      FActiveBNum := dgCharacter.Row * dgCharacter.ColCount + dgCharacter.Col;
      if FActiveBNum < FCharacterCount then
        begin
          P := dgCharacter.ClientToScreen(Point(X + 2, Y + 2));
          pmArt.Popup(P.X, P.Y);
        end;
    end;
end;

function TfrmMain.GetArtIndexData : TIndexedData;
begin
  Result := Nil;
  case FActiveArtGrid of
    agLandscape:
      begin
        if FActiveBNum >= FLandscapeCount then Exit;
        Result := FIdxMain[FLandscapeStart + FActiveBNum];
      end;
    agStatic:
      begin
        if FActiveBNum >= FStaticCount then Exit;
        Result := FIdxMain[FStaticStart + FActiveBNum];
      end;
    agCharacter:
      begin
        if FActiveBNum >= FCharacterCount then Exit;
        Result := FIdxMain[FCharacterStart + FActiveBNum];
      end;
  end;
end;

function CheckDescription(Description : String) : String;
begin
  if Description = '' then
    Result := 'unlabled'
  else
    Result := Description
end;

procedure TfrmMain.mnuArtViewClick(Sender: TObject);
var
  frmArtView: TfrmArtView;
  Style : TBlockStyle;
  AWidth, AHeight : Word;
  Data : TIndexedData;
  ArtHandle : Integer;
begin
  if not ActiveFilesOK then Exit;

  Data := GetArtIndexData;
  if Data = Nil then Exit;

  if Data.Handle < $4000 then
    ArtHandle := Data.Handle
  else if Data.Handle < $8000 then
    ArtHandle := Data.Handle - $4000
  else
    ArtHandle := Data.Handle - $8000;

  GetBlockInfo(Data, AWidth, AHeight, Style);

  if Style <> bsInvalid then
    begin
      EstablishBitmap(AWidth, AHeight, clWindow);
      DrawBlock(Data, GlobalBitmap, 0, 0);
      frmArtView := TfrmArtView.Create(Application);
      frmArtView.ShowBitmap(
        GetDescription(Data.Handle, UnlabeledArtBlock),
        GlobalBitmap,
        'Index:',
        ArtHandle, False);
    end;
end;

procedure TfrmMain.mnuArtSaveAsClick(Sender: TObject);
var
  Style : TBlockStyle;
  AWidth, AHeight : Word;
  Data : TIndexedData;
begin
  if not ActiveFilesOK then Exit;

  Data := Nil;
  case FActiveArtGrid of
    agLandscape:
      begin
        if FActiveBNum >= FLandscapeCount then Exit;
        Data := FIdxMain[FLandscapeStart + FActiveBNum];
        SavePicture.FileName := GetDescription(Data.Handle, UnlabeledArtBlock);
      end;
    agStatic:
      begin
        if FActiveBNum >= FStaticCount then Exit;
        Data := FIdxMain[FStaticStart + FActiveBNum];
        SavePicture.FileName := GetDescription(Data.Handle, UnlabeledArtBlock);
      end;
    agCharacter:
      begin
        if FActiveBNum >= FCharacterCount then Exit;
        Data := FIdxMain[FCharacterStart + FActiveBNum];
        SavePicture.FileName := 'CharacterTest';
      end;
  end;
  Assert(Data <> Nil);
  SavePicture.FileName := SavePicture.FileName + '.bmp';

  GetBlockInfo(Data, AWidth, AHeight, Style);

  if SavePicture.Execute and TestFileOverwrite(savePicture.FileName) then
    begin
      EstablishBitmap(AWidth, AHeight, clWindow);
      DrawBlock(Data, GlobalBitmap, 0, 0);
      GlobalBitMap.SaveToFile(savePicture.FileName);
    end;
end;

procedure WriteLine( var TextFile : TFileStream; Line : String );
var
  PLine : PChar;
begin
  Line := Line + #13 + #10;
  PLine := PChar(Line);
  if (PLine <> Nil) and (Length(Line) > 0) then
    TextFile.Write(PLine[0], Length(Line));
end;

procedure TfrmMain.mnuArtSaveAllClick(Sender: TObject);
var
  I : Integer;
  Data : TIndexedData;
  ArtHandle : Integer;
  Desc : String;
  TextFile : TFileStream;
  State : TArtGroup;
begin
  if not ActiveFilesOK then Exit;
  sdArtAll.FileName := 'UO Art Information.xml';
  if sdArtAll.Execute and TestFileOverwrite(sdArtAll.FileName) then
    begin
      Screen.Cursor := crHourglass;
      Application.ProcessMessages;
      try
        TextFile := TFileStream.Create(sdArtAll.FileName, fmCreate or fmShareExclusive);
        try
          State := agLandscape;
          WriteLine(TextFile, '<!-- This file was generated by InsideUO -->');
          WriteLine(TextFile, '<UltimaOnline>');
          WriteLine(TextFile, '  <Art>');
          WriteLine(TextFile, Format('    <Group Key="%s">', [ArtGroupStrings[State]]));
          for I := 0 to FIdxMain.Count-1 do
            begin
              Data := FIdxMain[I];
              if Data <> Nil then
                begin
                  if Data.Handle < $4000 then
                    begin
                      ArtHandle := Data.Handle;
                      if State <> agLandscape then
                        raise Exception.Create('Art blocks out of order');
                    end
                  else if Data.Handle < $8000 then
                    begin
                      ArtHandle := Data.Handle - $4000;
                      if State = agLandscape then
                        begin
                          State := agStatic;
                          WriteLine(TextFile, '    </Group>');
                          WriteLine(TextFile, Format('    <Group Key="%s">', [ArtGroupStrings[State]]));
                        end
                      else if State = agCharacter then
                        raise Exception.Create('Art blocks out of order');
                    end
                  else
                    begin
                      ArtHandle := Data.Handle - $8000;
                      if State = agLandscape then
                        raise Exception.Create('Missing art blocks');
                      if State = agStatic then
                        begin
                          State := agCharacter;
                          WriteLine(TextFile, '    </Group>');
                          WriteLine(TextFile, Format('    <Group Key="%s">', [ArtGroupStrings[State]]));
                        end;
                    end;
                  Desc := GetDescription(Data.Handle, '');
                  if (State = agCharacter) then
                    Desc := '';
                  if (Desc = '') then
                    WriteLine(TextFile, Format('      <Block Key="0x%x" />', [ArtHandle]))
                  else
                    WriteLine(TextFile, Format('      <Block Key="0x%x">%s</Block>', [ArtHandle, Desc]));
                end;
              end;
            WriteLine(TextFile, '    </Group>');
            WriteLine(TextFile, '  </Art>');
            WriteLine(TextFile, '</UltimaOnline>');
          finally
            TextFile.Free;
          end;
        finally
          Screen.Cursor := crDefault;        
        end;
    end;
end;

procedure TfrmMain.mnuArtCopyClick(Sender: TObject);
var
  Style : TBlockStyle;
  AWidth, AHeight : Word;
  Data : TIndexedData;
begin

   { THIS SHOULD BE COMBINABLE WITH ABOVE ROUTINE }

  if not ActiveFilesOK then Exit;

  Data := Nil;
  case FActiveArtGrid of
    agLandscape:
      begin
        if FActiveBNum >= FLandscapeCount then Exit;
        Data := FIdxMain[FLandscapeStart + FActiveBNum];
      end;
    agStatic:
      begin
        if FActiveBNum >= FStaticCount then Exit;
        Data := FIdxMain[FStaticStart + FActiveBNum];
      end;
    agCharacter:
      begin
        if FActiveBNum >= FCharacterCount then Exit;
        Data := FIdxMain[FCharacterStart + FActiveBNum];
      end;
  end;
  Assert(Data <> Nil);

  GetBlockInfo(Data, AWidth, AHeight, Style);
  EstablishBitmap(AWidth, AHeight, clWindow);
  DrawBlock(Data, GlobalBitmap, 0, 0);
  Clipboard.Assign(GlobalBitMap);
end;

function TfrmMain.FindByName(Start : Integer; BlkName : String) : Integer;
var
  I, Loc, PhyBNum : Integer;
  Desc : Array[0..19] of Char;
  SDesc : String;
  FS : TFileStream;
begin
   // Search through TILEDATA for an art block that contains BlkName
  Screen.Cursor := crHourglass;
  try
    Result := -1;
    FS := TFileStream.Create(frmOptions.GetPath('TILEDATA.MUL'), fmOpenRead or fmShareDenyNone);
    try
      case FActiveArtGrid of
        agLandscape:
          begin
            I := Start;
            SDesc := '';
            while (I < FLandscapeCount) and (Pos(BlkName, SDesc) = 0) do
              begin
                PhyBNum := FIdxMain[FLandscapeStart+I].Handle;
                Loc := ((PhyBNum shr 5) + 1) * 4 + PhyBNum * 26;
                FS.Seek(Loc + 6, soFromBeginning);
                FS.Read(Desc, SizeOf(Desc));
                SDesc := LowerCase(Desc);
                Inc(I);
              end;
            if I < FLandscapeCount then
              begin
                Dec(I);
                dgLandscape.Col := I mod dgLandscape.ColCount;
                dgLandscape.Row := I div dgLandscape.ColCount;
                Result := I;
              end
            else
              raise Exception.Create(ErrorTileNotFound);
          end;
        agStatic:
          begin
            I := Start;
            SDesc := '';
            while (I < FStaticCount) and (Pos(BlkName, SDesc) = 0) do
              begin
                PhyBNum := FIdxMain[FStaticStart+I].Handle - 16384;
                Loc := $68800 + ((PhyBNum shr 5) + 1) * 4 + PhyBNum * 37 + 17;
                FS.Seek(Loc, soFromBeginning);
                FS.Read(Desc, SizeOf(Desc));
                SDesc := LowerCase(Desc);
                Inc(I);
              end;
            if I < FStaticCount then
              begin
                Dec(I);
                dgStatic.Col := I mod dgStatic.ColCount;
                dgStatic.Row := I div dgStatic.ColCount;
                Result := I;
              end
            else
              raise Exception.Create(ErrorTileNotFound);
          end;
      end;
    finally
      FS.Free;
    end;
  finally
    Screen.Cursor := crDefault;
  end;
end;

function TfrmMain.FindByNumber(Start, BlkNum : Integer) : Integer;
var
  I : Integer;
begin
   // Search through the index for an art handle that contains BlkNum
  Result := -1;
  case FActiveArtGrid of
    agLandscape:
      begin
        I := 0;
        while (I < FLandscapeCount) and
              (FIdxMain[FLandscapeStart + I].Handle <> BlkNum) do
          Inc(I);
        if I < FLandscapeCount then
          begin
            dgLandscape.Col := I mod dgLandscape.ColCount;
            dgLandscape.Row := I div dgLandscape.ColCount;
            Result := I;
          end
        else
          raise Exception.Create(ErrorTileNotFound);
      end;
    agStatic:
      begin
        I := 0;
        while (I < FStaticCount) and
              (FIdxMain[FStaticStart + I].Handle <> BlkNum) do
          Inc(I);
        if I < FStaticCount then
          begin
            dgStatic.Col := I mod dgStatic.ColCount;
            dgStatic.Row := I div dgStatic.ColCount;
            Result := I;
          end
        else
          raise Exception.Create(ErrorTileNotFound);
          // Result := -1;
      end;
  end;
end;

procedure TfrmMain.mnuArtFindbyNameClick(Sender: TObject);
begin
  if frmFindBlockName.ShowModal <> mrOK then Exit;
  FindBlockName := LowerCase(frmFindBlockName.edtFindBlockName.Text);
  Application.ProcessMessages;

  FindTag := FindByName(0, FindBlockName);
  if FindTag < 0 then
    FindActive := faNo
  else
    FindActive := faName;
  mnuArtFindNext.Enabled := FindActive <> faNo;
end;

procedure TfrmMain.mnuArtFindbyNumberClick(Sender: TObject);
begin
  if frmFindBlockNum.ShowModal <> mrOK then Exit;
  FindBlockNum := StrToInt(frmFindBlockNum.edtBlockNum.Text);
  Application.ProcessMessages;

  FindTag := FindByNumber(0, FindBlockNum);
  if FindTag < 0 then
    FindActive := faNo
  else
    FindActive := faNumber;
  mnuArtFindNext.Enabled := FindActive <> faNo;
end;

procedure TfrmMain.mnuArtFindNextClick(Sender: TObject);
begin
  Application.ProcessMessages;
  case FindActive of
    faNumber : FindTag := FindByNumber(FindTag+1, FindBlockNum);
    faName : FindTag := FindByName(FindTag+1, FindBlockName);
  end;

   // We didn't find a next entry
  if FindTag < 0 then
    begin
      FindActive := faNo;
      mnuArtFindNext.Enabled := False;
    end;
end;

type
  TTileDataFlags = (
    tdfBackground,
    tdfWeapon,
    tdfTransparent,
    tdfTranslucent,
    tdfWall,
    tdfDamaging,
    tdfImpassable,
    tdfWet,
    tdfUnknown,
    tdfSurface,
    tdfBridge,
    tdfGeneric,
    tdfWindow,
    tdfNoShoot,
    tdfPrefixA,
    tdfPrefixAn,
    tdfInternal,
    tdfFoliage,
    tdfPartialHue,
    tdfUnknown1,
    tdfMap,
    tdfContainer,
    tdfWearable,
    tdfLightSource,
    tdfAnimated,
    tdfNoDiagonal,
    tdfUnknown2,
    tdfArmor,
    tdfRoof,
    tdfDoor,
    tdfStairBack,
    tdfStairRight);
  TTileDataFlagSet = set of TTileDataFlags;

const
  TileDataFlagStrings : Array[TTileDataFlags] of String = (
    'Background',
    'Weapon',
    'Transparent',
    'Translucent',
    'Wall',
    'Damaging',
    'Impassable',
    'Wet',
    'Unkown',
    'Surface',
    'Bridge',
    'Generic',
    'Window',
    'NoShoot',
    'PrefixA',
    'PrefixAn',
    'Internal',
    'Foliage',
    'PartialHue',
    'Unknown1',
    'Map',
    'Container',
    'Wearable',
    'LightSource',
    'Animated',
    'NoDiagonal',
    'Unknown2',
    'Armor',
    'Roof',
    'Door',
    'StairBack',
    'StairRight' );

function ConvertIntegerToFlags(I : Integer) : TTileDataFlagSet;
begin
  result := [];
  if (I AND $00000001 > 0) then Include(result, tdfBackground);
  if (I AND $00000002 > 0) then Include(result, tdfWeapon);
  if (I AND $00000004 > 0) then Include(result, tdfTransparent);
  if (I AND $00000008 > 0) then Include(result, tdfTranslucent);
  if (I AND $00000010 > 0) then Include(result, tdfWall);
  if (I AND $00000020 > 0) then Include(result, tdfDamaging);
  if (I AND $00000040 > 0) then Include(result, tdfImpassable);
  if (I AND $00000080 > 0) then Include(result, tdfWet);
  if (I AND $00000100 > 0) then Include(result, tdfUnknown);
  if (I AND $00000200 > 0) then Include(result, tdfSurface);
  if (I AND $00000400 > 0) then Include(result, tdfBridge);
  if (I AND $00000800 > 0) then Include(result, tdfGeneric);
  if (I AND $00001000 > 0) then Include(result, tdfWindow);
  if (I AND $00002000 > 0) then Include(result, tdfNoShoot);
  if (I AND $00004000 > 0) then Include(result, tdfPrefixA);
  if (I AND $00008000 > 0) then Include(result, tdfPrefixAn);
  if (I AND $00010000 > 0) then Include(result, tdfInternal);
  if (I AND $00020000 > 0) then Include(result, tdfFoliage);
  if (I AND $00040000 > 0) then Include(result, tdfPartialHue);
  if (I AND $00080000 > 0) then Include(result, tdfUnknown1);
  if (I AND $00100000 > 0) then Include(result, tdfMap);
  if (I AND $00200000 > 0) then Include(result, tdfContainer);
  if (I AND $00400000 > 0) then Include(result, tdfWearable);
  if (I AND $00800000 > 0) then Include(result, tdfLightSource);
  if (I AND $01000000 > 0) then Include(result, tdfAnimated);
  if (I AND $02000000 > 0) then Include(result, tdfNoDiagonal);
  if (I AND $04000000 > 0) then Include(result, tdfUnknown2);
  if (I AND $08000000 > 0) then Include(result, tdfArmor);
  if (I AND $10000000 > 0) then Include(result, tdfRoof);
  if (I AND $20000000 > 0) then Include(result, tdfDoor);
  if (I AND $40000000 > 0) then Include(result, tdfStairBack);
  if (I AND $80000000 > 0) then Include(result, tdfStairRight);
end;

function GetFlagsString(FlagSet : TTileDataFlagSet) : String;
var
  Flag : TTileDataFlags;
begin
  for Flag := tdfBackground to tdfStairRight do
    begin
      if Flag in FlagSet then
        begin
          if result <> '' then
            result := result + ', ';
          result := result + TileDataFlagStrings[Flag];
        end;
    end;
end;

procedure TfrmMain.mnuArtDetailsClick(Sender: TObject);
var
  S : Integer;
  Data : TIndexedData;
  Group : String;
  FS : TFileStream;
  Loc : Integer;
  SectionCount : Integer;
  GroupIndex : Integer;
  BlockNum : Integer;
  ArtName : String;
  LandData : TLandTileData;
  StaticData : TStaticTileData;
  FlagSet : TTileDataFlagSet;
begin
  if not ActiveFilesOK then Exit;

  Data := GetArtIndexData;

  Application.CreateForm(TfrmDetails, frmDetails);
  with frmDetails, Data do
    begin
      //Caption := 'Details for ' + GetDescription(Handle, UnlabeledArtBlock);
      Memo.Clear;
      S := Memo.SelStart;

      Memo.Lines.Add(Format('Overall Index: %d (0x%x)', [Handle, Handle]));
      BoldIt(Memo, S, 14); S := Memo.SelStart;

      Memo.Lines.Add(Format('Tag: %0.4x',[Tag]));
      BoldIt(Memo, S, 4); S := Memo.SelStart;

      if Handle < $4000 then
        begin
          Group := 'Landscape Tiles';
          GroupIndex := Handle;
        end
      else if Handle < $8000 then
        begin
          Group := 'Static Tiles';
          GroupIndex := Handle - $4000;
        end
      else
        begin
          Group := 'Character Tests';
          GroupIndex := Handle - $8000;
        end;
      Memo.Lines.Add(Format('Group: %s', [Group]));
      BoldIt(Memo, S, 6); S := Memo.SelStart;

      Memo.Lines.Add(Format('Group Relative Lookup Offset: %d (0x%x)', [FActiveBNum, FActiveBNum]));
      BoldIt(Memo, S, 29); S := Memo.SelStart;

      Memo.Lines.Add(Format('Group Index: %d (0x%x)', [GroupIndex, GroupIndex]));
      BoldIt(Memo, S, 12); S := Memo.SelStart;

      Memo.Lines.Add('');

      FS := TFileStream.Create(frmOptions.GetPath('TILEDATA.MUL'),
                               fmOpenRead or fmShareDenyNone);
      try
        if Handle < 16384 then // Land tiles
          begin
            BlockNum := Handle;
            SectionCount := (BlockNum shr 5) + 1;
            Loc := SectionCount * 4 + BlockNum * 26;
            FS.Seek(Loc, soFromBeginning);
            FS.Read(LandData, SizeOf(LandData));

            FlagSet := ConvertIntegerToFlags(LandData.Flags);

            ArtName := Trim(StrPas(LandData.Name));
            if ArtName = '' then
              ArtName := UnlabeledArtBlock
            else if tdfPrefixA in FlagSet then
              Caption := 'Details for a ' + ArtName
            else if tdfPrefixAn in FlagSet then
              Caption := 'Details for an ' + ArtName
            else
              Caption := 'Details for ' + ArtName;

            Memo.Lines.Add(Format('Flags: %s (0x%x)', [GetFlagsString(FlagSet), LandData.Flags]));
            BoldIt(Memo, S, 6); S := Memo.SelStart;

            Memo.Lines.Add(Format('Texture Id: %d (0x%x)', [LandData.TextureId, LandData.TextureId]));
            BoldIt(Memo, S, 11); S := Memo.SelStart;

            Memo.Lines.Add(Format('Name: %s', [LandData.Name]));
            BoldIt(Memo, S, 5); S := Memo.SelStart;
          end
        else // Static tiles
          begin
            BlockNum := Handle - 16384;
            SectionCount := (BlockNum shr 5) + 1;
            Loc := $68800 + SectionCount * 4 + BlockNum * 37;
            FS.Seek(Loc, soFromBeginning);
            FS.Read(StaticData, SizeOf(StaticData));

            FlagSet := ConvertIntegerToFlags(StaticData.Flags);

            ArtName := Trim(StrPas(StaticData.Name));
            if ArtName = '' then
              ArtName := UnlabeledArtBlock
            else if tdfPrefixA in FlagSet then
              Caption := 'Details for a ' + ArtName
            else if tdfPrefixAn in FlagSet then
              Caption := 'Details for an ' + ArtName
            else
              Caption := 'Details for ' + ArtName;

            Memo.Lines.Add(Format('Flags: %s (0x%x)', [GetFlagsString(FlagSet), StaticData.Flags]));
            BoldIt(Memo, S, 6); S := Memo.SelStart;

            Memo.Lines.Add(Format('Weight: %d (0x%x)', [StaticData.Weight, StaticData.Weight]));
            BoldIt(Memo, S, 7); S := Memo.SelStart;

            Memo.Lines.Add(Format('Quality: %d (0x%x)', [StaticData.Quality, StaticData.Quality]));
            BoldIt(Memo, S, 8); S := Memo.SelStart;

            Memo.Lines.Add(Format('Unknown: %d (0x%x)', [StaticData.Unknown, StaticData.Unknown]));
            BoldIt(Memo, S, 8); S := Memo.SelStart;

            Memo.Lines.Add(Format('Unknown1: %d (0x%x)', [StaticData.Unknown1, StaticData.Unknown1]));
            BoldIt(Memo, S, 9); S := Memo.SelStart;

            Memo.Lines.Add(Format('Quantity: %d (0x%x)', [StaticData.Quantity, StaticData.Quantity]));
            BoldIt(Memo, S, 9); S := Memo.SelStart;

            Memo.Lines.Add(Format('Anim Id: %d (0x%x)', [StaticData.AnimId, StaticData.AnimId]));
            BoldIt(Memo, S, 8); S := Memo.SelStart;

            Memo.Lines.Add(Format('Unknown2: %d (0x%x)', [StaticData.Unknown2, StaticData.Unknown2]));
            BoldIt(Memo, S, 9); S := Memo.SelStart;

            Memo.Lines.Add(Format('Hue: %d (0x%x)', [StaticData.Hue, StaticData.Hue]));
            BoldIt(Memo, S, 4); S := Memo.SelStart;

            Memo.Lines.Add(Format('Unknown3: %d (0x%x)', [StaticData.Unknown3, StaticData.Unknown3]));
            BoldIt(Memo, S, 9); S := Memo.SelStart;

            Memo.Lines.Add(Format('Height: %d (0x%x)', [StaticData.Height, StaticData.Height]));
            BoldIt(Memo, S, 7); S := Memo.SelStart;

            Memo.Lines.Add(Format('Name: %s', [StaticData.Name]));
            BoldIt(Memo, S, 5); S := Memo.SelStart;
          end;
      finally
        FS.Free;
      end;

      Memo.SelStart := 0;
      Show;
    end;
end;

procedure TfrmMain.SelectCell(Sender: TObject; ACol, ARow: Integer;
  var CanSelect: Boolean);
var
  Data : TIndexedData;
  ArtHandle : Integer;
begin
  if Sender = dgLandscape then
    FActiveArtGrid := agLandscape
  else if Sender = dgStatic then
    FActiveArtGrid := agStatic
  else if Sender = dgCharacter then
    FActiveArtGrid := agCharacter;
  FActiveBNum := ARow * (Sender as TDrawGrid).ColCount + ACol;

  Data := GetArtIndexData;
  if Data = Nil then Exit;

   // Set the status bar
  FindTag := ARow * (Sender as TDrawGrid).ColCount + ACol;
  if Data.Handle < $4000 then
    ArtHandle := Data.Handle
  else if Data.Handle < $8000 then
    ArtHandle := Data.Handle - $4000
  else
    ArtHandle := Data.Handle - $8000;

  sbMain.Panels[1].Text := GetDescription(Data.Handle, UnlabeledArtBlock);
  sbMain.Panels[3].Text := Format('%d (0x%x)', [ArtHandle, ArtHandle]);
  sbMain.Refresh;
end;

procedure TfrmMain.dgStaticKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = 114 then // F3
    mnuArtFindNextClick(Self)
  else if (ssCtrl in Shift) and (Key = 70) then // Ctrl-F
    mnuArtFindByNameClick(Self);
end;

procedure TfrmMain.dgLandscapeKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = 114 then // F3
    mnuArtFindNextClick(Self)
  else if (ssCtrl in Shift) and (Key = 70) then // Ctrl-F
    mnuArtFindByNameClick(Self);
end;

 //-----------------------------------------------------------------------------
 //  FONTS
 //-----------------------------------------------------------------------------

procedure TfrmMain.ActivateFonts;
var
  FCell : TFontCellHeader;
  MemIdx : TMemoryStream;
  CharSet : TCharacterSet;
  I : Integer;
  Item : TListItem;
  CharSetCount : Integer;
begin
  if not frmOptions.VerifyFile('FONTS.MUL') then exit;

  pnlFonts.BringToFront;
  ActiveSection := asFonts;
  ActiveFilesOK := False;
  FontHue := -1;

  FMain := Nil;
  MemIdx := TMemoryStream.Create;
  try
    FMain := TFileStream.Create(frmOptions.GetPath('FONTS.MUL'),
                                fmOpenRead or fmShareDenyNone);
    try
      MemIdx.LoadFromStream(FMain);
    finally
      FMain.Free;
    end;

    CharSetCount := 0;
    MemIdx.Seek(0, soFromBeginning);
    while MemIdx.Position < MemIdx.Size do
      begin
        Item := lvFonts.Items.Add;
        Item.Caption := IntToStr(CharSetCount);
        CharSet := TCharacterSet.Create;
        MemIdx.Read(CharSet.Header, SizeOf(CharSet.Header));
        For I := 0 to 223 do
          begin
            MemIdx.Read(FCell, SizeOf(FCell));
            with CharSet.Characters[I] do
              begin
                Width := FCell.fWidth;
                Height := FCell.fHeight;
                Unknown := FCell.fUnk;
                GetMem(Data, Width * Height * 2);
                MemIdx.Read(Data^, Width * Height * 2);
              end;
          end;
        Item.Data := CharSet;
        Inc(CharSetCount);
      end;
  finally
    MemIdx.Free;
  end;

  ActiveFilesOK := True;
  pnlFontsResize(Self);
  lvFonts.Selected := lvFonts.Items[0];
  edtFontPhrase.SetFocus;
end;

procedure TfrmMain.DeactivateFonts;
begin
  ActiveFilesOK := False;
  lvFonts.Items.BeginUpdate;
  try
    lvFonts.Items.Clear;
  finally
    lvFonts.Items.EndUpdate;
  end;
  ActiveSection := asNone;
end;

type
  TWordArr = Array[0..0] of Word;
  PWordArr = ^TWordArr;

procedure TfrmMain.DrawFontCell(C : TCanvas; R : TRect; Character : TCharacter);
var
  X, Y : Integer;
  Out_Line, In_Line : PWordArr;
  Bitmap : TBitmap;
begin
  if (Character.Width = 0) or (Character.Height = 0) then Exit;
  Bitmap := CreateBitmap(Character.Width, Character.Height, 0);
  try
    For Y := 0 to Character.Height-1 do
      begin
        Out_Line := PWordArr(Bitmap.ScanLine[Y]);
        In_Line := PWordArr(PChar(Character.Data) + Y * Character.Width * 2);
        For X := 0 to Character.Width-1 do
          if In_Line^[X] <> 0 then
            begin
              if FontHue > 0 then
                Out_Line^[X] := FontHueEntry.ColorTable[ConvertToGray(In_Line^[X])]
              else
                Out_Line^[X] := In_Line^[X];
            end
          else
            Out_Line^[X] := 0;
      end;
    Bitmap.Transparent := True;
    Bitmap.TransparentMode := tmFixed;
    Bitmap.TransparentColor := clBlack;
    C.Draw(R.Left, R.Top, Bitmap);
  finally
    Bitmap.Free;
  end;
end;

procedure TfrmMain.dgFontsDrawCell(Sender: TObject; Col, Row: Integer;
  Rect: TRect; State: TGridDrawState);
var
  CellNum : Integer;
  C : Char;
  Size : TSize;
  CL : LongInt;
begin
  if not ActiveFilesOK or (lvFonts.Selected = Nil) then Exit;
  with TCharacterSet(lvFonts.Selected.Data) do
    begin
      CellNum := Row * dgFonts.ColCount + Col;
      if CellNum < 224 then
        begin
          if(CellNum = 0) then
            C := ' '
          else
            C := Chr(CellNum + 32);

          with dgFonts.Canvas do
            begin
              Brush.Color := GridColors[gdSelected in State];
              Brush.Style := bsSolid;
              CL := ColorToRGB(Brush.Color);
              Font.Color := TColor(DecreaseColorBrightness(CL, 15));
              Font.Name := 'Arial';
              Font.Height := Rect.Bottom - Rect.Top;
              Size := TextExtent(C);
              TextRect(Rect,(Rect.Right+Rect.Left-Size.cx) div 2,Rect.Top,C);
              DrawFontCell(dgFonts.Canvas, Rect, Characters[CellNum]);
            end;
        end;
    end;
end;

procedure TfrmMain.dgFontsDblClick(Sender: TObject);
var
  CellNum: Integer;
  C : Char;
begin
  CellNum := dgFonts.Row * dgFonts.ColCount + dgFonts.Col;
  if CellNum < 224 then
    begin
      if(CellNum = 0) then
        C := ' '
      else
        C := Chr(CellNum + 32);

      edtFontPhrase.Text := edtFontPhrase.Text + C;
    end;
end;

procedure TfrmMain.pnlFontsResize(Sender: TObject);
var
  W : Integer;
begin
  if ActiveSection <> asFonts then Exit;

  W := dgFonts.ClientWidth - 16;
  dgFonts.ColCount := W div dgFonts.DefaultColWidth;

  if not ActiveFilesOK then
    dgFonts.RowCount := 0
  else
    dgFonts.RowCount := 224 div dgFonts.ColCount + 1;

  dgFonts.Refresh;
end;

procedure TfrmMain.lvFontsChange(Sender: TObject; Item: TListItem;
  Change: TItemChange);
begin
  if (ActiveItem <> Item) then
    begin
      pnlFontsResize(Self);
      ActiveItem := Item;
    end;
end;

procedure TfrmMain.lvFontsDeletion(Sender: TObject; Item: TListItem);
var
  I : Integer;
begin
  with TCharacterSet(Item.Data) do
    begin
      For I := 0 to 223 do
        FreeMem(Characters[I].Data,
                Characters[I].Width * Characters[I].Height * 2);
      Free;
    end;
end;

procedure TfrmMain.edtFontPhraseKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then
    begin
      Key := #00;
      actDrawPhraseExecute(self);
    end;
end;

procedure TfrmMain.actDrawPhraseExecute(Sender: TObject);
var
  S : String;
  W, H : Integer;
  CharNum : Byte;
  I : Integer;
  X : Integer;
  CS : TCharacterSet;
  frmArtView: TfrmArtView;
begin
  CS := TCharacterSet(lvFonts.Selected.Data);
  S := edtFontPhrase.Text;
  W := 0;
  H := 0;
  for I := 1 to Length(S) do
    begin
      if S[I] = ' ' then CharNum := 0 else CharNum := Ord(S[I])-32;
      Inc(W, CS.Characters[CharNum].Width);
      if H < CS.Characters[CharNum].Height then
        H := CS.Characters[CharNum].Height;
    end;
  W := Max(32, W);
  H := Max(16, H);

  EstablishBitmap(W,H,clWindow);

  X := 0;
  for I := 1 to Length(S) do
    begin
      if S[I] = ' ' then CharNum := 0 else CharNum := Ord(S[I])-32;
      W := CS.Characters[CharNum].Width;
      DrawFontCell(GlobalBitmap.Canvas,
                   Rect(X, H-CS.Characters[CharNum].Height, X+W, H),
                   CS.Characters[CharNum]);
      Inc(X, W);
    end;

  frmArtView := TfrmArtView.Create(Application);
  frmArtView.ShowBitmap(S, GlobalBitmap, 'Font Number:', lvFonts.Selected.Index, False);
end;

procedure TfrmMain.actFontColorExecute(Sender: TObject);
begin
  ShowHues(FontHue, FontHueEntry);
  dgFonts.Refresh;
end;

 //-----------------------------------------------------------------------------
 //  GUMPS
 //-----------------------------------------------------------------------------

{
type
  TGumpData = class (TObject)
    Number, Lookup, Length : Integer;
    Height : Word;
    Width : Word;
  end;
}

type
  TGumpData = class (TObject)
    Index : Integer;
    Handle : Integer;
  end;

procedure TfrmMain.ActivateGumps;
var
  I : Integer;
  Item : TListItem;
  GStart : Integer;
  Handle, Tag : Integer;
  Source : TStream;
  Panel : TStatusPanel;
  GumpData : TGumpData;
begin
  if not frmOptions.VerifyFile('GUMPIDX.MUL') then exit;
  if not frmOptions.VerifyFile('GUMPART.MUL') then exit;

  pnlGumps.BringToFront;
  ActiveSection := asGumps;
  ActiveFilesOK := False;
  FIdxMain := Nil;
  sbMain.SimpleText := MsgLoadingIndex;
  Screen.Cursor := crHourglass;
  Application.ProcessMessages;
  try
    FIdxMain := TIndexedFile.Create(frmOptions.GetPath('GUMPIDX.MUL'),
                                    frmOptions.GetPath('GUMPART.MUL'),
                                    FPatchFile, $0C, True);

    FIdxMain.GetRangeInfo(0, 2147483647, GStart, FGumpCount);

  finally
    Screen.Cursor := crDefault;
    sbMain.SimpleText := '';
    Application.ProcessMessages;
  end;
  ActiveFilesOK := True;

  lvGumps.Items.BeginUpdate;
  try
    For I := 0 to FGumpCount-1 do
      begin
        FIdxMain.GetEntryInfo(I, Handle, Tag, Source);
        Item := lvGumps.Items.Add;
        Item.Caption := Format('%.8x', [Handle]);
        if Source = FPatchFile then
          Item.Caption := Item.Caption + ' (p)';
        GumpData := TGumpData.Create();
        GumpData.Index := I;
        GumpData.Handle := Handle;
        Item.Data := GumpData;
      end;
  finally
    lvGumps.Items.EndUpdate;
  end;

   // Set up the status bar
  sbMain.SimplePanel := False;
  sbMain.Panels.Clear;

  Panel := sbMain.Panels.Add;
  Panel.Width := 60;
  Panel.Bevel := pbNone;
  Panel.Text := 'Handle.';

  Panel := sbMain.Panels.Add;
  Panel.Width := 150;
  Panel.Text := '';

  Panel := sbMain.Panels.Add;
  Panel.Width := 10;
  Panel.Bevel := pbNone;  
  Panel.Text := '';

  ActiveFilesOK := True;
end;

procedure TfrmMain.DeactivateGumps;
var
  I : Integer;
  GumpData : TGumpData;
  Item : TListItem;
begin
  ActiveFilesOK := False;
  sbMain.SimplePanel := True;
  lvGumps.Items.BeginUpdate;
  try
    for I := 0 to lvGumps.Items.Count-1 do
      begin
        Item := lvGumps.Items[I];
        GumpData := TGumpData(Item.Data);
        Item.Data := nil;
        GumpData.Free;
      end;
    lvGumps.Items.Clear;
  finally
    lvGumps.Items.EndUpdate;
  end;
  FIdxMain.Free;
  ActiveSection := asNone;
end;

procedure TfrmMain.pnlInGumpsResize(Sender: TObject);
begin
  imgGumps.Left := (pnlInGumps.ClientWidth - imgGumps.Width) div 2;
  imgGumps.Top := (pnlInGumps.ClientHeight - imgGumps.Height) div 2;
end;

procedure TfrmMain.lvGumpsDeletion(Sender: TObject; Item: TListItem);
begin
  Item.Data := Nil;
end;

procedure TfrmMain.lvGumpsChange(Sender: TObject; Item: TListItem;
  Change: TItemChange);
var
  Index : TIndexedData;
  W, H, L, T : Integer;
  GumpData : TGumpData;
begin
  if not ActiveFilesOK or (lvGumps.Selected = Nil) then Exit;

   // Load all the gump data into a memory stream before drawing it to the
   // bitmap (slight speed boost).
  GumpData := TGumpData(Item.Data);
  Index := FIdxMain[GumpData.Index];
  Index.Data.Seek(0, soFromBeginning);
  W := Index.Tag shr 16;
  H := Index.Tag and $FFFF;
  EstablishBitmap(W, H, clWindow);
  DrawGumpToBitmap(Index.Data, 0, Index.Data.Size, W, H, GlobalBitmap);

  L := (pnlInGumps.ClientWidth-W) div 2;
  T := (pnlInGumps.ClientHeight-H) div 2;

  imgGumps.Picture := Nil;
  imgGumps.SetBounds(L-1,T-1,W+2,H+2);
  with imgGumps.Canvas do
    begin
      Brush.Style := bsSolid;
      Brush.Color := clWindow;
      FillRect(Rect(1,1,W,H));
      Pen.Color := TColor(DecreaseColorBrightness(ColorToRGB(Brush.Color), 15));
      Pen.Mode  := pmCopy;
      Pen.Style := psSolid;
      Rectangle(0,0,W+2,H+2);
      Draw(1,1, GlobalBitmap);
    end;

  sbMain.Panels[1].Text := Format('%d (0x%x)', [GumpData.Handle, GumpData.Handle]);
end;

procedure TfrmMain.mnuGumpsSaveAsClick(Sender: TObject);
var
  Index : TIndexedData;
  W, H : Integer;
  GumpData : TGumpData;
begin
  if not ActiveFilesOK or (FGumpItem = Nil) then Exit;

  SavePicture.FileName := 'Gump.bmp';
  if SavePicture.Execute and TestFileOverwrite(SavePicture.FileName) then
    begin
      GumpData := TGumpData(FGumpItem.Data);
      Index := FIdxMain[GumpData.Index];
      Index.Data.Seek(0, soFromBeginning);
      W := Index.Tag shr 16;
      H := Index.Tag and $FFFF;
      EstablishBitmap(W, H, clWindow);
      DrawGumpToBitmap(Index.Data, 0, Index.Data.Size, W, H, GlobalBitmap);
      GlobalBitmap.SaveToFile(SavePicture.FileName);
    end;
end;

procedure TfrmMain.mnuGumpsDetailsClick(Sender: TObject);
begin
  raise Exception.Create(ErrorFeatureNotImplemented);
end;

procedure TfrmMain.lvGumpsContextPopup(Sender: TObject; MousePos: TPoint;
  var Handled: Boolean);
begin
  FGumpItem := lvGumps.GetItemAt(MousePos.X, MousePos.Y);
  if FGumpItem = Nil then
    begin
      Handled := True;
      Exit;
    end;
  lvGumps.Selected := FGumpItem;
end;

 //-----------------------------------------------------------------------------
 //  HUES
 //-----------------------------------------------------------------------------

procedure TfrmMain.ActivateHues;
var
  FileName : String;
  F : TFileStream;
  GroupCount : Integer;
  I : Integer;
begin
  if not frmOptions.VerifyFile('HUES.MUL') then exit;

  pnlHues.BringToFront;
  ActiveSection := asHues;
  ActiveFilesOK := False;
  sbMain.SimpleText := 'Loading hue information...';

  // Load
  HuesGrid.ColWidths[0] := 80;
  HuesGrid.ColWidths[1] := 100;

  FileName := frmOptions.GetPath('HUES.MUL');
  F := TFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
  try
    GroupCount := F.Size div SizeOf(THueGroup);
    SetLength(FRange, GroupCount);
    For I := 0 to GroupCount-1 do
      F.Read(FRange[I], SizeOf(THueGroup));
  finally
    F.Free;
  end;
  HuesGrid.RowCount := GroupCount * 8;

  sbMain.SimpleText := 'Hues loaded...';
  ActiveFilesOK := True;

  HuesGrid.Refresh;
  HuesGrid.Row := FStartingRow;
end;

procedure TfrmMain.DeactivateHues;
begin
  ActiveFilesOK := False;
  // Unload
  ActiveSection := asNone;
end;

procedure TfrmMain.pnlHuesResize(Sender: TObject);
var
  I, W : Integer;
begin
  W := (HuesGrid.ClientWidth - 180) div (HuesGrid.ColCount-2);
  for I := 2 to HuesGrid.ColCount-1 do
    HuesGrid.ColWidths[I] := W;
end;

procedure TfrmMain.HuesGridDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var
  S : String;
  Group : Integer;
  Entry : Integer;
begin
   // Draw any requested cell (Unless the data is not loaded)
  if (ActiveSection <> asHues) or not ActiveFilesOK then Exit;
  with HuesGrid.Canvas do
    begin
      Brush.Style := bsSolid;
      Brush.Color := clBtnFace;
      Group := ARow div 8;
      Entry := ARow mod 8;
      if ACol = 0 then with HuesGrid.Canvas do
        begin
          Pen.Style := psClear;
          if HuesGrid.Row = ARow then
            begin
              Brush.Color := clHighlight;
              Font.Color := clHighlightText;
            end
          else
            begin
              Brush.Color := clWindow;
              Font.Color := clWindowText;
            end;
          if ARow <= 0 then
            S := ''
          else
            S := Format('%d (0x%x)', [ARow-1, ARow-1]);
          FillRect(Rect);
          DrawText(Handle, PChar(S), Length(S), Rect, DT_CENTER or DT_VCENTER or DT_SINGLELINE);
        end
      else if ACol = 1 then with HuesGrid.Canvas do
        begin
          Pen.Style := psClear;
          if HuesGrid.Row = ARow then
            begin
              Brush.Color := clHighlight;
              Font.Color := clHighlightText;
            end
          else
            begin
              Brush.Color := clWindow;
              Font.Color := clWindowText;
            end;
          S := FRange[Group].Entries[Entry].Name;
          FillRect(Rect);
          DrawText(Handle, PChar(S), Length(S), Rect, DT_LEFT or DT_VCENTER or DT_SINGLELINE);
        end
      else
        begin
          with FRange[Group] do
            begin
              Brush.Color := Color15To32(Entries[Entry].ColorTable[ACol-2]);
              FillRect(Rect);
            end;
        end;

      if (HuesGrid.Row = ARow) and (FOldRow <> ARow) then
        begin
          Rect := HuesGrid.CellRect(0, ARow);
          InvalidateRect(HuesGrid.Handle, @Rect, True);
          Rect := HuesGrid.CellRect(0, FOldRow);
          InvalidateRect(HuesGrid.Handle, @Rect, True);
          FOldRow := ARow;
        end;
      Pen.Style := psSolid;
    end;
end;

procedure TfrmMain.HuesGridSelectCell(Sender: TObject; ACol, ARow: Integer;
  var CanSelect: Boolean);
begin
  FStartingRow := ARow;
end;

 //-----------------------------------------------------------------------------
 //  MAP
 //-----------------------------------------------------------------------------

procedure TfrmMain.ActivateMap;
var
  MapsNode : TXMLNode;
  MapNode : TXMLNode;
  MapName : String;
  I : Integer;
  DefaultIdx : Integer;
  MapFile : String;
  StaIdxFile : String;
  StaticsFile : String;
  MapDifLocFile : String;
  MapDifFile : String;
  StaDifLocFile : String;
  StaDifFile : String;
  StaticsDifFile : String;
  ViewName : String;
  ViewAreaStr : String;
  ViewArea : RMapViewArea;
begin
  pnlMap.BringToFront;
  ActiveSection := asMap;

  ActiveFilesOK := False;
  FIndex := Nil; FMain := Nil;
  pnlMap.BevelInner := bvLowered;

  DefaultIdx := 0;
  WhichMapIdx := -1;

  MapsNode := frmOptions.InsideUOData.GetNode('InsideUO/Maps');
  if MapsNode.Count <= 0 then
    raise Exception.Create(ErrorXMLMissingMap);

  // Load the names for the maps
  if not MapsLoadedOnce then
    begin
      Maps := TStringList.Create;
      try
        for I := 0 to MapsNode.Count-1 do
          begin
            MapNode := MapsNode[I];
            if (MapNode.TagName = 'MAP') then
              begin
                MapName := MapNode.GetAttributeValue('NAME');
                Maps.Add(MapName);
                if MapName = 'Default' then
                  DefaultIdx := I;
              end;
          end;
      except
        on E: Exception do
          begin
            Maps.Free;
            Maps := Nil;
            pnlSectionTitle.Caption := 'Map - Error';
            lblMapError.Caption := Format(ErrorMapLoadWrapper, [e.Message]);
            raise;
          end;
      end;
      MapsLoadedOnce := true;
    end;

  // Find which map we're currently on based on the WhichMap string
  WhichMapIdx := 0;
  while (WhichMapIdx < Maps.Count) and (Maps[WhichMapIdx] <> WhichMap) do
    Inc(WhichMapIdx);
  if WhichMapIdx >= Maps.Count then
    WhichMapIdx := DefaultIdx;
  WhichMap := Maps[WhichMapIdx];
  MapNode := MapsNode[WhichMapIdx];

  // Base mapping file names.
  MapFile := MapNode.GetAttributeValue('MAP_FILE');
  if MapFile = '' then
    raise Exception.CreateFmt(ErrorMapFileTagNotSet, [WhichMap]);
  StaIdxFile := MapNode.GetAttributeValue('STAIDX_FILE');
  if StaIdxFile = '' then
    raise Exception.CreateFmt(ErrorStaIdxFileTagNotSet, [WhichMap]);
  StaticsFile := MapNode.GetAttributeValue('STATICS_FILE');
  if StaticsFile = '' then
    raise Exception.CreateFmt(ErrorStaticsFileTagNotSet, [WhichMap]);

  // Optional diffing file names.
  MapDifLocFile := MapNode.GetAttributeValue('MAPDIFLOC_FILE');
  MapDifFile := MapNode.GetAttributeValue('MAPDIF_FILE');
  StaDifLocFile := MapNode.GetAttributeValue('STADIFLOC_FILE');
  StaDifFile := MapNode.GetAttributeValue('STADIF_FILE');
  StaticsDifFile := MapNode.GetAttributeValue('STATICSDIF_FILE');

  FUOMap.Free;
  FUOMap := TUOMap.Create(pnlMap);
  try
    with FUOMap do
      begin
        Parent := pnlMap;
        Align := alClient;

        // Handle map file
        MapPath := frmOptions.GetPath(MapFile);
        if(MapPath = '') then
          raise Exception.CreateFmt(ErrorMapFilePathNotSet, [MapFile]);
        if not FileExists(MapPath) then
          raise Exception.CreateFmt('Map file could not be found (MAP?.MUL): %s', [MapFile]);

        // Handle statics index file
        StaticsIndexPath := frmOptions.GetPath(StaIdxFile);
        if(StaticsIndexPath = '') then
          raise Exception.CreateFmt(ErrorStaIdxPathNotSet, [StaIdxFile]);
        if not FileExists(StaticsIndexPath) then
          raise Exception.CreateFmt('Statics index file could not be found (STAIDX?.MUL): %s', [StaticsIndexPath]);

        // Handle statics file
        StaticsPath := frmOptions.GetPath(StaticsFile);
        if(StaticsPath = '') then
          raise Exception.CreateFmt(ErrorStaticsPathNotSet, [StaticsFile]);
        if not FileExists(StaticsPath) then
          raise Exception.CreateFmt('Statics file could not be found (STATICS?.MUL): %s', [StaticsPath]);

        RadarPath := frmOptions.GetPath('RADARCOL.MUL');
        if(RadarPath = '') then Exit;

        // Optional map paths
        MapDifLocPath := frmOptions.GetPath(MapDifLocFile);
        MapDifPath := frmOptions.GetPath(MapDifFile);
        StaDifLocPath := frmOptions.GetPath(StaDifLocFile);
        StaDifPath := frmOptions.GetPath(StaDifFile);
        StaticsDifPath := frmOptions.GetPath(StaticsDifFile);

        OnMapClick := DoMapClick;

        ResetViews;
        for I := 0 to MapNode.Count-1 do
          if MapNode[I].TagName = 'AREA' then
            SetOverallArea(ParseArea(MapNode[I].Data))
          else if MapNode[I].TagName = 'VIEW' then
            begin
              ViewName := MapNode[I].GetAttributeValue('NAME');
              ViewAreaStr := MapNode[I].GetAttributeValue('AREA');
              ViewArea := ParseArea(ViewAreaStr);
              AddView(ViewName, ViewArea);
            end;
        ResetMap(0);

        pnlSectionTitle.Caption := 'Map - ' + WhichMap;
      end;
  except
    on E: Exception do
      begin
        FUOMap.Free;
        FUOMap := Nil;
        lblMapError.Caption := Format(ErrorMapLoadWrapper, [e.Message]);
        raise;
      end;
  end;
  pnlMap.BevelInner := bvNone;
  ActiveFilesOK := True;
end;

procedure TfrmMain.DeactivateMap;
begin
  ActiveFilesOK := False;
  FUOMap.Free;
  FUOMap := Nil;
  ActiveSection := asNone;
end;

procedure TfrmMain.BuildMapsMenu(AMenu : TMenu);
var
  I : Integer;
  MenuItem : TMenuItem;
  Prefix : String;
begin
  Prefix := AMenu.Name + '_Alt';
  I := 0;
  while I < AMenu.Items.Count do
    begin
      if(StrLeft(AMenu.Items[I].Name,Length(Prefix)) = Prefix) then
        begin
          MenuItem := AMenu.Items[I];
          AMenu.Items.Delete(I);
          MenuItem.Free;
        end
      else
        Inc(I);
    end;

  for I := 0 to Maps.Count-1 do
    begin
      MenuItem := TMenuItem.Create(Self);
      MenuItem.Name := Format('%s%d', [Prefix, I]);
      MenuItem.Caption := Maps[I];
      MenuItem.OnClick := mnuMapViewAlt;
      MenuItem.Tag := I;
      if WhichMapIdx = I then
        MenuItem.Checked := True;
      AMenu.Items.Add(MenuItem);
    end;
end;

procedure TfrmMain.pnlMapContextPopup(Sender: TObject; MousePos: TPoint;
  var Handled: Boolean);
begin
  BuildMapsMenu(pmMapsEmpty);
end;

procedure TfrmMain.DoMapClick(Sender: TObject; Button: TMouseButton;
                              Shift: TShiftState; X, Y, AX, AY : Integer);
begin
  if (Button = mbRight) and ActiveFilesOK then
    begin
      MapAX := AX;
      MapAY := AY;
      BuildMapsMenu(pmMap);
      pmMap.Popup(X+2, Y+2);
    end;
end;

procedure TfrmMain.ZoomClick(Sender: TObject);
begin
  if Sender = mnuMapZoom1x1 then
    FUOMap.Zoom := zm1x1
  else
    FUOMap.Zoom := zm1x8;

  TMenuItem(Sender).Checked := True;
end;

procedure TfrmMain.mnuMapNormalClick(Sender: TObject);
begin
  mnuMapNormal.Checked := True;
  FUOMap.AltSetting := asNormal;
end;

procedure TfrmMain.mnuMapNormalWithAltClick(Sender: TObject);
begin
  mnuMapNormalWithAlt.Checked := True;
  FUOMap.AltSetting := asNormalWithAlt;
end;

procedure TfrmMain.mnuMapAltitudeClick(Sender: TObject);
begin
  mnuMapAltitude.Checked := True;
  FUOMap.AltSetting := asAltitude;
end;

procedure TfrmMain.mnuMapDrawStaticsClick(Sender: TObject);
begin
  mnuMapDrawStatics.Checked := not mnuMapDrawStatics.Checked;
  if mnuMapDrawStatics.Checked then
    FUOMap.Flags := FUOMap.Flags + [flStatics]
  else
    FUOMap.Flags := FUOMap.Flags - [flStatics];
end;

procedure TfrmMain.mnuMapDrawPatchedClick(Sender: TObject);
begin
  mnuMapDrawPatched.Checked := not mnuMapDrawPatched.Checked;
  if mnuMapDrawPatched.Checked then
    FUOMap.Flags := FUOMap.Flags + [flPatch]
  else
    FUOMap.Flags := FUOMap.Flags - [flPatch];
end;

procedure TfrmMain.mnuMapXRayViewClick(Sender: TObject);
begin
  mnuMapXRayView.Checked := not mnuMapXRayView.Checked;
  if mnuMapXRayView.Checked then
    FUOMap.Flags := FUOMap.Flags + [flXRay]
  else
    FUOMap.Flags := FUOMap.Flags - [flXRay];
end;

procedure TfrmMain.mnuMapBlackWaterClick(Sender: TObject);
begin
  mnuMapBlackWater.Checked := not mnuMapBlackWater.Checked;
  if mnuMapBlackWater.Checked then
    FUOMap.Flags := FUOMap.Flags + [flNulWater]
  else
    FUOMap.Flags := FUOMap.Flags - [flNulWater];
end;

procedure TfrmMain.mnuMapViewAlt(Sender: TObject);
begin
  if Sender is TMenuItem then
    begin
      WhichMapIdx := TMenuItem(Sender).Tag;
      WhichMap := Maps[WhichMapIdx];
      ActivateMap;
    end;
end;

function TfrmMain.GetDescription(BlockNum : Integer; Default : String) : String;
var
  FS : TFileStream;
  Loc : Integer;
  Desc : Array[0..19] of Char;
begin
  FS := TFileStream.Create(frmOptions.GetPath('TILEDATA.MUL'), fmOpenRead or fmShareDenyNone);
  try
    if BlockNum < 16384 then
      begin
        Loc := ((BlockNum shr 5) + 1) * 4 + BlockNum * 26;
        FS.Seek(Loc + 6, soFromBeginning);
        FS.Read(Desc, SizeOf(Desc));
      end
    else if BlockNum < 32768 then
      begin
        BlockNum := BlockNum - 16384;
        Loc := $68800 + ((BlockNum shr 5) + 1) * 4 + BlockNum * 37;
        FS.Seek(Loc + 17, soFromBeginning);
        FS.Read(Desc, SizeOf(Desc));
      end
    else
      Result := '';

    Result := Trim(StrPas(Desc));
    if Result = '' then
      Result := Default;
  finally
    FS.Free;
  end;
end;

procedure TfrmMain.mnuMapDetailsClick(Sender: TObject);
var
  List : TList;
  Static : TStaticPosition;
  I, X, Y : Integer;
  Str : String;
  BlockId : Integer;
  Block : PBlock;
begin
  frmMapPointData.lblXDat.Caption := IntToStr(MapAX);
  frmMapPointData.lblYDat.Caption := IntToStr(MapAY);
  frmMapPointData.Memo.Clear;

  BlockId := FUOMap.GetBlockIdAt(MapAX, MapAY);
  frmMapPointData.Memo.Lines.Add(Format('Block Id: %d', [BlockId]));

  Block := FUOMap.GetBlockAt(BlockId);
  if Block <> Nil then
    try
      // Add block
      frmMapPointData.Memo.Lines.Add('');
      frmMapPointData.Memo.Lines.Add('Base Data');
      frmMapPointData.Memo.Lines.Add(Format('Header: 0x%x', [Block^.Header]));
      for Y := 0 to 7 do
        begin
          Str := '';
          for X := 0 to 7 do
            begin
              I := Y * 8 + X;
              Str := Str + Format(' %d (%d)', [Block^.Cells[I].Value, Block^.Cells[I].Altitude]);
              if ((MapAX and $07) = X) and ((MapAY and $07) = Y) then
                Str := Str + '*';
            end;
          frmMapPointData.Memo.Lines.Add(Str);
        end;

      // Add info about specific base tile
      frmMapPointData.Memo.Lines.Add('');
      with Block.Cells[(MapAY and $07) * 8 + (MapAX and $07)] do
        begin
          Str := Format('Base Tile = %s (%d), Alt = %d',[GetDescription(Value, UnlabeledArtBlock), Value, Altitude]);
          frmMapPointData.Memo.Lines.Add(Str);
        end;
    finally
      Dispose(Block);
    end;

  Block := FUOMap.GetPatchedBlockAt(BlockId);
  if Block <> Nil then
    try
      // Add block
      frmMapPointData.Memo.Lines.Add('');
      frmMapPointData.Memo.Lines.Add('Patched Base Data');
      frmMapPointData.Memo.Lines.Add(Format('Header: 0x%x', [Block^.Header]));
      for Y := 0 to 7 do
        begin
          Str := '';
          for X := 0 to 7 do
            begin
              I := Y * 8 + X;
              Str := Str + Format(' %d (%d)', [Block^.Cells[I].Value, Block^.Cells[I].Altitude]);
              if ((MapAX and $07) = X) and ((MapAY and $07) = Y) then
                Str := Str + '*';
            end;
          frmMapPointData.Memo.Lines.Add(Str);
        end;

      // Add info about specific base tile
      frmMapPointData.Memo.Lines.Add('');
      with Block.Cells[(MapAY and $07) * 8 + (MapAX and $07)] do
        begin
          Str := Format('Base Tile = %s (%d), Alt = %d',[GetDescription(Value, UnlabeledArtBlock), Value, Altitude]);
          frmMapPointData.Memo.Lines.Add(Str);
        end;
    finally
      Dispose(Block);
    end;

  // Add statics
  List := TList.Create;
  try
    FUOMap.GetStaticsAt(MapAX, MapAY, List);
    frmMapPointData.Memo.Lines.Add('');
    frmMapPointData.Memo.Lines.Add('Statics');
    For I := 0 to List.Count-1 do
      begin
        Static := TStaticPosition(List[I]^);
        Str := Format('Static Object = %s (%d), Alt = %d',[GetDescription(16384 + Static.BlockNum, UnlabeledArtBlock), Static.BlockNum, Static.Altitude]);
        frmMapPointData.Memo.Lines.Add(Str);
      end;
  finally
    For I := 0 to List.Count-1 do
      Dispose(List[I]);
    List.Free;
  end;

  // Add patched statics
  List := TList.Create;
  try
    FUOMap.GetPatchedStaticsAt(MapAX, MapAY, List);
    frmMapPointData.Memo.Lines.Add('');
    frmMapPointData.Memo.Lines.Add('Patched Statics');
    For I := 0 to List.Count-1 do
      begin
        Static := TStaticPosition(List[I]^);
        Str := Format('Static Object = %s (%d), Alt = %d',[GetDescription(16384 + Static.BlockNum, UnlabeledArtBlock), Static.BlockNum, Static.Altitude]);
        frmMapPointData.Memo.Lines.Add(Str);
      end;
  finally
    For I := 0 to List.Count-1 do
      Dispose(List[I]);
    List.Free;
  end;

  frmMapPointData.ShowModal;
end;

procedure TfrmMain.mnuMapSaveClick(Sender: TObject);
begin
  frmMapSave.Execute(FUOMap);
end;

 //-----------------------------------------------------------------------------
 //  MULTI
 //-----------------------------------------------------------------------------

type
  TMultiLookup = class
  public
    FileIndex : Integer;
    Index : Integer;
  end;


procedure TfrmMain.ActivateMulti;
var
  I, J : Integer;
  Item : TListItem;
  Handle, Tag : Integer;
  Source : TStream;
  MultisNode : TXMLNode;
  MultiSetNode : TXMLNode;
  MultiNode : TXMLNode;
  MultiFileName : String;
  MultiIdxFileName : String;
  MultiIdx, Index : Integer;
  Lookup : TMultiLookup;
  FileIndex : Integer;
begin
  if not frmOptions.VerifyFile('ARTIDX.MUL') then exit;
  if not frmOptions.VerifyFile('ART.MUL') then exit;

  pnlMulti.BringToFront;
  ActiveSection := asMulti;
  ActiveFilesOK := False;

  ShowFloors := tbMulti.Position;
  FIndex := Nil; FMain := Nil;

  sbMain.SimpleText := 'Loading multi index...';

  Assert(MultiFiles = Nil);
  MultiFiles := TIndexFileList.Create;

  Screen.Cursor := crHourglass;
  Application.ProcessMessages;
  try
    MultisNode := frmOptions.InsideUOData.GetNode('InsideUO/Multis');
    try
      for I := 0 to MultisNode.Count-1 do
        begin
          MultiSetNode := MultisNode[I];
          if MultiSetNode.TagName = 'MULTISET' then
            begin
              MultiFileName := MultiSetNode.GetAttributeValue('MULTI_FILE');
              MultiIdxFileName := MultiSetNode.GetAttributeValue('MULTIIDX_FILE');

              FileIndex := MultiFiles.Find(MultiFileName, MultiIdxFileName);
              if FileIndex < 0 then
                begin
                  if FileExists( frmOptions.GetPath(MultiFileName) ) and
                     FileExists( frmOptions.GetPath(MultiIdxFileName) ) then
                    FileIndex := MultiFiles.Add(
                      MultiFileName, MultiIdxFileName,
                      frmOptions.GetPath(MultiFileName),
                      frmOptions.GetPath(MultiIdxFileName),
                      FPatchFile, $0E);
                end;

              // TODO: Store FileIndex so that we can use it later.

              lvMulti.Items.BeginUpdate;
              try
                // Find the first <MULTI> tag and set Index equal to it.
                Index := -1;
                MultiIdx := 0;
                while MultiIdx < MultiSetNode.Count do
                  begin
                    MultiNode := MultiSetNode[MultiIdx];
                    Inc(MultiIdx);
                    if MultiNode.TagName = 'MULTI' then
                      begin
                        Index := StrToInt(MultiNode.GetAttributeValue('INDEX'));
                        break;
                      end;
                  end;

                // Run through each of the multi entries in the index.
                if FileIndex >= 0 then
                  for J := 0 to MultiFiles.Get(FileIndex).Count-1 do
                    begin
                      MultiFiles.Get(FileIndex).GetEntryInfo(J, Handle, Tag, Source);

                      // Find the next matching entry in the tags (if there is one)
                      MultiNode := Nil;
                      while Index < Handle do
                        begin
                          Index := -1;
                          while (MultiIdx < MultiSetNode.Count) do
                            begin
                              MultiNode := MultiSetNode[MultiIdx];
                              Inc(MultiIdx);
                              if MultiNode.TagName = 'MULTI' then
                                begin
                                  Index := StrToInt(MultiNode.GetAttributeValue('INDEX'));
                                  break;
                                end;
                            end;
                          if Index = -1 then
                            break;
                        end;

                      Lookup := TMultiLookup.Create;
                      Lookup.FileIndex := FileIndex;
                      Lookup.Index := J;

                      // Add a new item to the list.
                      Item := lvMulti.Items.Add;
                      Item.Data := Pointer(Lookup);
                      if (Index = Handle) and
                         (MultiNode <> Nil) and
                         (MultiNode.Data <> '') then
                        begin
                          Item.Caption := MultiNode.Data;
                        end
                      else // Couldn't find an entry.  Show the number.
                        begin
                          Item.Caption := Format('%.8x', [Handle]);
                        end;
                      Item.SubItems.Add(Format('%.8x', [Handle]));
                      if Source = FPatchFile then
                        Item.Caption := Item.Caption + ' (p)';
                    end;
              finally
                lvMulti.Items.EndUpdate;
              end;
            end;
        end;
    except
      on E: Exception do
        raise Exception.CreateFmt(ErrorXMLWrapper, [e.Message]);
    end;
  finally
    Screen.Cursor := crDefault;
    sbMain.SimpleText := '';
    Application.ProcessMessages;
  end;

  ArtIndex := TFileStream.Create(frmOptions.GetPath('ARTIDX.MUL'), fmOpenRead or fmShareDenyNone);
  ArtMain := TFileStream.Create(frmOptions.GetPath('ART.MUL'), fmOpenRead or fmShareDenyNone);
  ArtCacheIndex := TArtCacheIndex.Create(ArtIndex, ArtMain);

  ActiveFilesOK := True;
end;

procedure TfrmMain.DeactivateMulti;
var
  I : Integer;
begin
  ActiveFilesOK := False;

  MultiFiles.Free;
  MultiFiles := Nil;

  lvMulti.Items.BeginUpdate;
  try
    for I := 0 to lvMulti.Items.Count-1 do
      TMultiLookup(lvMulti.Items[I].Data).Free;
    lvMulti.Items.Clear;
  finally
    lvMulti.Items.EndUpdate;
  end;

  ArtCacheIndex.Free;
  ArtMain.Free;
  ArtIndex.Free;

  ActiveSection := asNone;
end;

procedure TfrmMain.DrawMulti(Item : TListItem);
var
  Data : TStream;
  R : TRect;
  Lookup : TMultiLookup;
begin
  if not ActiveFilesOK or (Item = Nil) then Exit;

  // Get the multi as a bitmap
  Lookup := TMultiLookup(Item.Data);
  Data := MultiFiles.Get(Lookup.FileIndex)[Lookup.Index].Data;
  Assert(Data <> Nil);
  Data.Seek(0, soFromBeginning);
  DrawMultiToBitmap(Data, ArtCacheIndex, ShowFloors, GlobalBitmap);

  // Draw it onto imgMulti
  imgMulti.Picture := Nil;
  imgMulti.SetBounds(
      Max((sbMulti.Width-GlobalBitmap.Width-2) div 2,0) + 1,
      Max((sbMulti.Height-GlobalBitmap.Height-2) div 2,0) + 1,
      GlobalBitmap.Width + 2, GlobalBitmap.Height + 2);
  R := Rect(0,0,imgMulti.Width,imgMulti.Height);
  with imgMulti.Canvas do
    begin
      Pen.Color := TColor(DecreaseColorBrightness(ColorToRGB(clWindow), 15));
      Rectangle(R);
      Draw(1,1,GlobalBitmap);
    end;
end;

procedure TfrmMain.sbMultiResize(Sender: TObject);
begin
  imgMulti.Left := Max((sbMulti.Width-imgMulti.Width-2) div 2,0) - sbMulti.HorzScrollBar.Position;
  imgMulti.Top := Max((sbMulti.Height-imgMulti.Height-2) div 2,0) - sbMulti.VertScrollBar.Position;
end;

procedure TfrmMain.lvMultiChange(Sender: TObject; Item: TListItem;
  Change: TItemChange);
begin
  if Item.Data <> Nil then
    DrawMulti(Item);
end;

procedure TfrmMain.tbMultiChange(Sender: TObject);
begin
  if ShowFloors = tbMulti.Position then Exit;

  ShowFloors := tbMulti.Position;
  if (lvMulti.Selected <> Nil) and (lvMulti.Selected.Data <> Nil) then
    DrawMulti(lvMulti.Selected);
end;

procedure TfrmMain.mnuMultiSaveAsClick(Sender: TObject);
var
  Data : TStream;
  Lookup : TMultiLookup;
begin
  if not ActiveFilesOK or (lvMulti.Selected = Nil) then Exit;

  SavePicture.FileName := 'Multi.bmp';
  if SavePicture.Execute and TestFileOverwrite(SavePicture.FileName) then
    begin
      // Get the multi as a bitmap
      Lookup := TMultiLookup(lvMulti.Selected.Data);
      Data := MultiFiles.Get(Lookup.FileIndex)[Lookup.Index].Data;
      Assert(Data <> Nil);
      Data.Seek(0, soFromBeginning);
      DrawMultiToBitmap(Data, ArtCacheIndex, ShowFloors, GlobalBitmap);
      GlobalBitmap.SaveToFile(SavePicture.FileName);
    end;
end;

procedure TfrmMain.MultiMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  P : TPoint;
begin
  if (Button = mbRight) and ActiveFilesOK and (lvMulti.Selected <> Nil) then
    begin
      P := (Sender as TControl).ClientToScreen(Point(X + 2, Y + 2));
      pmMulti.Popup(P.X, P.Y);
    end;
end;

procedure TfrmMain.mnuMultiDetailsClick(Sender: TObject);
var
  Data : TIndexedData;
  frmDetails : TfrmDetails;
  S : Integer;
  Lookup : TMultiLookup;
begin
  if not ActiveFilesOK or (lvMulti.Selected = Nil) then Exit;

  // Pull up details on the selected item
  Lookup := TMultiLookup(lvMulti.Selected.Data);
  Data := MultiFiles.Get(Lookup.FileIndex)[Lookup.Index];

  Assert(Data <> Nil);

  Application.CreateForm(TfrmDetails, frmDetails);
  with frmDetails, Data do
    begin
      Caption := 'Details for ' + lvMulti.Selected.Caption;
      Memo.Clear;
      S := Memo.SelStart;
      Memo.Lines.Add(Format('Number: %d (0x%x)', [Handle, Handle]));
      BoldIt(Memo, S, 7); S := Memo.SelStart;
      Memo.Lines.Add(Format('Tag: 0x%x', [Tag]));
      BoldIt(Memo, S, 7); S := Memo.SelStart;
      Memo.Lines.Add(Format('Length: %d', [Data.Size]));
      BoldIt(Memo, S, 7); {S :=} Memo.SelStart;
      Memo.SelStart := 0;
      Show;
    end;
end;

 //-----------------------------------------------------------------------------
 //  SKILLS
 //-----------------------------------------------------------------------------

procedure TfrmMain.ActivateSkills;
var
  SkillsStart : Integer;
  SkillsCount : Integer;
  Source : TStream;
  I : Integer;
  Handle, Tag : Integer;
  Data : TStream;
  Skill : TSkill;
  Item : TListItem;
begin
  if not frmOptions.VerifyFile('SKILLS.IDX') then exit;
  if not frmOptions.VerifyFile('SKILLS.MUL') then exit;

  pnlSkills.BringToFront;
  ActiveSection := asSkills;
  ActiveFilesOK := False;
  FIdxMain := Nil;
  sbMain.SimpleText := 'Loading skills index...';
  Screen.Cursor := crHourglass;
  Application.ProcessMessages;
  try
    FIdxMain := TIndexedFile.Create(frmOptions.GetPath('SKILLS.IDX'),
                                    frmOptions.GetPath('SKILLS.MUL'),
                                    FPatchFile, $10, True);

    FIdxMain.GetRangeInfo(0, 2147483647, SkillsStart, SkillsCount);
  finally
    Screen.Cursor := crDefault;
    sbMain.SimpleText := '';
    Application.ProcessMessages;
  end;
  ActiveFilesOK := True;

  lvSkills.Items.Clear;
  lvSkills.Items.BeginUpdate;
  try
    For I := 0 to SkillsCount-1 do
      begin
        FIdxMain.GetEntryInfo(I, Handle, Tag, Source);
        Data := FIdxMain[I].Data;
        Assert(Data <> Nil);
        Data.Seek(0, soFromBeginning);
        FillChar(Skill,SizeOf(Skill),0);
        Data.Read(Skill, Data.Size);
        Item := lvSkills.Items.Add;
        Item.Caption := Skill.Name;
        if Skill.Button > 0 then
          Item.ImageIndex := 0
        else
          Item.ImageIndex := -1;
      end;
    lvSkills.SortType := stText; 
  finally
    lvSkills.Items.EndUpdate;
  end;
  lvSkills.Width := 10;
end;

procedure TfrmMain.DeactivateSkills;
begin
  ActiveFilesOK := False;
  ActiveSection := asNone;
end;

 //-----------------------------------------------------------------------------
 //  SOUND
 //-----------------------------------------------------------------------------

type
  TSoundData = class (TObject)
    FileName : String;
    SoundIndexRec : TSoundIndexRec;
  end;

procedure BreakUp(S : String; var First, Second : String);
var
  I, L : Integer;
begin
  L := Length(S); I := 1;
  While (I < L) and not (S[I] in ['0'..'9', '.']) do
    Inc(I);

  If I < L then
    begin
      First := Copy(S, 1, I-1);
      Second := Copy(S,I,L-I+1);
    end
  else
    begin
      First := S;
      Second := '';
    end;
end;

procedure TfrmMain.ActivateSound;
var
  IndexRec : TSoundIndexRec;
  SoundHeader : TSoundHeader;
  Node, SubNode : TTreeNode;
  SndIndex : TMemoryStream;
  SndData : TSoundData;
  HashTable : TListHashTable;
  S1, S2 : String;
  HNode : THashNode;
  SL : TStringList;
  I, J : Integer;
  List : TList;
begin
  if not frmOptions.VerifyFile('SOUNDIDX.MUL') then exit;
  if not frmOptions.VerifyFile('SOUND.MUL') then exit;

  pnlSound.BringToFront;
  ActiveSection := asSound;

  ActiveFilesOK := False;

   // Populate the sound list
  sbMain.SimpleText := 'Loading sound list...';
  Screen.Cursor := crHourglass;
  Application.ProcessMessages;
  try
    HashTable := TListHashTable.Create;
    try
      FIndex := Nil; FMain := Nil;
      FIndex := TFileStream.Create(frmOptions.GetPath('SOUNDIDX.MUL'), fmOpenRead or fmShareDenyNone);
      try
        SndIndex := TMemoryStream.Create;
        try
          SndIndex.CopyFrom(FIndex, 0);
        except
          SndIndex.Free;
          raise;
        end;
      finally
        FIndex.Free;
        FIndex := Nil;
      end;

       // Try for the SndIndex
      try
        FMain := TFileStream.Create(frmOptions.GetPath('SOUND.MUL'), fmOpenRead or fmShareDenyNone);

        ActiveFilesOK := True;

         // Load up the hash table with the index
        SndIndex.Seek(0, soFromBeginning);
        while SndIndex.Position < SndIndex.Size do
          begin
            SndIndex.Read(IndexRec, SizeOf(IndexRec));
            if IndexRec.Lookup <> -1 then
              begin
                SndData := TSoundData.Create;
                Move(IndexRec, SndData.SoundIndexRec, SizeOf(IndexRec));
                FMain.Seek(IndexRec.Lookup, soFromBeginning);
                FMain.Read(SoundHeader, SizeOf(SoundHeader));
                SndData.FileName := StrPas(SoundHeader.Name);

                BreakUp(SndData.FileName, S1, S2);
                HNode := HashTable.LookupOrAppend(S1);
                HNode.Data := SndData;
              end;
          end;
      finally
        SndIndex.Free;
      end;

       // Load up the tree view with the hash table
      tvSound.Items.BeginUpdate;
      try
        SL := TStringList.Create;
        try
          HashTable.CopyToStringList(SL);
          For I := 0 to SL.Count-1 do
            begin
              List := TList(SL.Objects[I]);
              if List.Count = 1 then
                begin
                  SubNode := tvSound.Items.AddObject(Nil, TSoundData(THashNode(List[0]).Data).FileName, THashNode(List[0]).Data);
                  THashNode(List[0]).Data := Nil;
                  SubNode.ImageIndex := 1;
                  SubNode.SelectedIndex := 1;
                end
              else
                begin
                  S1 := LowerCase(SL[I]);
                  S1[1] := UpCase(S1[1]);
                  Node := tvSound.Items.AddObject(Nil, S1, Nil);
                  Node.ImageIndex := 0;
                  Node.SelectedIndex := 0;
                  For J := 0 to List.Count-1 do
                    begin
                      SubNode := tvSound.Items.AddChildObject(Node, TSoundData(THashNode(List[J]).Data).FileName, THashNode(List[J]).Data);
                      THashNode(List[J]).Data := Nil;
                      SubNode.ImageIndex := 1;
                      SubNode.SelectedIndex := 1;
                    end;
                end;
            end;
        finally
          SL.Free;
        end;
      finally
        tvSound.Items.EndUpdate;
      end;

    finally
      HashTable.Free;
    end;
  finally
    Screen.Cursor := crDefault;
    sbMain.SimpleText := '';
    Application.ProcessMessages;
  end;
end;

procedure TfrmMain.DeactivateSound;
begin
  tvSound.Items.BeginUpdate;
  try
    tvSound.Items.Clear;
  finally
    tvSound.Items.EndUpdate;
  end;
  FMain.Free;
  ActiveFilesOK := False;
  ActiveSection := asNone;
end;

procedure TfrmMain.tvSoundContextPopup(Sender: TObject; MousePos: TPoint;
  var Handled: Boolean);
begin
  if not ActiveFilesOK then
    begin
      Handled := True;
      Exit;
    end;

  FActiveSoundNode := tvSound.GetNodeAt(MousePos.X, MousePos.Y);
  if FActiveSoundNode <> Nil then
    begin
      tvSound.Selected := FActiveSoundNode;
      if FActiveSoundNode.Data = Nil then
        Handled := True;
    end
  else
    Handled := True;
end;
              
procedure TfrmMain.PlayRequestedSound(Item : TTreeNode);
var
  Data : TSoundData;
  Buffer : Array[0..1024] of Char;
  TempPath : String;
begin
  // Play sound from FActiveSoundItem
  if not ActiveFilesOK or (Item.Data = Nil) then Exit;
  Data := TSoundData(Item.Data);
  if Data.SoundIndexRec.Lookup <> -1 then
    begin
      FMain.Seek(Data.SoundIndexRec.Lookup, soFromBeginning);

      if GetTempPath(SizeOf(Buffer), Buffer) = 0 then
        RaiseLastWin32Error;
      TempPath := String(Buffer);

      if GetTempFileName(PChar(TempPath), 'IUO', 1, Buffer) = 0 then
        RaiseLastWin32Error;

      SaveSound(FMain, Data.SoundIndexRec.Length, Buffer);
      PlaySound(Buffer, 0, SND_FILEName);
      DeleteFile(Buffer);
    end;
end;

procedure TfrmMain.mnuSoundPlayClick(Sender: TObject);
begin
  if not ActiveFilesOK then Exit;
  PlayRequestedSound(FActiveSoundNode);
end;

procedure TfrmMain.mnuSoundSaveAsClick(Sender: TObject);
var
  Data : TSoundData;
begin
  // Save sound at FActiveSoundItem
  if not ActiveFilesOK then Exit;
  sdSound.FileName := FActiveSoundNode.Text;
  if sdSound.Execute and TestFileOverwrite(sdSound.FileName) then
    begin
      Data := TSoundData(FActiveSoundNode.Data);
      if Data.SoundIndexRec.Lookup <> -1 then
        begin
          FMain.Seek(Data.SoundIndexRec.Lookup, soFromBeginning);
          SaveSound(FMain, Data.SoundIndexRec.Length, sdSound.FileName);
        end;
    end;
end;

procedure TfrmMain.mnuSoundDetailsClick(Sender: TObject);
var
  Data : TSoundData;
  frmDetails : TfrmDetails;
  S : Integer;
begin
  if not ActiveFilesOK then Exit;

  // Pull up details on FActiveSoundItem
  Data := TSoundData(FActiveSoundNode.Data);

  Application.CreateForm(TfrmDetails, frmDetails);
  with frmDetails, Data.SoundIndexRec do
    begin
      Caption := 'Details for ' + FActiveSoundNode.Text;
      Memo.Clear;
      Memo.SelAttributes.Style := [];
      Memo.SelAttributes.Color := clWindowText;

      S := Memo.SelStart;
      Memo.Lines.Add(Format('Number: %d (0x%x)', [Index, Index]));
      BoldIt(Memo, S, 7); S := Memo.SelStart;
      Memo.Lines.Add(Format('Lookup: %x', [Lookup]));
      BoldIt(Memo, S, 7); S := Memo.SelStart;
      Memo.Lines.Add(Format('Length: %d', [Length]));
      BoldIt(Memo, S, 7); S := Memo.SelStart;
      Memo.Lines.Add(Format('Unknown: %d', [Unknown]));
      BoldIt(Memo, S, 8);
      Memo.SelStart := 0;
      Show;
    end;
end;

procedure TfrmMain.tvSoundDblClick(Sender: TObject);
begin
  if (tvSound.Selected <> Nil) and ActiveFilesOK then
    PlayRequestedSound(tvSound.Selected);
end;

procedure TfrmMain.tvSoundDeletion(Sender: TObject; Node: TTreeNode);
begin
  TSoundData(Node.Data).Free;
end;

procedure TfrmMain.mnuSoundCopyClick(Sender: TObject);
var
  Data : TSoundData;
begin
  // Save sound at FActiveSoundItem
  if not ActiveFilesOK then Exit;

  Data := TSoundData(FActiveSoundNode.Data);
  if Data.SoundIndexRec.Lookup <> -1 then
    begin
       FMain.Seek(Data.SoundIndexRec.Lookup, soFromBeginning);
       CopySound(FMain, Data.SoundIndexRec.Length);
    end;
end;

 //-----------------------------------------------------------------------------
 //  TEXTURE MAPS
 //-----------------------------------------------------------------------------

type
  TTexMapsIndexBuffer = class(TObject)
    BlockNum : Integer;
    Lookup : Integer;
    Length : Integer;
    Unknown : Integer;
  end;

procedure TfrmMain.ActivateTexMaps;
begin
  if not frmOptions.VerifyFile('TEXIDX.MUL') then exit;
  if not frmOptions.VerifyFile('TEXMAPS.MUL') then exit;

  pnlTexMaps.BringToFront;
  ActiveSection := asTexMaps;

  ActiveFilesOK := False;

   // Load startup data for texture maps
  sbMain.SimpleText := 'Loading texture maps...';
  Screen.Cursor := crHourglass;
  Application.ProcessMessages;
  try
    FIdxMain := TIndexedFile.Create(frmOptions.GetPath('TEXIDX.MUL'),
                                    frmOptions.GetPath('TEXMAPS.MUL'),
                                    FPatchFile, $0A, True);
    ActiveFilesOK := True;
  finally
    Screen.Cursor := crDefault;
    sbMain.SimpleText := '';
    Application.ProcessMessages;
  end;

  pnlTexMapsResize(Self);
end;

procedure TfrmMain.DeactivateTexMaps;
begin
  FIdxMain.Free;
  ActiveFilesOK := False;
  ActiveSection := asNone;
end;

procedure TfrmMain.pnlTexMapsResize(Sender: TObject);
var
  W : Integer;
begin
  if ActiveSection <> asTexMaps then Exit;

  W := dgTexMaps.ClientWidth - 16;
  dgTexMaps.ColCount := W div dgTexMaps.DefaultColWidth;

  if not ActiveFilesOK then
    dgTexMaps.RowCount := 0
  else
    dgTexMaps.RowCount := FIdxMain.Count div dgTexMaps.ColCount + 1;

  dgTexMaps.Refresh;
end;

function TfrmMain.BuildTexBitmap(Index : Integer) : TBitmap;
var
  W, H : Word;
  Row : PWordArray;
  I : Integer;
  IdxData : TIndexedData;
begin
  Result := TBitmap.Create;
  try
    if Index >= FIdxMain.Count then raise Exception.Create('Index out of bounds');
    IdxData := FIdxMain[Index];
    IdxData.Data.Seek(0, soFromBeginning);
    case IdxData.Data.Size of
      $2000 : begin W := 64; H := 64; end;
      $8000 : begin W := 128; H := 128; end;
    else
      raise Exception.CreateFmt('Unknown block size (%d)', [IdxData.Data.Size]);
    end;

    Result.Height := H;
    Result.Width := W;
    Result.PixelFormat := pf15bit;
    Result.HandleType := bmDIB;

    For I := 0 to H-1 do
      begin
        Row := Result.ScanLine[I];
        if IdxData.Data.Position < IdxData.Data.Size then
          IdxData.Data.Read(Row^, W * 2);
      end;
  except
    Result.Free;
    raise;
  end;
end;

procedure TfrmMain.dgTexMapsDblClick(Sender: TObject);
begin
  mnuTexMapsViewClick(Self);
end;

procedure TfrmMain.dgTexMapsDrawCell(Sender: TObject; Col, Row: Integer;
  Rect: TRect; State: TGridDrawState);
var
  Index : Integer;
  Bitmap : TBitmap;
begin
  if not ActiveFilesOK then Exit;
  Index := Row * dgTexMaps.ColCount + Col;
  if Index >= FIdxMain.Count then Exit;
  Bitmap := BuildTexBitmap(Index);
  try
    with dgTexMaps.Canvas do
      begin
        Bitmap.Width := 64;
        Bitmap.Height := 64;
        Draw(Rect.Left, Rect.Top, Bitmap);
      end;
  finally
    Bitmap.Free;
  end;
end;

procedure TfrmMain.dgTexMapsMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  P : TPoint;
  ACol, ARow : Integer;
begin
  if (Button = mbRight) and ActiveFilesOK then
    begin
      dgTexMaps.MouseToCell(X, Y, ACol, ARow);
      dgTexMaps.Row := ARow;
      dgTexMaps.Col := ACol;

      FActiveBNum := dgTexMaps.Row * dgTexMaps.ColCount + dgTexMaps.Col;
      if FActiveBNum < FIdxMain.Count then
        begin
          P := dgTexMaps.ClientToScreen(Point(X + 2, Y + 2));
          pmTexMaps.Popup(P.X, P.Y);
        end;
    end;
end;

procedure TfrmMain.mnuTexMapsViewClick(Sender: TObject);
var
  frmArtView: TfrmArtView;
  Bitmap : TBitmap;
  H : INteger;
begin
  if not ActiveFilesOK then Exit;
  FActiveBNum := dgTexMaps.Row * dgTexMaps.ColCount + dgTexMaps.Col;
  Bitmap := BuildTexBitmap(FActiveBNum);
  try
    H := FIdxMain[FActiveBNum].Handle;
    frmArtView := TfrmArtView.Create(Application);
    frmArtView.ShowBitmap('Texture Map', Bitmap, 'Index:', H, True);
  finally
    BitMap.Free;
  end;
end;

procedure TfrmMain.mnuTexMapsCopyClick(Sender: TObject);
begin
  if not ActiveFilesOK or (FActiveBNum >= FIdxMain.Count) then Exit;
  Clipboard.Assign(BuildTexBitmap(FActiveBNum));
end;

procedure TfrmMain.mnuTexMapsSaveAsClick(Sender: TObject);
begin
  if not ActiveFilesOK or (FActiveBNum >= FIdxMain.Count) then Exit;
  SavePicture.FileName := 'TexMap.bmp';
  if SavePicture.Execute and TestFileOverwrite(SavePicture.FileName) then
    BuildTexBitmap(FActiveBNum).SaveToFile(SavePicture.FileName);
end;

procedure TfrmMain.mnuTexMapsDetailsClick(Sender: TObject);
var
  frmDetails : TfrmDetails;
  S : Integer;
  IdxData : TIndexedData;
  W, H : Integer;
begin
  FActiveBNum := dgTexMaps.Row * dgTexMaps.ColCount + dgTexMaps.Col;
  if not ActiveFilesOK or (FActiveBNum >= FIdxMain.Count) then Exit;

  IdxData := FIdxMain[FActiveBNum];

  // Pull up details on the selected item
  Application.CreateForm(TfrmDetails, frmDetails);
  with frmDetails, IdxData do
    begin
      case Data.Size of
        $2000 : begin W := 64; H := 64; end;
        $8000 : begin W := 128; H := 128; end;
      else
        begin W := 0; H := 0; end;
      end;

      Caption := 'Details for Texture Map';
      Memo.Clear;
      S := Memo.SelStart;
      Memo.Lines.Add(Format('Number: %d (0x%x)', [Handle, Handle]));
      BoldIt(Memo, S, 7); S := Memo.SelStart;
//      Memo.Lines.Add(Format('Lookup: 0x%x', [Lookup]));
//      BoldIt(Memo, S, 7); S := Memo.SelStart;
      Memo.Lines.Add(Format('Length: %d', [Data.Size]));
      BoldIt(Memo, S, 7); S := Memo.SelStart;
      Memo.Lines.Add(Format('Tag: 0x%x, %d, %d, %d, %d', [Tag, Tag and $FF, (Tag shr 8) and $FF, (Tag shr 16) and $FF, (Tag shr 24) and $FF]));
      BoldIt(Memo, S, 8); S := Memo.SelStart;
      Memo.Lines.Add(Format('Width: %d', [W]));
      BoldIt(Memo, S, 6); S := Memo.SelStart;
      Memo.Lines.Add(Format('Height: %d', [H]));
      BoldIt(Memo, S, 7);
      Memo.SelStart := 0;
      Show;
    end;
end;

end.
