unit UOSound;

interface

uses
  Classes;

type
 {$A-}
  TSoundIndexRec = record
    Lookup : Integer;
    Length : Integer;
    Index : Word;
    Unknown : Word;
  end;

  TSoundHeader = record
    Name : Array[0..15] of Char;
    Unk1,
    Unk2,
    Unk3,
    Unk4 : Integer;
  end;

  TWaveChunk = Record
    WaveStr : Array[0..7] of Char;
    FmtLength : Integer;
    Tag : Word;
    Channels : Word;
    Rate : Integer;    // Samples per second
    BPS : Integer;     // Average Bytes per second
    Align : Word;      // Block size of data
    Bits : Word;       // Number of bits per sample of mono data
  end;

  TDataChunk = Record
    DataStr : Array[0..3] of Char;
    DataLength : Integer;
  end;

  TRiffHeader = Record
    RiffStr : Array[0..3] of Char;
    Length : Integer;
  end;
 {$A+}

procedure SaveSound(Stream : TFileStream; Length : Integer; Name : String);
procedure CopySound(Stream : TFileStream; Length : Integer);

implementation

uses
  SysUtils, Windows, ClipBrd;

procedure SaveSoundToStream(Incoming, Outgoing : TStream; IncomingLen : Integer);
var
  SoundHeader : TSoundHeader;
  RiffHeader : TRiffHeader;
  WaveChunk : TWaveChunk;
  DataChunk : TDataChunk;
  DataSize : Integer;
  Buffer : Pointer;
begin
  Incoming.Read(SoundHeader, SizeOf(SoundHeader));
  DataSize := IncomingLen - SizeOf(SoundHeader);

  with RiffHeader do
    begin
      RiffStr := 'RIFF';
      Length := DataSize + SizeOf(WaveChunk) + SizeOf(DataChunk);
    end;

  with WaveChunk do
    begin
      WaveStr := 'WAVEfmt ';
      FmtLength := SizeOf(TWaveChunk) - 12;
      Tag := 1;
      Channels := 1;//2;  // Sterio
      Bits := 16;
      Rate := 22050;// 11025;
      BPS := Rate * 4;  // Bytes per sec
      Align := 4;
    end;

  with DataChunk do
    begin
      DataStr := 'data';
      DataLength := DataSize;
    end;

  GetMem(Buffer, DataSize);
  try
    Incoming.Read(Buffer^, DataSize);
    Outgoing.Write(RiffHeader, SizeOf(RiffHeader));
    Outgoing.Write(WaveChunk, SizeOf(WaveChunk));
    Outgoing.Write(DataChunk, SizeOf(DataChunk));
    Outgoing.Write(Buffer^, DataSize);
  finally
    FreeMem(Buffer, DataSize);
  end;
end;

procedure SaveSound(Stream : TFileStream; Length : Integer; Name : String);
var
  OutFile : TFileStream;
  Padding : Integer;
begin
  OutFile := TFileStream.Create(Name, fmCreate or fmShareExclusive);
  try
    SaveSoundToStream(Stream, OutFile, Length);
    Padding := 0;
    OutFile.Write(Padding, SizeOf(Padding));
  finally
    OutFile.Free;
  end;
end;  // SaveSound

procedure CopySound(Stream : TFileStream; Length : Integer);
var
  MemStream : TMemoryStream;
  Data: THandle;
  DataPtr: Pointer;
begin
  MemStream := TMemoryStream.Create;
  try
    MemStream.Size :=
      SizeOf(TRiffHeader) +
      SizeOf(TWaveChunk) +
      SizeOf(TDataChunk) +
      (Length - SizeOf(TSoundHeader));

    SaveSoundToStream(Stream, MemStream, Length);

    Data := GlobalAlloc(GMEM_MOVEABLE, MemStream.Size);
    try
      DataPtr := GlobalLock(Data);
      try
        Move(MemStream.Memory^, DataPtr^, MemStream.Size);
        Clipboard.SetAsHandle(CF_WAVE, Data);
      finally
        GlobalUnlock(Data);
      end;
    except
      GlobalFree(Data);
      raise;
    end;
  finally
    MemStream.Free;
  end;
end;  // CopySound

end.
