unit UOArt;

interface

uses
  Classes, Graphics, Dialogs, IndexedFile;

const
  CACHE_SIZE = 50;

type
 {$A-}
  TArtIndexRec = record
    Lookup : Integer;
    Length : Integer;
    Unknown : Integer;
  end;
 {$A+}

  TBlockStyle = (bsInvalid, bsRaw, bsRun);

  TArtCacheNode = class (TObject)
  private
    ID : Integer;
    Width, Height : Word;
    Style : TBlockStyle;
    Data : TMemoryStream;
  public
    constructor Create(AID, ASize : Integer; Stream : TStream);
    destructor Destroy; override;
  end;

  TArtCache = class (TObject)
  private
    Cache : Array[0..CACHE_SIZE-1] of TArtCacheNode;
    procedure NeedCacheBlock(I : Integer);
  public
    destructor Destroy; override;
    constructor DrawCacheBlock(Idx : Integer; Bitmap : TBitmap; AX, AY : Integer);
    function GetCacheBlockInfo(Idx : Integer; var Width, Height : Word; var Style : TBlockStyle) : Boolean;
  end;

  TArtCacheLookup = class (TArtCache)
  private
    FData : TList;
    procedure NeedBlock(Lookup, Size, Source : Integer);
  public
    constructor Create(Data : Array of TStream);
    procedure DrawBlock(Lookup, Size, Source : Integer; Bitmap : TBitmap; AX, AY : Integer);
    function GetBlockInfo(Lookup, Size, Source : Integer; var Width, Height : Word; var Style : TBlockStyle) : Boolean;
  end;

  TArtCacheIndex = class (TArtCache)
  private
    FData,
    FIndex : TStream;
    procedure NeedBlock(BlockNum : Integer);
  public
    constructor Create(Index, Data : TStream);
    procedure DrawBlock(BlockNum : Integer; Bitmap : TBitmap; AX, AY : Integer);
    function GetBlockInfo(BlockNum : Integer; var Width, Height : Word; var Style : TBlockStyle) : Boolean;
  end;

procedure GetBlockInfo(Data : TIndexedData; var Width, Height : Word; var Style : TBlockStyle);
procedure DrawBlock(Data : TIndexedData; Bitmap : TBitmap; AX, AY : Integer);

implementation

uses
  SysUtils, Utility;

procedure LoadRaw(Stream : TStream; Size : Integer; BitMap : TBitMap; AX, AY : Integer);
var
  Y : Integer;
  P : PWordArray;
begin
  For Y := 0 to 21 do
    begin
      P := BitMap.ScanLine[AY+Y];
      Stream.Read(P^[AX+22-(Y+1)], (Y+1) * 4);
    end;
  For Y := 0 to 21 do
    begin
      P := BitMap.ScanLine[AY+22+Y];
      Stream.Read(P^[AX+Y], (22-Y) * 4);
    end;
end;

procedure LoadRun(Stream : TStream; Size : Integer; BitMap : TBitmap; AX, AY : Integer);
const
  LineOffsetsSize = 5120;
var
  XOffs, Run : Word;
  X, Y, Dif : Integer;
  Width, Height : Word;
  DataSize, Trigger : Word;
  DataStart : Integer;
  LineDone : Boolean;
  P : PWordArray;
  Cnt, DCnt, V : Integer;
  LineOffsets : Array[0..LineOffsetsSize-1] of Word;

begin
  Stream.Read(DataSize, SizeOf(DataSize));  // Size of overall block??
  Stream.Read(Trigger, SizeOf(Trigger));

   // Width
  Stream.Read(Width, SizeOf(Width));
  if Width = 0 then Exit;

   // Height
  Stream.Read(Height, SizeOf(Height));
  if (Height = 0) or (Height * 2 > LineOffsetsSize) then Exit;
  Stream.Read(LineOffsets, Height * 2);

  DataStart := Stream.Position;

   // Data
  Y := 0;
  while (Y < Height) and (Y < BitMap.Height) do
    begin
      Stream.Seek(DataStart + LineOffsets[Y] * 2, soFromBeginning);
      Stream.Read(XOffs, SizeOf(XOffs));
      Stream.Read(Run, SizeOf(Run));
      X := XOffs;
      if (AY+Y >= 0) and (AY+Y<Bitmap.Height) then
        begin
          P := BitMap.ScanLine[AY+Y];

          LineDone := False;
          while not LineDone do
            begin
              if Run > 0 then
                begin
                  Cnt := Run;
                  if Cnt*2 > SizeOf(LineOffsets) then Exit;

                  if AX+X < BitMap.Width then
                    begin
                      Dif := Min(AX+X, 0);
                      if Dif < 0 then
                        Stream.Seek(-Dif*2, soFromCurrent);
                      DCnt := Min(BitMap.Width - (AX+X), Run + Dif);
                      if DCnt > 0 then
                        Stream.Read(P^[AX+X-Dif], DCnt*2);
                      Cnt := Cnt - DCnt;
                    end;

                  Stream.Seek(Cnt*2, soFromCurrent);
                  if Cnt > 0 then LineDone := True;
                  Stream.Read(V, SizeOf(V));
                  X := X + Run;
                end
              else
                V := 0;

              if V = 0 then
                LineDone := True
              else
                begin
                  XOffs := V and $FFFF;
                  Run := (V shr 16) and $FFFF;
                  X := X + XOffs;
                end;
            end;
        end;

      Inc(Y);
    end;
end;

constructor TArtCacheNode.Create(AID, ASize : Integer; Stream : TStream);
var
  Flag : Integer;
begin
  inherited Create;
  ID := AID;

  Data := TMemoryStream.Create;
  //Data.Size := ASize;
  Data.Seek(0, soFromBeginning);
  Data.CopyFrom(Stream, ASize);

  Data.Seek(0, soFromBeginning);
  Data.Read(Flag, SizeOf(Flag));
  if (Flag > $FFFF) or (Flag = 0) then
    begin
      Style := bsRaw;
      Width := 44;
      Height := 44;
    end
  else
    begin
      Style := bsRun;
      Data.Read(Width, SizeOf(Width));
      Data.Read(Height, SizeOf(Height));
      if (Width >= 1024) or (Height >= 1024) then
        begin
          Style := bsInvalid;
          Width := 0;
          Height := 0;
        end;
    end;
end;

destructor TArtCacheNode.Destroy;
begin
  Data.Free;
  inherited Destroy;
end;

 //----------

destructor TArtCache.Destroy;
var
  I : Integer;
begin
  For I := 0 to CACHE_SIZE-1 do
    Cache[I].Free;
  inherited Destroy;
end;

procedure TArtCache.NeedCacheBlock(I : Integer);
var
  CacheNode : TArtCacheNode;
begin
  if (I > 0) and (I < CACHE_SIZE) then
    begin
      CacheNode := Cache[I];
      Move(Cache[0], Cache[1], I * SizeOf(Pointer));
      Cache[0] := CacheNode;
    end;
end;

constructor TArtCache.DrawCacheBlock(Idx : Integer; Bitmap : TBitmap; AX, AY : Integer);
begin
  if Cache[Idx] = Nil then Exit;
  Cache[Idx].Data.Seek(0, soFromBeginning);
  Case Cache[Idx].Style of
    bsRaw : LoadRaw(Cache[Idx].Data, Cache[Idx].Data.Size, Bitmap, AX, AY);
    bsRun : LoadRun(Cache[Idx].Data, Cache[Idx].Data.Size, Bitmap, AX, AY);
  end;
end;

function TArtCache.GetCacheBlockInfo(Idx : Integer; var Width, Height : Word; var Style : TBlockStyle) : Boolean;
begin
  if Cache[Idx] = Nil then
    begin
      Width := 0;
      Height := 0;
      Style := bsInvalid;
      Result := False;
    end
  else
    begin
      Width := Cache[Idx].Width;
      Height := Cache[Idx].Height;
      Style := Cache[Idx].Style;
      Result := Cache[Idx].Style <> bsInvalid;
    end;
end;


 //------

constructor TArtCacheLookup.Create(Data : Array of TStream);
var
  I : Integer;
begin
  inherited Create;
  FData := TList.Create;
  For I := Low(Data) to High(Data) do
    FData.Add(Data[I]);
end;

procedure TArtCacheLookup.NeedBlock(Lookup, Size, Source : Integer);
var
  CacheNode : TArtCacheNode;
  I : Integer;
begin
  I := 0;
  while (I < CACHE_SIZE) and ((Cache[I] = NIL) or (Cache[I].ID <> Lookup)) do
    Inc(I);

  if (I >= CACHE_SIZE) then
    begin
      if Lookup < 0 then
        CacheNode := Nil
      else
        begin
          TStream(FData[Source]).Seek(Lookup, soFromBeginning);
          CacheNode := TArtCacheNode.Create(Lookup, Size, TStream(FData[Source]));
        end;

      Cache[CACHE_SIZE-1].Free;
      Move(Cache[0], Cache[1], (CACHE_SIZE-1) * SizeOf(Pointer));
      Cache[0] := CacheNode;
    end
  else
    NeedCacheBlock(I);
end;

procedure TArtCacheLookup.DrawBlock(Lookup, Size, Source : Integer; Bitmap : TBitmap; AX, AY : Integer);
begin
  NeedBlock(Lookup, Size, Source);
  DrawCacheBlock(0, Bitmap, AX, AY);
end;

function TArtCacheLookup.GetBlockInfo(Lookup, Size, Source : Integer; var Width, Height : Word; var Style : TBlockStyle) : Boolean;
begin
  NeedBlock(Lookup, Size, Source);
  Result := GetCacheBlockInfo(0, Width, Height, Style);
end;

 //------

constructor TArtCacheIndex.Create(Index, Data : TStream);
begin
  inherited Create;
  FData := Data;
  FIndex := Index;
end;

procedure TArtCacheIndex.NeedBlock(BlockNum : Integer);
var
  I : Integer;
  ArtIndexRec : TArtIndexRec;
  CacheNode : TArtCacheNode;
begin
  I := 0;
  while (I < CACHE_SIZE) and ((Cache[I] = NIL) or (Cache[I].ID <> BlockNum)) do
    Inc(I);

  if (I >= CACHE_SIZE) then
    begin
      FIndex.Seek(BlockNum * 12, soFromBeginning);
      FIndex.Read(ArtIndexRec, SizeOf(ArtIndexRec));
      if ArtIndexRec.Lookup < 0 then
        CacheNode := Nil
      else
        begin
          FData.Seek(ArtIndexRec.Lookup, soFromBeginning);
          CacheNode := TArtCacheNode.Create(BlockNum, ArtIndexRec.Length, FData);
        end;

      Cache[CACHE_SIZE-1].Free;
      Move(Cache[0], Cache[1], (CACHE_SIZE-1) * SizeOf(Pointer));
      Cache[0] := CacheNode;
    end
  else
    NeedCacheBlock(I);
end;

procedure TArtCacheIndex.DrawBlock(BlockNum : Integer; Bitmap : TBitmap; AX, AY : Integer);
begin
  NeedBlock(BlockNum);
  DrawCacheBlock(0, Bitmap, AX, AY);
end;

function TArtCacheIndex.GetBlockInfo(BlockNum : Integer; var Width, Height : Word; var Style : TBlockStyle) : Boolean;
begin
  NeedBlock(BlockNum);
  Result := GetCacheBlockInfo(0, Width, Height, Style);
end;

procedure GetBlockInfo(Data : TIndexedData; var Width, Height : Word; var Style : TBlockStyle);
var
  Flag : Integer;
begin
  if Data = Nil then Exit;
  with Data do
    begin
      Data.Seek(0, soFromBeginning);
      Data.Read(Flag, SizeOf(Flag));
      if (Flag > $FFFF) or (Flag = 0) then
        begin
          Style := bsRaw;
          Width := 44;
          Height := 44;
        end
      else
        begin
          Style := bsRun;
          Data.Read(Width, SizeOf(Width));
          Data.Read(Height, SizeOf(Height));
          if (Width >= 1024) or (Height >= 1024) then
            begin
              Style := bsInvalid;
              Width := 0;
              Height := 0;
            end;
        end;
    end;
end;

procedure DrawBlock(Data : TIndexedData; Bitmap : TBitmap; AX, AY : Integer);
var
  Width, Height : Word;
  Style : TBlockStyle;
begin
  if Data = Nil then Exit;
  GetBlockInfo(Data, Width, Height, Style);
  Data.Data.Seek(0, soFromBeginning);
  Case Style of
    bsRaw : LoadRaw(Data.Data, Data.Data.Size, Bitmap, AX, AY);
    bsRun : LoadRun(Data.Data, Data.Data.Size, Bitmap, AX, AY);
  end;
end;



end.
